/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.dataobj;
import ostore.util.*;
import ostore.security.*;

/**
 * A certificate mapping an AGUID to the public key of an inner ring.
 *
 * @author  Sean C. Rhea
 * @version $Id: PRCert.java,v 1.16 2002/07/21 20:57:34 srhea Exp $
 */
public class PRCert implements QuickSerializable {

    public SecureHash aguid;
    public QSPublicKey pr_key;

    public PRCert(SecureHash aguid, QSPublicKey pr_key) {
	this.aguid = aguid;
	this.pr_key = pr_key;
    }

    public PRCert (InputBuffer buffer) throws QSException {
	aguid = (SecureHash) buffer.nextObject ();
	pr_key = (QSPublicKey) buffer.nextObject ();
    }

    public void serialize (OutputBuffer buffer) {
        buffer.add (aguid);
        buffer.add (pr_key);
    }

    public boolean equals (Object obj) {
	if (obj == null)
	    return false;
	PRCert rhs = (PRCert) obj;
	if (! aguid.equals (rhs.aguid))
	    return false;
	return pr_key.equals (rhs.pr_key);
    }

    public String toString () {
	return "(PRCert " + aguid + " " + new SHA1Hash (pr_key) + ")";
    }
}

