/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.monitor;

import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.SinkIF;
import ostore.util.*;
import java.util.*;

/**
 * Resource managed dynamically by VM or OS.
 *
 * @author  Jeff Pang
 * @version $Id: DynamicResource.java,v 1.2 2003/02/18 05:33:20 jeffpang Exp $
 **/
public abstract class DynamicResource extends Resource
{
    private int sampleInterval;
    
    /**
     * @param sampleInterval time between samples in ms
     * @param initial initial resource utilization
     */
    public DynamicResource(NodeId id, String name, 
			   double[] initial, int sampleInterval)
    {
	super(id, name, initial);
	this.sampleInterval = sampleInterval;
    }

    public DynamicResource(NodeId id, String name, 
			   double initial, int sampleInterval)
    {
	super(id, name, initial);
	this.sampleInterval = sampleInterval;
    }
    
    /**
     * Take a sample at the current time.
     */
    public abstract void sample();

    /**
     * Get the sample interval
     */
    public int getSampleInterval()
    {
	return sampleInterval;
    }
}
