/**
 * Network.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.network;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StreamTokenizer;
import java.net.InetAddress;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.dispatch.Signal;
import ostore.util.AssertionViolatedException;
import ostore.util.ByteUtils;
import ostore.util.Carp;
import ostore.util.CountBuffer;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.Pair;
import ostore.util.PriorityQueue;
import ostore.util.QSException;
import ostore.util.Queue;
import ostore.util.QuickSerializable;
import ostore.util.QuickSerializable;
import ostore.util.QuickSerializable;
import ostore.util.TypeTable;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.NoSuchStageException;
import seda.sandStorm.api.NoSuchStageException;
import seda.sandStorm.api.ProfilableIF;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SingleThreadedEventHandlerIF;
import seda.sandStorm.api.SinkClosedEvent;
import seda.sandStorm.api.SinkClosedException;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.SinkFlushedEvent;
import seda.sandStorm.api.SinkFullException;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.StagesInitializedSignal;
import seda.sandStorm.core.BufferElement;
import seda.sandStorm.internal.SinkProxy;
import seda.sandStorm.lib.aSocket.ATcpClientSocket;
import seda.sandStorm.lib.aSocket.ATcpConnectFailedEvent;
import seda.sandStorm.lib.aSocket.ATcpConnection;
import seda.sandStorm.lib.aSocket.ATcpInPacket;
import seda.sandStorm.lib.aSocket.ATcpListenFailedEvent;
import seda.sandStorm.lib.aSocket.ATcpListenSuccessEvent;
import seda.sandStorm.lib.aSocket.ATcpServerSocket;
import seda.sandStorm.lib.aSocket.aSocketInputStream;
import soss.network.TransitStub;

/**
 * TODO
 *
 * @author     Sean C. Rhea
 * @version    $Id: SimpleTest.java,v 1.2 2002/08/09 23:19:20 srhea Exp $
 */
public class SimpleTest 
implements SingleThreadedEventHandlerIF, EventHandlerIF {

    protected String TAG = "SimpleTest";
    protected static final boolean DEBUG = false;

    protected NodeId lowest_nodeid;
    protected SinkIF this_sink;
    protected Map nodeid_to_classifier;
    protected LinkedList peers = new LinkedList ();
    protected Random rand;
    protected int msgnum, msgcnt, rcvcnt;
    protected boolean start;

    public static class Msg extends NetworkMessage {
	public int value;
	public Msg (NodeId peer, int v) { super (peer, false); value = v; }
	public Msg (InputBuffer buffer) throws QSException {
	    super (buffer);
	    value = buffer.nextInt ();
	}
	public void serialize (OutputBuffer buffer) {
	    super.serialize (buffer);
	    buffer.add (value);
	}
	public Object clone () throws CloneNotSupportedException {
	    Msg result = (Msg) super.clone ();
	    result.value = value;
	    return result;
	}
    }

    public void init (ConfigDataIF config) throws Exception {

	// Set up some debug stuff.

	TAG = config.getStage().getName();
        if (DEBUG) Debug.turnOn (TAG);

	// Set this_sink.

	this_sink = config.getStage().getSink();

	// Create all of our filters.

	Filter nm_filter = new Filter ();
	if (! nm_filter.requireType (
		    Class.forName ("ostore.network.SimpleTest$Msg")))
	    BUG (TAG + ": could not require type.");
	if (! nm_filter.requireValue ("inbound", new Boolean (true)))
	    BUG (TAG + ": could not require inbound.");

	// Read in all of the node_ids we represent, find their
	// classifiers, and register our filters.

	int node_id_count = config.getInt ("node_id_count");
	nodeid_to_classifier = new HashMap (node_id_count);
	for (int i = 0; i < node_id_count; ++i) {
	    NodeId n = new NodeId (config.getString ("node_id_" + i));
	    if (DEBUG) Debug.printtagln (TAG, n.toString ());
	    Classifier c = Classifier.getClassifier (n);
	    nodeid_to_classifier.put (n, Classifier.getClassifier (n));
	    c.subscribe (nm_filter, this_sink);
	    if (i == 0) {
		Filter init_filter = new Filter ();
		if (! init_filter.requireType (Class.forName (
				"seda.sandStorm.api.StagesInitializedSignal")))
		    BUG (TAG + ": could not require type.");
		c.subscribe (init_filter, this_sink);
	    }
	}

	int peer_count = config.getInt ("peer_count");
	for (int i = 0; i < peer_count; ++i) {
	    NodeId n = new NodeId (config.getString ("peer_" + i));
	    peers.addLast (n);
	}

	msgnum = 0;
	rand = new Random (config.getInt ("seed"));
	start = config.getBoolean ("start");
	msgcnt = config.getInt ("msgcnt");

	TypeTable.register_type ("ostore.network.SimpleTest$Msg");
    }

    public void handleEvent (QueueElementIF item) 
    throws EventHandlerException {

	if (DEBUG) Debug.printtagln (TAG, "event " + item);

	if (item instanceof StagesInitializedSignal) {
	    if (start) {
		send_new_msg ();
		send_new_msg ();
	    }
	}
	else if (item instanceof Msg) {
	    ++rcvcnt;
	    if (rcvcnt % 2 == 0) {
		send_new_msg ();
		send_new_msg ();
	    }
	}
	else {
	    Carp.die (TAG + ": got unknown event of type " + 
		    item.getClass ().getName () + ": " + item + ".");
	}
    }

    public void handleEvents(QueueElementIF items[]) 
    throws EventHandlerException {
	for(int i=0; i<items.length; i++)
	    handleEvent(items[i]);
    }

    public void destroy() throws Exception {
    }

    protected void send_new_msg () {

	if ((! start) && (msgnum >= msgcnt)) {
	    System.err.println ("Success!");
	    System.exit (0);
	}

	int s = (0x7ffffff & rand.nextInt ()) % nodeid_to_classifier.size ();
	int d = (0x7ffffff & rand.nextInt ()) % peers.size ();
	Iterator i = nodeid_to_classifier.keySet ().iterator ();
	while (i.hasNext () && (s-- > 0))
	    i.next ();
	NodeId src = (NodeId) i.next ();
	i = peers.iterator ();
	while (i.hasNext () && (d-- > 0))
	    i.next ();
	NodeId dst = (NodeId) i.next ();

	Classifier classifier = (Classifier) nodeid_to_classifier.get (src);
	Msg m = new Msg (dst, msgnum++);

	if (DEBUG) Debug.printtagln (TAG, "Sending " + m);

	try {
	    classifier.dispatch (m);
	}
	catch (SinkFullException e) {
	    BUG (TAG + ": SFE");
	}
    }

    protected static final void BUG (String msg) {
	Carp.die (msg + "  This is a bug in ostore.network.Network.  " +
		"Please let Sean know about it ASAP.");
    }
}

