/**
 * MonitorStopMsg.java
 *
 * @author Jeremy Stribling
 * @version    $Id: MonitorStopMsg.java,v 1.3 2003/05/08 01:55:05 strib Exp $
 *
 * Copyright (c) 2003 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network.patchwork;
import ostore.util.NodeId;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkIF;

/**
 * A <code>MonitorStopMsg</code> is sent by a stage when it knows it
 * will no longer (at least for a while) want statistics from Patchwork
 * about a given set of nodes.  Patchwork will stop monitoring them until
 * further notice.
 *
 * @author Jeremy Stribling
 * @version    $Id: MonitorStopMsg.java,v 1.3 2003/05/08 01:55:05 strib Exp $
 */
public class MonitorStopMsg implements QueueElementIF {

    public NodeId[] peers;
    public Object userData;
    public SinkIF sink;

    public MonitorStopMsg (NodeId peer, Object userData, SinkIF sink) {
	NodeId[] tmp = {peer}; 
	peers = tmp;
	this.userData = userData;
	this.sink = sink;
    }

    public MonitorStopMsg (NodeId[] peers, Object userData, SinkIF sink) {
	this.peers = peers;
	this.userData = userData;
	this.sink = sink;
    }

    public Object clone () throws CloneNotSupportedException {
	MonitorStopMsg newone = (MonitorStopMsg) super.clone ();
	newone.peers = peers;
	newone.userData = userData;
	newone.sink = sink;
	return newone;
    }

    public String toString () {
	String s = "(MonitorStopMsg, peers: {";
	for( int i = 0; i < peers.length; i++ ) {
	    s += peers[i] + " ";
	}
	s += "}, userData: " + userData + ")";
	return s;
    }
}
