/**
 * Key with a type and provider.
 *
 * @author     Sean C. Rhea
 * @version    $Id: GenerateRSAKeyPair.java,v 1.6 2002/08/09 00:13:12 czerwin Exp $
 */

package ostore.security;
import ostore.util.*;
import java.security.*;
import java.io.*;

public class GenerateRSAKeyPair {

    /**
     * Generate a key pair with the default seed.  This is the preferred
     * method for generating keys.
     *
     * @param size The size of the key to generate
     * @param debug Show debug output.
     *
     * @return The keypair generated.
     **/

    public static KeyPair genkeypair (int size, boolean debug) {
      return genkeypair(size,null,debug);
    }

    /**
     * Generate a key pair with the following seed.  Note, specifying
     * your own seed will result in less random keys then using the default
     * seeding mechanism.
     *
     * @param size The size of the key to generate
     * @param seed The seed to use.  If null, seeds in default manner.
     * @param debug Show debug output.
     *
     * @return The keypair generated.
     **/
  
    public static KeyPair genkeypair (int size, byte seed[], boolean debug) {
	KeyPair keypair = null;
	try {
	    Provider cryptix = new cryptix.jce.provider.CryptixCrypto ();
	    Security.addProvider (cryptix);

	    Provider cryptixRandom = new cryptix.jce.provider.CryptixRandom ();
	    Security.addProvider (cryptixRandom);
	    SecureRandom secure_random = null;

            if (seed == null)
                secure_random = SecureRandom.getInstance ("DevRandom", "CryptixRandom");
            else
              secure_random = new SecureRandom(seed);
	    KeyPairGenerator kpgen = 
		KeyPairGenerator.getInstance ("RSA", "CryptixCrypto");
	    kpgen.initialize (size, secure_random);
	    keypair = kpgen.generateKeyPair ();
	}
	catch (Exception e) {
	    if (debug) Carp.out ("Caught exception " + e 
		    + " while trying to generate a key pair.");
	    return null;
	}
	return keypair;
    }

    public static QSPublicKey qs_pkey (KeyPair kp) {
	return new QSRsaPublicKey (kp.getPublic ());
    }

    public static QSPrivateKey qs_skey (KeyPair kp) {
	return new QSRsaPrivateKey (kp.getPrivate ());
    }

    public static void main (String [] argv) {

	if (argv.length < 2) {
	    System.err.println ("usage: genkey <prefix> <size>");
	    System.exit (1);
	}

	int which = 0;
	String prefix = argv[which++];
	int keysize = Integer.parseInt (argv[which++]);

	KeyPair keypair = genkeypair (keysize, true);
	if (keypair == null) {
	    System.exit (1);
	}

	QSPublicKey  pub  = qs_pkey (keypair);
	QSPrivateKey priv = qs_skey (keypair);

	try {
	    QSIO.write (pub, prefix + ".pub");
	}
	catch (IOException e) {
	    System.err.println ("Could not write " + prefix + ".pub");
	    System.exit (1);
	}

	try {
	    QSIO.write (priv, prefix + ".priv");
	}
	catch (IOException e) {
	    System.err.println ("Could not write " + prefix + ".priv");
	    System.exit (1);
	}
    }
}
