/**
 * QSDHPublicKey.
 *
 * @author      Gifford Cheung
 * @version     $Id: QSDHPublicKey.java,v 1.12 2002/08/02 19:39:56 geels Exp $
 */

package ostore.security;

import java.math.BigInteger;
import java.security.*;
import java.util.*;
import javax.crypto.interfaces.*;
import javax.crypto.spec.*;

import ostore.util.*;

/**
 * QuickSerializable Diffie-Hellman public key. Based on the same
 * the Cryptix JCE DHPublicCryptix v1.1 and QSPublicKey
 *
 * @version $Revision 1.0$
 * @author Gifford Cheung 
 */
public class QSDHPublicKey implements QSPublicKey {

	public QSDHPublicKey (DHPublicKey key) {
		if( key == null ) Carp.die( "QSDHPublicKey: key must be non-null.");
		_key = new MyRawDHPublicKey(key.getY(), key.getParams());
	}

	public QSDHPublicKey(InputBuffer buffer) throws QSException {
		BigInteger y,p,g;
		int l;
		y = buffer.nextBigInteger ();
		p = buffer.nextBigInteger ();
		g = buffer.nextBigInteger ();
		l = buffer.nextInt ();
		_key = new MyRawDHPublicKey( y, new DHParameterSpec( p, g, l));
	}

	public void serialize (OutputBuffer buffer ) {
	    buffer.add (_key.getY());
	    buffer.add (_key.getParams().getP());
	    buffer.add (_key.getParams().getG());
	    buffer.add (_key.getParams().getL());
	}
	
	public PublicKey key() { return _key; }

	public boolean equals( Object obj) {
		boolean e;
		if (obj == null) 
			e = false;
		else if (obj == this) 
			e = true;
		else if (!(obj instanceof QSDHPublicKey))
			e = false;
		else {
			QSDHPublicKey otherkey = (QSDHPublicKey) obj;
			DHPublicKey key = (DHPublicKey) otherkey.key();
			// y 
			boolean y =
			_key.getY().equals( key.getY());
			// p
			boolean p =
			_key.getParams().getP().equals( key.getParams().getP());
			// g
			boolean g =
			_key.getParams().getG().equals( key.getParams().getG());
			// l
			boolean l =
			_key.getParams().getL() == key.getParams().getL();
			e = (y && p && g && l);
		}
		return e;
			
	}

	DHPublicKey _key;

	private class MyRawDHPublicKey implements DHPublicKey {

		private final BigInteger y;
		private final DHParameterSpec params;

		public MyRawDHPublicKey(BigInteger y, DHParameterSpec params) {
			this.y = y;
			this.params = params;
		}

		public BigInteger getY() { return this.y; }
		public DHParameterSpec getParams() { return this.params; }
		public String getAlgorithm() { return "DH"; }
		public String getFormat() { return "RAW"; }
		public byte[] getEncoded() {
			//TODO uh... not used? 
			Carp.die("Not implemented");
			return new byte[0];
		}
	}
}
	
