
#include <jni.h>
#include "ostore_security_thresh_NativeIF.h"
#include "../gmp-interface.h"
#include <stdio.h>
#include <gmp.h>
#include <iostream>
#include <openssl/evp.h>
#include <assert.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>

#define PROFILE false

JNIEXPORT void JNICALL 
Java_ostore_security_thresh_NativeIF_sign_1using_1c_1helper (
	JNIEnv *env, 
	jclass not_used, 
	jobject jx,
	jobject jn,
	jobject jr,
	jobject jgroupVerifier,
	jobject jdelta,
	jobject jverifier,
	jobject jsignVal,
	jobject jsecret,
	jobjectArray results,
	jboolean debug) 
{
    jclass big_int_class = env->FindClass ("java/math/BigInteger");
    jmethodID big_int_init = 
	env->GetMethodID (big_int_class, "<init>", "([B)V");
    if (big_int_init == 0) 
	return;
    jmethodID big_int_to_byte_array = 
	env->GetMethodID (big_int_class, "toByteArray", "()[B");
    if (big_int_to_byte_array == 0) 
	return;

    mpz_t TWO, FOUR;
    mpz_init_set_ui (TWO, 2);
    mpz_init_set_ui (FOUR, 4);

    mpz_t x, n, r, groupVerifier, delta, verifier, signVal, secret;
    mpz_init (x);
    mpz_init (n);
    mpz_init (r);
    mpz_init (groupVerifier);
    mpz_init (delta);
    mpz_init (verifier);
    mpz_init (signVal);
    mpz_init (secret);
    bigint2mpz (x, env, jx, big_int_to_byte_array);
    if (debug) print_mpz ("X", x);
    bigint2mpz (n, env, jn, big_int_to_byte_array);
    if (debug) print_mpz ("N", n);
    bigint2mpz (r, env, jr, big_int_to_byte_array);
    if (debug) print_mpz ("R", r);
    bigint2mpz (groupVerifier, env, jgroupVerifier, big_int_to_byte_array);
    if (debug) print_mpz ("GROUPERIFIER", groupVerifier);
    bigint2mpz (delta, env, jdelta, big_int_to_byte_array);
    if (debug) print_mpz ("DELTA", delta);
    bigint2mpz (verifier, env, jverifier, big_int_to_byte_array);
    if (debug) print_mpz ("VERIFIER", verifier);
    bigint2mpz (signVal, env, jsignVal, big_int_to_byte_array);
    if (debug) print_mpz ("SIGNVAL", signVal);
    bigint2mpz (secret, env, jsecret, big_int_to_byte_array);
    if (debug) print_mpz ("SECRET", secret);

    mpz_t tmp1, tmp2;
    mpz_init (tmp1);
    mpz_init (tmp2);

    struct timeval start, finish, tdiff;
    if (PROFILE) {
	memset (&start, 0, sizeof (start));
	memset (&finish, 0, sizeof (start));
	gettimeofday (&start, 0);
    }

    // BigInteger vprime = groupVerifier.modPow (r,n);
    mpz_t vprime;
    mpz_init (vprime);
    mpz_powm (vprime, groupVerifier, r, n);

    if (PROFILE) {
	gettimeofday (&finish, 0);
	timersub (&finish, &start, &tdiff);
	cout << "1. Took " << tdiff.tv_sec << " seconds and "
	    << tdiff.tv_usec << " microseconds." << endl;
	gettimeofday (&start, 0);
    }

    // BigInteger xtilde = x.modPow(FOUR.multiply(delta),n);
    mpz_mul (tmp1, FOUR, delta);

    if (PROFILE) {
	gettimeofday (&finish, 0);
	timersub (&finish, &start, &tdiff);
	cout << "2. Took " << tdiff.tv_sec << " seconds and "
	    << tdiff.tv_usec << " microseconds." << endl;
	gettimeofday (&start, 0);
    }

    mpz_t xtilde;
    mpz_init (xtilde);
    mpz_powm (xtilde, x, tmp1, n);

    if (PROFILE) {
	gettimeofday (&finish, 0);
	timersub (&finish, &start, &tdiff);
	cout << "3. Took " << tdiff.tv_sec << " seconds and "
	    << tdiff.tv_usec << " microseconds." << endl;
	gettimeofday (&start, 0);
    }

    // BigInteger xprime = xtilde.modPow(r,n);
    mpz_t xprime;
    mpz_init (xprime);
    mpz_powm (xprime, xtilde, r, n);

    if (PROFILE) {
	gettimeofday (&finish, 0);
	timersub (&finish, &start, &tdiff);
	cout << "4. Took " << tdiff.tv_sec << " seconds and "
	    << tdiff.tv_usec << " microseconds." << endl;
	gettimeofday (&start, 0);
    }

    EVP_MD_CTX md;
    EVP_DigestInit (&md, EVP_sha1 ());
    int len = 0;
    unsigned char *buf = 0;

    // md.update(groupVerifier.mod(n).toByteArray());

    mpz_mod (tmp1, groupVerifier, n);
    if (debug) print_mpz ("V", tmp1);
    add_to_hash (&md, tmp1, debug);
    // md.update(xtilde.toByteArray());
    if (debug) print_mpz ("XTILDE", xtilde);
    add_to_hash (&md, xtilde, debug);
    // md.update(verifier.mod(n).toByteArray());

    if (PROFILE) {
	gettimeofday (&finish, 0);
	timersub (&finish, &start, &tdiff);
	cout << "5. Took " << tdiff.tv_sec << " seconds and "
	    << tdiff.tv_usec << " microseconds." << endl;
	gettimeofday (&start, 0);
    }

    mpz_mod (tmp1, verifier, n);
    if (debug) print_mpz ("VI", tmp1);
    add_to_hash (&md, tmp1, debug);
    // md.update(x.modPow(signVal,n).modPow(TWO,n).toByteArray());
    mpz_t xi;
    mpz_init (xi);

    if (PROFILE) {
	gettimeofday (&finish, 0);
	timersub (&finish, &start, &tdiff);
	cout << "6. Took " << tdiff.tv_sec << " seconds and "
	    << tdiff.tv_usec << " microseconds." << endl;
	gettimeofday (&start, 0);
    }

    mpz_powm (xi, x, signVal, n);

    if (PROFILE) {
	gettimeofday (&finish, 0);
	timersub (&finish, &start, &tdiff);
	cout << "7. Took " << tdiff.tv_sec << " seconds and "
	    << tdiff.tv_usec << " microseconds." << endl;
	gettimeofday (&start, 0);
    }

    mpz_powm (tmp2, xi, TWO, n);
    if (debug) print_mpz ("XI^2", tmp2);
    add_to_hash (&md, tmp2, debug);
    // md.update(vprime.toByteArray());
    if (debug) print_mpz ("V'", vprime);
    add_to_hash (&md, vprime, debug);
    // md.update(xprime.toByteArray());
    if (debug) print_mpz ("X'", xprime);
    add_to_hash (&md, xprime, debug);

    // c = new BigInteger(md.digest()).mod(n);
    unsigned char md_result [EVP_MAX_MD_SIZE];
    EVP_DigestFinal (&md, md_result, 0);

    if (debug) {
	cout << "c: hash = " << hex;
	for (int i = 0; i < 20; ++i) 
	    cout << (((int) md_result [i]) & 0xff);
	cout << dec << endl;
    }

    mag2mpz (tmp1, md_result, 20);

    if (PROFILE) {
	gettimeofday (&finish, 0);
	timersub (&finish, &start, &tdiff);
	cout << "8. Took " << tdiff.tv_sec << " seconds and "
	    << tdiff.tv_usec << " microseconds." << endl;
	gettimeofday (&start, 0);
    }

    mpz_mod (tmp2, tmp1, n);
    if (debug) print_mpz ("C", tmp2);
    env->SetObjectArrayElement (results, 2, 
	mpz2bigint (env, tmp2, big_int_class, big_int_init));

    if (PROFILE) {
	gettimeofday (&finish, 0);
	timersub (&finish, &start, &tdiff);
	cout << "9. Took " << tdiff.tv_sec << " seconds and "
	    << tdiff.tv_usec << " microseconds." << endl;
	gettimeofday (&start, 0);
    }

    // z = (c.multiply(secret)).add(r);
    mpz_mul (tmp1, tmp2, secret);
    mpz_add (tmp2, tmp1, r);
    if (debug) print_mpz ("Z", tmp2);
    env->SetObjectArrayElement (results, 1,
	mpz2bigint (env, tmp2, big_int_class, big_int_init));

    if (debug) print_mpz ("XUPS", xi);
    env->SetObjectArrayElement (results, 0,
	mpz2bigint (env, xi, big_int_class, big_int_init));

    if (PROFILE) {
	gettimeofday (&finish, 0);
	timersub (&finish, &start, &tdiff);
	cout << "10. Took " << tdiff.tv_sec << " seconds and "
	    << tdiff.tv_usec << " microseconds." << endl;
    }

    mpz_clear (TWO);
    mpz_clear (FOUR);
    mpz_clear (x);
    mpz_clear (n);
    mpz_clear (r);
    mpz_clear (groupVerifier);
    mpz_clear (delta);
    mpz_clear (verifier);
    mpz_clear (signVal);
    mpz_clear (secret);
    mpz_clear (tmp1);
    mpz_clear (tmp2);
    mpz_clear (vprime);
    mpz_clear (xtilde);
    mpz_clear (xprime);
    mpz_clear (xi);
}

JNIEXPORT jobject JNICALL 
Java_ostore_security_thresh_NativeIF_package_1up_1c_1helper (
	JNIEnv *env, 
	jclass not_used, 
	jint l, 
	jobject jn, 
	jobject je, 
	jobjectArray jsigarray, 
	jintArray jlambdaarray, 
	jboolean debug) 
{
    jclass big_int_class = env->FindClass ("java/math/BigInteger");
    jmethodID big_int_init = 
	env->GetMethodID (big_int_class, "<init>", "([B)V");
    if (big_int_init == 0) 
	return 0;
    jmethodID big_int_to_byte_array = 
	env->GetMethodID (big_int_class, "toByteArray", "()[B");
    if (big_int_to_byte_array == 0) 
	return 0;
    
    mpz_t w, n, e;
    mpz_init_set_ui (w, 1);
    mpz_init (e);
    mpz_init (n);
    bigint2mpz (n, env, jn, big_int_to_byte_array);
    bigint2mpz (e, env, je, big_int_to_byte_array);

    jsize len = env->GetArrayLength (jsigarray);
    jint *jlambdas = env->GetIntArrayElements (jlambdaarray, 0);
    for (int i = 0; i < len; ++i) {
	mpz_t sig, lambda;
	mpz_init (sig);
	jobject jsig = env->GetObjectArrayElement (jsigarray, i);
	int jlambda = jlambdas [i];
	bigint2mpz (sig, env, jsig, big_int_to_byte_array);
	mpz_init_set_si (lambda, jlambda);
	if (debug) {
	    printf ("lambda [%d] = %d\n", i, jlambda);
	    print_mpz ("lambda", lambda);
	    print_mpz ("sig", sig);
	}
	mpz_powm (sig, sig, lambda, n);
	if (debug)
	    print_mpz ("smpl", sig);
	mpz_mul (w, w, sig);
	mpz_mod (w, w, n);
	if (debug)
	    print_mpz ("w", w);

	mpz_clear (sig);
	mpz_clear (lambda);
    }
    env->ReleaseIntArrayElements (jlambdaarray, jlambdas, 0);

    jobject result = mpz2bigint (env, w, big_int_class, big_int_init);

    mpz_clear (w);
    mpz_clear (e);
    mpz_clear (n);
    
    return result;
}

JNIEXPORT jboolean JNICALL 
Java_ostore_security_thresh_NativeIF_verify_1c_1helper (
	JNIEnv *env, 
	jclass not_used, 
	jbyteArray jdata, 
	jbyteArray jcombined_sig, 
	jint l, 
	jobject jn, 
	jobject je)
{
    jclass big_int_class = env->FindClass ("java/math/BigInteger");
    jmethodID big_int_init = 
	env->GetMethodID (big_int_class, "<init>", "([B)V");
    if (big_int_init == 0) 
	return 0;
    jmethodID big_int_to_byte_array = 
	env->GetMethodID (big_int_class, "toByteArray", "()[B");
    if (big_int_to_byte_array == 0) 
	return 0;
    
    mpz_t w, x, delta, n, e, eprime, we;
    mpz_init (w);
    mpz_init (x);
    mpz_init (delta);
    mpz_init (n);
    bigint2mpz (n, env, jn, big_int_to_byte_array);
    mpz_init (e);
    bigint2mpz (e, env, je, big_int_to_byte_array);
    mpz_init (eprime);
    mpz_init (we);

    jsize len = env->GetArrayLength (jdata);
    jbyte *data = env->GetByteArrayElements (jdata, 0);
    mag2mpz (x, (unsigned char*) data, len);
    env->ReleaseByteArrayElements (jdata, data, 0);

    len = env->GetArrayLength (jcombined_sig);
    data = env->GetByteArrayElements (jcombined_sig, 0);
    mag2mpz (w, (unsigned char*) data, len);
    env->ReleaseByteArrayElements (jcombined_sig, data, 0);

    mpz_mod (x, x, n);
    mpz_fac_ui (delta, l);
    mpz_mul (eprime, delta, delta);
    mpz_mul_2exp (eprime, eprime, 2);

    mpz_powm (eprime, x, eprime, n);
    mpz_powm (we, w, e, n);

    jboolean result = mpz_cmp (eprime, we) == 0;

    mpz_clear (w);
    mpz_clear (x);
    mpz_clear (delta);
    mpz_clear (n);
    mpz_clear (e);
    mpz_clear (eprime);
    mpz_clear (we);

    return result;
}
    
int main () {

    mpz_t pie;
    mpz_init_set_str (pie, "123456789deadbeef987654321", 16);
    mpz_out_str (0, 16, pie);

    unsigned int len;
    unsigned char *bytes;

    mpz2mag (pie, &bytes, &len);

    for (int i = 0; i < len; ++i) 
	printf ("%02x", bytes [i]);
    printf ("\n");

    mpz_t cake;
    mpz_init (cake);
    mag2mpz (cake, bytes, len);

    mpz_out_str (0, 16, cake);

    mpz_clear (pie);
    mpz_clear (cake);
}

