/**
 *  TapestryLocationFailure.java
 *
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.api;
import ostore.network.NetworkMessage;
import ostore.util.SecureHash;
import seda.sandStorm.api.QueueElementIF;
import ostore.util.TypeTable;
import ostore.util.QSException;
import ostore.util.ByteUtils;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;

/**
 * An event which indicates that a <code>TapestryLocateMsg</code> was not
 * successful.  In other words, that the {@link TapestryLocateMsg} made its
 * way all the way root without getting a positive query result.
 *
 * <p>NOTE: This event should be treated in much the same way as a DNS
 * lookup failure.  It does not necessary indicate that the query
 * <i>shouldn't</i> have succeeded, only that it didn't.  For instance, an
 * intermediate node between the query source and the root could well cause
 * such a failure, perhaps maliciously.  <i>Use with caution; you have been
 * warned.</i>
 *
 * @author     Ben Y. Zhao
 * @version    $Id: TapestryLocateFailure.java,v 1.16 2002/09/05 00:00:14 srhea Exp $
 */
public class TapestryLocateFailure
extends NetworkMessage implements QuickSerializable {

  /**
   * The hash of the original request. Used to distinquish this request.
   */
  public SecureHash id;
  /**
   * The guid where the request was determined to have failed. Should be the
   * dynamically mapped root of the object ID originally requested for.
   */
  public SecureHash root;

  public SecureHash obj_guid;
  
  /**
   * Number of hops taken by the message on its way to the application
   */
  public int hopCount;

  public TapestryLocateFailure (SecureHash root, SecureHash id,
				SecureHash obj_guid, NodeId req, int hop) {
    super(req, false /*outbound*/);
    this.id = id;
    this.root = root;
    this.obj_guid = obj_guid;
    this.hopCount = hop;
  }
  
  public TapestryLocateFailure (InputBuffer buffer) 
    throws QSException {
    super(buffer);
    root = (SecureHash) buffer.nextObject();
    id = (SecureHash) buffer.nextObject();
    obj_guid = (SecureHash) buffer.nextObject();
    hopCount = buffer.nextInt();
  }
  

  public void serialize(OutputBuffer buffer) {
    super.serialize(buffer);
    buffer.add(root);
    buffer.add(id);
    buffer.add(obj_guid);
    buffer.add(hopCount);
  }

  public String toString () {
      return "(TapestryLocateFailure id=" + id + ", root=" + root + 
	  ", obj_guid=" + obj_guid + ", super=" + super.toString () + ")";
  }
}

