/**
 * TapestryRouteMsg.
 *
 * @author     Sean C. Rhea
 * @version    $Id: TapestryRouteMsg.java,v 1.11 2002/08/18 22:39:41 srhea Exp $
 */
package ostore.tapestry.api;
import ostore.util.SecureHash;
import ostore.util.QuickSerializable;
import ostore.util.OutputBuffer;
import ostore.util.InputBuffer;
import seda.sandStorm.api.QueueElementIF;

/**
 * Messages sent from one OceanStore machine to any other.  To send a
 * message from one OceanStore machine to any other, one simply inherits
 * from this class, filling in the proper serialization code for the
 * inherited class's data.  (The Tapestry stage will take care of the
 * fields in this class itself.)  
 *
 * <p> Implementation-wise, this is simply a route-to-root operation.
 */
public abstract class TapestryRouteMsg 
implements QueueElementIF, QuickSerializable {

    /** 
     * If <code>inbound</code>, the sender; if <code>! inbound</code> the
     * intended recipient.
     */
    public SecureHash peer;

    /**
     * Whether this message is being received (<code>true</code>) or sent
     * (<code>false</code>).
     */
    public boolean inbound;

    /**
     * Number of hops taken by the message on its way to the application
     */
    public int hopCount;

    /**
     * Number of hops this message should take before delivering payload.
     */
    public int TTL;

    /**
     * The delivery qos of this msg.
     */

    public int qos;

    /**
     * Construct with the given destination.  Use for new outbound
     * messages.
     */
    protected TapestryRouteMsg (SecureHash peer) {
	this.peer = peer;
	this.inbound = false;
    }

    /**
     * Empty constructor.  For use by derived types when reconstructing
     * themselves from a byte array.
     */
    protected TapestryRouteMsg () {}

    protected TapestryRouteMsg (InputBuffer buffer) {}

    public void serialize (OutputBuffer buffer) {}
}

