/**
 * Channel class.
 *
 * @author     Ling Huang
 * @version    $Id: Channel.java,v 1.14 2003/03/03 18:54:14 hling Exp $
 */

package ostore.tapestry.channel;
import java.util.Map;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Set;
import java.util.Iterator;
import ostore.tapestry.impl.*;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkException;
import ostore.dispatch.Classifier;
import ostore.network.NetworkMessage;
import ostore.dispatch.Classifier;
import ostore.tapestry.patchwork.PatchworkModel;
import ostore.tapestry.patchwork.PatchworkChangePeriodMsg;
import ostore.tapestry.patchwork.PatchworkFaultMsg;
import ostore.tapestry.patchwork.RouterFaultMsg;
import ostore.util.AssertionViolatedException;
import ostore.util.ByteUtils;
import ostore.util.Carp;
import ostore.util.CountBuffer;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.QSArray;
import ostore.util.QSByteArray;
import ostore.util.QSException;
import ostore.util.QuickSerializable;
import ostore.util.SecureHash;
import ostore.util.TypeTable;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.tapestry.patchwork.PatchworkModel;
import ostore.tapestry.patchwork.PatchworkChangePeriodMsg;

public class Channel
{
    private String class_tag = "Channel";
    //    private static final boolean DEBUG = DebugFlags.TAPESTRY_ROUTER;
    private static final boolean DEBUG = DebugFlags.TAPESTRY_CHANNEL;

    //To which router this Channel belongs, for dispatch.
    private Router  _router;
    //Outstanding packets in this channel, waiting for ack and for retransmission.
    private HashMap _outstandings;

    //Now the channel is designed to be hop by hop. Every channel is associate
    //a node, and has nothing to do with messages.
    //A node has multiple channels. The number of channels equals to the number
    //of the node's neighbour.
    //A channel in a node is uniquely identified by its next hop in neighbour.
    protected NodeId     ID;

    protected int status;	//Status of this channel, including:
				//CHANNEL_CLOSED
				//CHANNEL_ESTABLISHED
				//CHANNEL_ESTABLISHING  //send out req, waiting for est.
				//CHANNEL_CLOSING       //send out req, waiting for close.
				//......
    public static final int CLOSED = 1000;
    public static final int ESTABLISHED  = 1001;
    public static final int ESTABLISHING = 1002;
    public static final int CLOSING = 1003;

    public static final int DEFAULT_RETRANS_CHECK_TIME = 1000;
    // default beaconing period (in milliseconds)
    public static final int DEFAULT_BEACON_PERIOD = 300;

//  Transmit timing stuff.
    protected long timer[];        //A set of timmrs in milliseconds for this channel, including:
				//timer[0]: connection-establishment timer
				//timer[1]: unit retransmission timer. If the message is divided
                                //into n ChannelMsgs, its retransmission timer should be
                                //n X timer[1].
				//timer[2]: reassembly timer
				//......

    protected long RTT;          //In unit of millisecond.
    protected long DTO;          //In unit of millisecond.
                                 //Drop packet time out. After DTO time if we don't get ack
                                 //for this packet, don't retransmit it any more and 
                                 //remove it from channel._outstandings.
                                 //This also should indicate that this channel is no 
                                 //longer alive.                             
    
//  Send sequence variables
    protected long sequence;   //starts at 0
//  protected int send_unack;
//  protected int send_next;
//  protected int send_urgent;

//  Receiver sequence variables. 
//  protected int recv_init;.
//  protected int recv_unack;
//  protected int recv_next;
//  protected int recv_urgent;

//Jemery: currently at least three kinds of information should provide
//to channel by patchwork. You can and should move them to patchwork field.
    protected long beacon_period;   //Beacon period in milliseconds.
    //For link qualities:
    protected double latency;      //what is its unit?
    protected double loss_rate;    //what is its unit?
                                   //This is the link quality we previously 
                                   //talk about, is it right?
                                   //What is its unit?
    protected double bandwidth;    //what is its unit?.

    protected int loss_beacon_period;   //Beacon period in milliseconds.
    protected int latency_beacon_period;   //Beacon period in milliseconds.

    //reassembly buffer, should contain a list of packets that 
    //need reassembly..
    protected NetworkMessage assm_buff[];    

//  Member functions:
//  Construction functions:
    public Channel()
    {
	ID   = null;

	status = CLOSED;
	RTT = 500;
	DTO = 15000;

	//	beacon_period = 10000;   //beacon probe every 10 second.
	//	link_quality = 1.0;

	sequence = 0;
	assm_buff = null;
	timer = null;
	_router = null;
	_outstandings = new HashMap();

	set_loss_beacon_period( DEFAULT_BEACON_PERIOD );
	set_latency_beacon_period( DEFAULT_BEACON_PERIOD );
    }

    public Channel(NodeId nexthop, Router router)
    {
	this.ID   = nexthop;

	this.status = CLOSED;
	RTT = 500;
	DTO = 10000;

	//	this.beacon_period = 10000;   //beacon probe every 10 second.
	//	this.link_quality = 1.0;

	sequence = 0;
	assm_buff = null;

    // Default vaules for Timers:
	timer = new long[3];
	timer[0] = 4000;       // Connection-establishment can be several RTT = 4000ms.
	timer[1] = 1000;       // Assume RTT is 500ms, unit retransmission timer 
	                       // should be 2 X RTT = 1000ms
	timer[2] = 1000;       // reassembly timer should be less or equal to 
	                       //retransmission timer.
	_router = router;
	_outstandings = new HashMap();

	set_loss_beacon_period( DEFAULT_BEACON_PERIOD );
	set_latency_beacon_period( DEFAULT_BEACON_PERIOD );
    }

    public Channel(NodeId nexthop, Router router, long timer[])
    {
	this.ID   = nexthop;

	this.status = CLOSED;
	RTT = 500;
	DTO = 10000;

	//	this.beacon_period = 10000;   //beacon probe every 10 second.
	//	this.link_quality = 1.0;

	sequence = 0;
	assm_buff = null;

	int i, num;
	num = timer.length;
	timer = new long[num];
	for (i = 0; i < num; i++)
	    this.timer[i] = timer[i];

	_router = router;
	_outstandings = new HashMap();

	set_loss_beacon_period( DEFAULT_BEACON_PERIOD );
	set_latency_beacon_period( DEFAULT_BEACON_PERIOD );
    }

    public void update_statistics()          
    {   
	//For successful delivery of every packet, or 
	//in some time frequency we should update the
	//statistics of the channel.

	final String tag = class_tag + ".update_statistics";
	long time = System.currentTimeMillis();

	//Integration of patchwork in channel:
	//Integration of patchwork in channel:
	//Update the link status of the channel.
	latency   = get_latency();
	loss_rate = get_loss_rate();

	if (latency > 0)
	    RTT = (long)java.lang.Math.ceil(2.0*latency);
	if (DEBUG) Debug.printtagln (tag, "Update link statistics at " + 
		   " at time " + time);
	if (DEBUG) Debug.printtagln (tag, "RTT = " + RTT +  
		   " loss_rate = " + loss_rate);
    }

    //Jemery:
    //Get information directly from patchwork table.
    //No calculation, because the table already have them stored.
    public double get_bandwidth()          {   return 10.0;   }

    /**
     * Get the period (in milliseconds) of the latency probing beacons
     */
    public int get_latency_beacon_period() { 
	return latency_beacon_period;
    }

    /**
     * Get the period (in milliseconds) of the loss probing beacons
     */
    public int get_loss_beacon_period() {   
	return loss_beacon_period;   
    }

    /**
     * Get the latency (in milliseconds) to the other end of the channel
     */
    public double get_latency() {   
	return getPWStat( PatchworkModel.LATENCY );
    }

    /**
     * Get the loss rate (in percentage of beacons lost) for this channel
     */
    public double get_loss_rate() {   
	return getPWStat( PatchworkModel.LOSS );
    }

    // private helper for getting statistics
    private double getPWStat( int type ) {
	if( _router == null ) {
	    return -1;
	} else {
	    //	    return _router.get_channel_handler().
	    //		_patchwork.lookupDouble( _router.get_routing_table().
	    //					 node_id_to_guid(ID), type );
	    return -1;
	}
    }

    //Jemery:
    //Explicit function call of probe and calculation for link quality. 
    //You may need to send out probe msgs to patchwork stage, calculate 
    //the link quality, and update those information in the patchwork 
    //table in ChannelHandler.
    //Without explicit function call, may be only do periodic calculation
    //for latency and loss_rate.
    //public double calculate_latency()               {   return 10.0;   }
    //public double calculate_loss_rate()             {   return 10.0;   }
    //public double calculate_bandwidth()             {   return 10.0;   }

    /**
     * Set the latency beaconing period
     *
     * @param period   new period in milliseconds
     */
    public void set_latency_beacon_period( int period ) {   
	latency_beacon_period = period;
	setPWPeriod( PatchworkChangePeriodMsg.LATENCY, period );
    }

    /**
     * Set the loss rate beaconing period
     *
     * @param period   new period in milliseconds
     */
    public void set_loss_beacon_period( int period ) {   
	loss_beacon_period = period;
	setPWPeriod( PatchworkChangePeriodMsg.LOSS, period );
    }

    // private helper function for setting periods
    private void setPWPeriod( int type, int period ) {
	dispatch( new PatchworkChangePeriodMsg( type, period, ID ) );
    }

    /**
     * Tell other end of channel to drop a certain percentage of both
     * Router and probe messages from this end.  Note the double taken
     * here is actually rounded off to an int, for now.
     *
     * @param quality   Percent of messages to keep, i.e. 100.0 = drop 
     *                  no messages, 0.0 = drop all messages
     */
    public void set_link_quality( double quality ) {   

	int loss_rate = (int) (100 - quality);
	dispatch( new PatchworkFaultMsg( ID, loss_rate, true ) );
	dispatch( new RouterFaultMsg( ID, loss_rate, true ) );

    }

//  Message fragmentation and reordering:
//  Functions to get member variables:
    public NodeId get_channel_id()        {   return ID; }
    public int  get_status()              {   return status; }
    public long get_sequence()            {   return sequence;  }
    public Router get_router()            {   return _router; }

// Get timer
    public long[]  get_timer()                  {    return timer;    }
    public long    get_timer_establishment()    {    return timer[0]; }
    public long    get_timer_retransmission()   {    return timer[1]; }
    public long    get_timer_reassembly()       {    return timer[2]; }
    //    public int   get_timer_beacon()           {    return timer[3]; }


//  Functions to set member variables:
    public void set_channel_id(NodeId id)          {   this.ID = id; }
    public void set_status(int status)             {   this.status = status; }
    public void set_sequence(int sequence)         {   this.sequence = sequence;  }
    public void set_router(Router router)          {   this._router = router; }

// Set timer
    public void set_timer(long timer[])
    {
	int  i, num;
	
	num = timer.length;
	timer = new long[num];
	for (i = 0; i < num; i ++)
	    this.timer[i] = timer[i];
    }
    public void set_timer_establishment(long est_timer)        {    timer[0] = est_timer; }
    public void set_timer_retransmission(long rtr_timer)       {    timer[1] = rtr_timer; }
    public void set_timer_reassembly(long asm_timer)           {    timer[2] = asm_timer; }
    //    public void set_timer_beacon(int beacon_timer)            {    timer[3] = beacon_timer; }


// Functions useful for others:
    public void init()
    {
    }

    public int ping()
    {
	return 0;
    }
    public int traceroute()
    {
	return 0;
    }

    protected void dispatch (QueueElementIF item) 
    {
	final String tag = class_tag + ".dispatch(QueueElementIF)";

	try 
	{
	    Classifier.getClassifier(_router.get_self_node_id()).dispatch(item);
	} 
	catch (SinkException e) 
	{
	    _router.BUG (class_tag + ".dispatch: could not dispatch " + item);
	}
    }

    public void send(NetworkMessage msg, SecureHash msg_id)
    {
	final String tag = class_tag + ".send(NetworkMsg)";
	long   rto, dto;

	//Send this message to next hop.
	dispatch(msg);

	//Record this message as outstanding.
	long time = System.currentTimeMillis();
	//	SecureHash msg_id = msg.get_msg_id();
	rto = time + 2*RTT;
	dto = time + DTO;
	OutstandingMsg ost_msg = new OutstandingMsg((NetworkMessage)msg, msg_id,
				 time, rto, dto);
	_outstandings.put(msg_id,ost_msg);

	if (DEBUG)
	{
	    Debug.printtagln (tag, " send out a msg to" +
		  msg.peer + " and put it in channel as outstanding " +
		  "with msg_id " + msg_id + "at time " + time);
	    Debug.printtagln (tag, " This packet will be " +
	           "retransmission timer out at " + rto + 
		   " and be dropping timer out at " + dto);

	    OutstandingMsg  retran_msg;
	    LinkedList  dispatched;

	    retran_msg = (OutstandingMsg)ost_msg.clone();
	    dispatched = _router.get_channel_handler()._dispatched;
	    dispatched.add(retran_msg);
	}

	return ;
    }

    public void recv(ChannelACKMsg msg)
    {
	SecureHash ack_id[];
	Object     rmvd_msg;
	int i,num;

	final String tag = class_tag + ".recv(ChannelACKMsg)";

	ack_id = msg.ack_id;
	num = ack_id.length;
	for (i = 0; i < num; i++)
	{
	    if (DEBUG) Debug.printtagln (tag, " receiver an ack from" +
		       msg.peer + " for msg_id " + ack_id[i] + ".");
	    rmvd_msg = _outstandings.remove(ack_id[i]);
	    if (rmvd_msg == null)
	    {
		if (DEBUG) Debug.printtagln (tag, " No outstanding msg" +
			   " for msg_id " + ack_id[i] + " exists!");
	    }
	    else
	    {
		if (DEBUG) Debug.printtagln (tag, " Remove outstanding msg" +
					     " with msg_id " + ack_id[i] + " .");
	    }
	}

	update_statistics();
	return ;
    }

    public int get_packet_number()
    {
	return _outstandings.size();
    }

    public void retransmit()
    {
	final String tag = class_tag + ".retransmit()";
	boolean retr_time_out = false;
	boolean drop_time_out = false;
       
	SecureHash     msg_id, drp_msg_id[];
	OutstandingMsg ost_msg;
	long time;
	int i, size, drp_msg_num;

	Set key_set = _outstandings.keySet();
	size = key_set.size();
	
	time = System.currentTimeMillis();
	if (DEBUG)
	{
	    Debug.printtagln (tag, "In channel " + ID + " " +
		   size + " packets are recorded as outstanding" + 
	           " at time " + time);
	}
	if (size > 0)
	{
	    drp_msg_id = new SecureHash[size];
	    drp_msg_num = 0;

	    Iterator iter = key_set.iterator ();
	    while ( iter.hasNext () ) 
	    {
		msg_id = (SecureHash)iter.next ();
		ost_msg = (OutstandingMsg)_outstandings.get(msg_id);
		if (ost_msg != null)
		{
		    if (ost_msg.RTO < time)
		    { 
			ost_msg.times++;
			ost_msg.RTO = time + 2*RTT;
			dispatch(ost_msg.msg);

			retr_time_out = true;
			if (DEBUG)
		        {
			    Debug.printtagln (tag, "Retransmission" + 
			    " timeout for packet " + ost_msg.msg_id +
			    " at time " + time);
			    
			    OutstandingMsg  retran_msg;
			    LinkedList  dispatched;

			    retran_msg = (OutstandingMsg)ost_msg.clone();
			    retran_msg.tran_time = time;

			    dispatched = _router.get_channel_handler()._dispatched;
			    dispatched.add(retran_msg);
			}
		    }

		    if (ost_msg.DTO < time)
		    {
			drp_msg_id[drp_msg_num] = msg_id;
			drp_msg_num++;
		    }
		}
	    }
	    
	    if (DEBUG) Debug.printtagln (tag, "Finish iteration " + 
		       "on all outstanding packets at time " + time);
	    if (retr_time_out == false)
	    {
		if (DEBUG) Debug.printtagln (tag, "No packet" + 
			   " retransmit at time " + time);
	    }

	    if (drp_msg_num > 0)
	    {
		for (i = 0; i < drp_msg_num; i++)
		{
		    _outstandings.remove(drp_msg_id[i]);
		    if (DEBUG) Debug.printtagln (tag, "Drop packet" + 
			       " timeout for packet " + drp_msg_id[i]  +
			       " at time " + time);
		}
	    }
	    else
	    {
		if (DEBUG) Debug.printtagln (tag, "No packet" + 
			   " dropped at time " + time);
	    }
	}
	else
	{
	    if (DEBUG) Debug.printtagln (tag, "No packet" + 
		       " for retransmission at time " + time);
	}
	
	return;
    }

//  Clone functions:
    public Object clone () throws CloneNotSupportedException 
    {
	//	Channel result = (Channel) super.clone ();
	Channel result = new Channel();
	
	result.ID = ID;

	result.status = status;
	result.sequence = sequence;

	//	result.beacon_period = beacon_period;
	//	result.link_quality = link_quality;

	//	result.assm_buff = assm_buff.clone();
	System.err.print("I don't know how to clone StringBuffer for assm_buff!\n");

	int i, num;
	num = timer.length;
	for (i = 0; i < num; i++)
	    result.timer[i] = timer[i];

	result._router = _router;
	return result;
    }

    public String toString () 
    {
	return "Channel " + super.toString () + " :" + ID;
    }
}















