/**
 * ChannelResponseMsg class.
 *
 * @author     Ling Huang
 * @version    $Id: ChannelResponseMsg.java,v 1.2 2002/07/20 19:18:15 srhea Exp $
 */

package ostore.tapestry.channel;
import ostore.tapestry.impl.*;
import ostore.network.NetworkMessage;
import ostore.util.AssertionViolatedException;
import ostore.util.ByteUtils;
import ostore.util.CountBuffer;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.QSArray;
import ostore.util.QSByteArray;
import ostore.util.QSException;
import ostore.util.QuickSerializable;
import ostore.util.SecureHash;
import ostore.util.TypeTable;


//Destination should send back ChannelResponseMsg to source for constructing a
//channel.
public class ChannelResponseMsg extends RouteMsg
{
    protected int channel_id;  // At this time, we make it local unique. 
    protected int status;        // Response error code for request

    // If status = CHANNEL_ID_OK, it is ok to use this id for two nodes to construct this channel.
    public static final int CHANNEL_ID_OK = 2000;
    // If status = CHANNEL_ID_EXIST, channel_id produced by sender exists at destination. 
    // Channel cannot be established in this situation. Destination recommends 
    // another id in channel_id.
    public static final int CHANNEL_ID_EXIST  = 2001;
    // If status = TOO_MUCH_CHANNEL, then destination is overloaded.
    public static final int CHANNEL_TOO_MUCH = 2002;
    public static final int CHANNEL_SUCCEED = 2003;
    // ..............

    public ChannelResponseMsg()
    {
	this.channel_id = 0;
	this.status = CHANNEL_ID_OK;

	this.prefix_only = false; //Channel never do prefix only routing.

	// Default values for hop and TTL: hop start from 0, and TTL is big enough!
	this.hop = 0;
	this.TTL = 100;
    }

    public ChannelResponseMsg (InputBuffer buffer) throws QSException 
    {
	super (buffer);
	channel_id = buffer.nextInt ();
	status = buffer.nextInt ();
    }

    public ChannelResponseMsg(NodeId peer, SecureHash src, SecureHash dest,int channel_id)
    {
	super(peer,src,dest);
	this.channel_id = channel_id;
	this.status = CHANNEL_ID_OK;

	this.prefix_only = false; //ChannelResponseMsg never do prefix only routing.

	// Default values for hop and TTL: hop start from 0, and TTL is big enough!
	this.hop = 0;
	this.TTL = 100;
    }

    public ChannelResponseMsg(NodeId peer, SecureHash src, SecureHash dest,int channel_id,
		      int hop, int ttl)
    {
	super(peer,src,dest);
	this.channel_id = channel_id;
	this.status = CHANNEL_ID_OK;

	this.prefix_only = false; //ChannelResponseMsg never do prefix only routing.

	this.hop = hop;
	this.TTL = ttl;
    }

    public ChannelResponseMsg(NodeId peer, SecureHash src, SecureHash dest,int channel_id,
			      int status, int hop, int ttl)
    {
	super(peer,src,dest);
	this.channel_id = channel_id;
	this.status = status;

	this.prefix_only = false; //ChannelResponseMsg never do prefix only routing.

	this.hop = hop;
	this.TTL = ttl;
    }

//  Functions to get member variables:
    public int get_channel_id()                   { return channel_id; }
    public int get_status()                       { return status; }
    public NodeId get_peer()                      { return peer; }
    public SecureHash get_src()                   { return src; }
    public SecureHash get_dest()                  { return dest; }
//  Functions to set member variables:
    public void set_channel_id(int id)            { channel_id = id; }
    public void set_status(int status)            { this.status  = status; }
    
//  Serialize functions:
    public void serialize (OutputBuffer buffer) 
    {
	super.serialize (buffer);

	buffer.add (channel_id);
	buffer.add (status);
    }

//  Clone functions:
    public Object clone () throws CloneNotSupportedException 
    {
	ChannelResponseMsg result = (ChannelResponseMsg) super.clone ();
	
	result.channel_id = channel_id;
	result.status = status;

	return result;
    }

//  Clone functions:
    public String toString () 
    {
	return "(ChannelResponseMsg " + super.toString () + " " + src + "->" + dest + 
	    " hop " + hop + " TTL " + TTL;
    }
}













