/**
 * DynamicMapMsg.java
 *
 * @author Ben Y. Zhao
 * @version $Id: DynamicMapMsg.java,v 1.21 2002/09/05 23:26:56 ravenben Exp $
 */

package ostore.tapestry.impl;
import ostore.network.NetworkMessage;
import ostore.util.NodeId;
import ostore.util.SecureHash;
import ostore.util.QSException;
import ostore.util.Carp;
import java.util.Iterator;
import java.util.ArrayList;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.QuickSerializable;

public class DynamicMapMsg
extends NetworkMessage implements QuickSerializable {

  public SecureHash src;  //the GUID of whoever sent this message
  public ArrayList route_level; //single level (hop) of the routing table
  public int sequence;    //which routing level is this
  public int total;       //how many total routing levels are there
  //the peer refers to the receiver, which in this case is the inserting node

  public DynamicMapMsg (NodeId peer, SecureHash src,
			ArrayList route_level, int seq, int tot) {
    super (peer, false /*outbound*/);
    this.src = src;
    this.route_level = route_level;
    this.sequence = seq;
    this.total = tot;
  }


  public DynamicMapMsg (InputBuffer buffer) throws QSException {
    super(buffer) ;
    src = (SecureHash) buffer.nextObject();
    SecureHash name;
    NodeId addr;
    int size = buffer.nextInt();
    route_level = new ArrayList(size);
    //initial capacity >= 16 / 0.75
    for (int i =0 ; i < size; i++) {
      name = (SecureHash) buffer.nextObject();
      addr = (NodeId) buffer.nextObject();
      if (!name.isNull()) 
	//if null guid, means pair of guid/addr is useless
	route_level.add(new GuidAddr(name, addr));
      else 
	route_level.add(new GuidAddr(null, addr));
    }
    sequence = buffer.nextInt();
    total = buffer.nextInt();
  }

  public void serialize (OutputBuffer buffer) {
    super.serialize (buffer);
    buffer.add(src);
    buffer.add(route_level.size());
    
    SecureHash name;
    NodeId addr;
    GuidAddr ga;
    for (Iterator iter = route_level.iterator(); iter.hasNext(); ) {
      ga = (GuidAddr) iter.next();
      name = ga.getGuid();
      if (name == null) 
	buffer.add(src.nullHash());
      else 
	buffer.add(name);
      addr = ga.getAddr();
      if (addr == null)
	Carp.die("Should never happen!!!!");
      buffer.add(addr);
    }

    buffer.add(sequence);
    buffer.add(total);
  }
  
  public Object clone () throws CloneNotSupportedException {
    DynamicMapMsg result = (DynamicMapMsg) super.clone();
    result.src = src;
    result.sequence = sequence;
    result.total = total;
    result.route_level.addAll(route_level);
    return result;
  }

  public String toString() {
    String s = "(DynamicMapMsg: map#"+sequence+" of "+total+"\n[";
    for (int i=0; i<route_level.size(); i++) {
      s += (GuidAddr) route_level.get(i) +", ";
    }
    return s;
  }

}
