/**
 *  FullTestDriver.java
 *
 *  @author     Ben Y. Zhao and Ling Huang
 *  @version    $Id: FullTestDriver.java,v 1.33 2003/05/22 23:45:26 hling Exp $
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.impl;
import java.util.*;
import java.io.*;
import java.net.*;
import java.security.*;
import ostore.apps.gnutella_trace.GnutellaMinuteEnterPDF;
import ostore.apps.gnutella_trace.GnutellaMinuteExitPDF;
import ostore.util.*;
import ostore.dispatch.*;
import ostore.tapestry.*;
import ostore.tapestry.api.*;
import ostore.tapestry.patchwork.CDF;
import ostore.network.patchwork.InjectFaultMsg;
import ostore.tapestry.patchwork.RouterFaultMsg;
import ostore.tapestry.channel.ICMPPongMsg;
import seda.sandStorm.api.*;
import seda.sandStorm.core.*;
import seda.sandStorm.internal.SinkProxy;

public class FullTestDriver implements EventHandlerIF, ProfilableIF {

    public String class_tag = "FullTestDriver";
    //    private static final boolean DEBUG = false;
    private static final boolean DEBUG = true;

    private void dispatch (QueueElementIF item) {
	try {
	    _classifier.dispatch(item);
	} catch (SinkException e) {
	    Carp.out ("Could not dispatch " + item);
	    System.exit (1);
	}
    }

    private void dispatchList (LinkedList list) {
      final String tag = class_tag+".dispatchList";
      ListIterator li = list.listIterator(0);

      //Modified by Ling Huang: if publishing is ongoing, all other actions
      //should wait for sometime (until publishing is finish)
      /*
      while (li.hasNext()) {
	dispatch ((QueueElementIF) li.next());
      }
      */

      long milis_to_wait = 1;
      if ( on_publishing )
      {
	  milis_to_wait = 60000;
	  //          milis_to_wait = 20000;
	  if (DEBUG)
	      Debug.printtagln(tag, "There is publishing going on, waiting for " +
			       milis_to_wait + " milliseconds...");
	  on_publishing = false;
      }

      while (li.hasNext()) {
	_classifier.dispatch_later((QueueElementIF) li.next(), milis_to_wait);
      }

    }

    private void dispatchList (LinkedList list, long time) 
    {
	final String tag = class_tag+".dispatchList";
	
	ListIterator li = list.listIterator(0);
	long milis_to_wait = time;
	while ( li.hasNext() ) {
	    _classifier.dispatch_later((QueueElementIF) li.next(), 
				       milis_to_wait);
	}
    }

    public void handleEvents(QueueElementIF items[]) 
    throws EventHandlerException {
	for(int i=0; i<items.length; i++)
	    handleEvent(items[i]);
    }

  public void printtagln(String tag, String msg) {
    System.err.print(tag+":"+msg+"\n");
  }

    public void init (ConfigDataIF config) throws Exception {

	on_publishing = false;
	responseCount = 0;

	Types.registerAllTypes ();
	rand = new Random();

	_classifier = Classifier.getClassifier (
                new NodeId (config.getString ("node_id")));
	
	_self_node_id = new NodeId (config.getString ("node_id"));

	class_tag = class_tag + ".["+_self_node_id.port()+"]";
	if (DEBUG) ostore.util.Debug.turnOn(class_tag);

	sim_running = config.getBoolean( "simulator_running" );

	final String tag = class_tag +".init";
	String on = config.getString("power");
	if (on != null && on.equalsIgnoreCase("off")) return;

	String stageName = config.getStage().getName();
	try {
	  if(config.getBoolean("Profile"))  {
	    config.getManager().getProfiler().add("ostore.tapestry.impl.FullTestDriver", this);
	    SinkIF sink =
	      config.getManager().getStage(stageName).getSink();
	    this_sink = new SinkProxy(sink,
				      config.getManager(),
				      config.getStage().getWrapper());
	  } else
	    this_sink = config.getManager().getStage(stageName).getSink();
	} catch (NoSuchStageException e) {
	  Carp.die (tag + ": could not get own sink.");
	}

	Filter filter;
	String [] event_types = {
	  "seda.sandStorm.api.StagesInitializedSignal",
	  "ostore.tapestry.impl.TapestryReadyMsg",
	  "ostore.tapestry.impl.TimeOutEvent",
	  "ostore.tapestry.api.TapestryDetachConfirm"
	};	

	for (int i=0; i< event_types.length; i++) {
	  filter = new Filter ();
	  if (! filter.requireType (Class.forName(event_types[i]))) {
	    Carp.out ("Could not require type " + event_types[i]);
	    System.exit (1);
	  }
	  _classifier.subscribe (filter, this_sink);
	}

	String [] msg_types = {
	  "ostore.tapestry.channel.ICMPPongMsg",
	  "ostore.tapestry.impl.TestReadyMsg",
	  "ostore.tapestry.impl.TestLocateMsg",
	  "ostore.tapestry.impl.TestFoundMsg",
	  "ostore.tapestry.api.TapestryLocateFailure",
	  "ostore.tapestry.impl.TestDeleteMsg",
	  "ostore.tapestry.impl.FullTestMemberMsg",
	  "ostore.tapestry.impl.FullTestRespMsg",
	  "ostore.tapestry.impl.DynamicMapMsg",
	  "ostore.tapestry.impl.DynamicObjMapMsg",
	  "ostore.tapestry.impl.DebugRespMsg"
	};

	for (int i=0; i< msg_types.length; i++) {
	  filter = new Filter ();
	  if (! filter.requireType (Class.forName(msg_types[i])))
	    Carp.die ("Could not require type " + msg_types[i]);
	  if (! filter.requireValue("inbound", new Boolean(true)))
	    Carp.die ("Could not require inbound = true for "+
		      msg_types[i]);
	  _classifier.subscribe (filter, this_sink);
	}

	String frequency = config.getString("probing_frequency");
	if (frequency != null)
	    probing_frequency = Integer.parseInt(frequency);
	Debug.printtagln(tag, "probing_frequency = " + probing_frequency);
	nodes = new NodeId [Integer.parseInt(config.getString("nodes"))];
	nodemap = new HashMap();
	unattached_nodes = new ArrayList();
	attached_nodes = new ArrayList();
	fixedNodes = new HashSet();
	waitingTimeouts = new HashMap();
	init_nodes = 0;
	tests = new LinkedList();
	//	tests.add("multicast_bandwidth");
	//	tests.add("misrouting_latency");
	//	tests.add("probing_bandwidth");
	//	tests.add("reachability");
	//	tests.add("route_switch");
	//	tests.add("lookup_results");
	//	tests.add("static_lookup");
	//	tests.add("rdp_results");
	//	tests.add("static_route");
	//	tests.add("multicast_route");
	//	tests.add("single_insertion");
	tests.add("parallel_insertion");
	//	tests.add("trace_driven");
	//	tests.add("single_insertion_bw");
	//	tests.add("single_insertion_bw_repeat");
	//	tests.add("routing_table_debug");
	//	tests.add("back_pointer_debug");
	//	tests.add("reverse_pointer_debug");
	//	tests.add("network_change");
	//      tests.add("verify_tables");
	//	tests.add("network_operation");
	//	tests.add("dynamic_route");
	//	tests.add("dynamic_lookup");
	System.out.println("FullTestDriver Tapestry stage initialized!!");
    }

  public void handleEvent(QueueElementIF item) 
    throws EventHandlerException {

    final String tag = class_tag+".handleEvent";

    if (item instanceof StagesInitializedSignal) {
      
	//do nothing;
    } else if (item instanceof TimeOutEvent) {
      TimeOutEvent toe = (TimeOutEvent) item;
      handle_timeoutevent(toe);
    } else if (item instanceof TapestryReadyMsg) {

    } else if (item instanceof FullTestMemberMsg) {
      FullTestMemberMsg  msg = (FullTestMemberMsg) item;

      if (DEBUG)
	Debug.printtagln(tag, msg + " : ready with number " + init_nodes + 
			 " of " + nodes.length + ".");
      if (msg.type == msg.STATIC) 
      {
	attached_nodes.add(msg.peer);
	fixedNodes.add(msg.peer);
      } 
      else if (msg.type == msg.DYNAMIC) 
	unattached_nodes.add(msg.peer);

      init_nodes++;
      if (msg.guid != null)
	  nodemap.put(msg.peer, msg.guid);

      if (init_nodes >= nodes.length) {
	//initialize the guids array
	guids=new SecureHash[attached_nodes.size()+unattached_nodes.size()];
	int i=0;
	for (; i<attached_nodes.size(); i++) {
	  guids[i] = (SecureHash) nodemap.get(attached_nodes.get(i));
	}
	for (int j=0; j<unattached_nodes.size(); j++) {
	  guids[i+j] = (SecureHash) nodemap.get(unattached_nodes.get(j));
	}
	//now we can start
	if (DEBUG)
	    Debug.printtagln(tag, msg + "All wanted " + init_nodes + 
			     "nodes come in. Begin to test...");
	next_test();
      }

    } else if (item instanceof FullTestRespMsg) {
      FullTestRespMsg msg = (FullTestRespMsg) item;
      handle_respmsg(msg);

    } else if (item instanceof DynamicMapMsg) {
      DynamicMapMsg msg = (DynamicMapMsg) item;
      handle_mapmsg(msg);
      if (waitingResponses.size() == 0) {
	  System.out.println("Routing Table Debug Test Passed.");
	  next_test();
      }

    } else if (item instanceof DynamicObjMapMsg) {
      DynamicObjMapMsg msg = (DynamicObjMapMsg) item;			       
      handle_objmapmsg(msg);
      if (responseCount == attached_nodes.size()) {
	  responseCount = 0;
	  System.out.println("Back Pointer Debug Test Passed.");
	  next_test();
      }

    } else if (item instanceof DebugRespMsg) {
      DebugRespMsg msg = (DebugRespMsg) item;
      handle_debugrespmsg(msg);
      if (currTest == ROUTING_TABLE_DEBUG) {
          if (waitingResponses.size() == 0) {
              next_test();
          }
      } else if (currTest == BACK_POINTER_DEBUG) {
	  if (responseCount == attached_nodes.size()) {
	      responseCount = 0;
	      System.out.println("Back Pointer Debug Test Passed.");
	      next_test();
	  }
      } else if (currTest == REVERSE_POINTER_DEBUG) {
          if (responseCount == attached_nodes.size()) {
	      responseCount = 0;
	      System.out.println("Reverse Pointer Debug Test Passed.");
              next_test();
          }
      }

    } else {
      Carp.die ("ostore.tapestry.impl.Test: Received unknown "
		+ "QueueElementIF: " + item);
    }
  }


  public void handle_timeoutevent(TimeOutEvent toe) {
      System.out.println("Timeout in FullTestDriver, some bad things happen...");
      if (toe.requestType == FullTestMsg.INTEGRATE_REQUEST) 
      {
	  if ( !timeout_already )
	  {
	      timeout_already = true;
	      begin_waiting_msgs_debug_test();
	  }

	  return;
      }
  }

  public void handle_mapmsg(DynamicMapMsg msg) {
      final String tag = "FullTestDriver.handle_mapmsg";
      
      if (currTest == ROUTING_TABLE_DEBUG) {
	  ArrayList al = msg.route_level;
	  GuidAddr ga;

	  // Display the entries
	  SecureHash guid;
	  for (int i=0; i<al.size(); i++) {
	      ga = (GuidAddr) al.get(i);
	      guid = ga.getGuid();
	      if (guid != null)
		  System.out.println("Received Routing Table Entry from " + msg.peer + " -- msg.sequence: " 
				     + msg.sequence + " ga.getAddr(): " + ga.getAddr() + " ga.getGuid(): " + 
				     ga.getGuid());
	      /* Do not add it to our routing table
		 _routing_tbl.write(msg.sequence, i, ga.getAddr(),
		 ga.getGuid(), new Long(1000));
	      */
	      
	  }
	  
	  waitingResponses.remove(msg.peer);
      }
  }

  public void handle_objmapmsg(DynamicObjMapMsg msg) {
      final String tag = "FullTestDriver.handle_objmapmsg";

      if (currTest == BACK_POINTER_DEBUG) {
	  // add one to the count if we're done
	  if (msg.objcount == msg.seq) {
	      responseCount++;
	  }
	  if (msg.objcount > 0) {  //in other words, not just a null msg
	      System.out.println("Received Back Pointer from " + msg.peer + " -- msg.objguid: " 
				 + msg.objguid + " msg.pinfo: " + msg.pinfo);
	      /* Do not add it to our backpointers list
		LinkedList sss;
		if (_backpointers.containsKey(msg.objguid)) {
		sss = (LinkedList) _backpointers.get(msg.objguid);
		sss.addAll(msg.pinfo);
		} else {
		_backpointers.put(msg.objguid, msg.pinfo);
		HashSet level =
		(HashSet) _objlevels.get(new Integer(Router.LOCALROOTOBJS));
		if (level == null) {
		level = new HashSet();
		_objlevels.put(new Integer(Router.LOCALROOTOBJS), level);
		}
		level.add(msg.objguid);
		}
		
		//got object from msg, now reply and confirm
		DynamicObjReqMsg dorm =
		new Dyn(0.05)amicObjReqMsg(msg.peer, _self_guid, _self_guid,
		DynamicObjReqMsg.CONFIRM);
		dispatch(dorm);
	      */
	  }
      }
  }

  public void handle_debugrespmsg(DebugRespMsg msg) {
      final String tag = "FullTestDriver.handle_debugrespmsg";

      if (currTest == ROUTING_TABLE_DEBUG && msg.requestType == DebugMsg.DONE) {
	  waitingResponses.remove(msg.peer);
      } else if (currTest == BACK_POINTER_DEBUG && msg.requestType == DebugMsg.DONE) {
	  responseCount++;
      } else if (currTest == REVERSE_POINTER_DEBUG) {
	  responseCount++;
	  System.out.println("Received Reverse Pointer from " + msg.peer + " -- Guid: " + msg.rev_s +
			     " NodeID: " + msg.rev_nid);
      }
      else
      {
	  waitingResponses.remove(msg.peer);
	  int size = waitingResponses.size();
	  System.out.println("Get DebugRespMsg(remaining = " + size+ "): " + msg);	  
      }
  }

  public void handle_respmsg(FullTestRespMsg msg) {
    final String tag = "FullTestDriver.handle_respmsg";
    if (currTest == DYNAMIC_ROUTE) {
      if (msg.requestType == FullTestMsg.DETACH_REQUEST) {
	if (attached_nodes.remove(msg.peer))
	  unattached_nodes.add(msg.peer);
	else if (DEBUG)
	  Debug.printtagln(tag, "Unexpected detach node received");
	if (waitingMembership.remove(msg.getID())==null && DEBUG)
	  Debug.printtagln(tag, "Unexpected response message: "+msg);
	else if (waitingMembership.size() == 0) {
	  //start the dynamic join requests
	  ListIterator li = unattached_nodes.listIterator();
	  LinkedList join = new LinkedList();
	  FullTestMsg ftm;
	  SecureHash id;
	  NodeId nid;
	  while (li.hasNext()) {
	    nid = (NodeId) li.next();
	    id = genID();
	    ftm = new FullTestMsg(nid, FullTestMsg.INTEGRATE_REQUEST,
				  id, 0, null, (NodeId)attached_nodes.get(0));
	    join.add(ftm);
	    waitingJoins.add(nid);
	  }
	  dispatchList(join);
	}
      } else if (msg.requestType == FullTestMsg.INTEGRATE_REQUEST) {
	if (unattached_nodes.remove(msg.peer))
	  attached_nodes.add(msg.peer);
	else if (DEBUG)
	  Debug.printtagln(tag, "Unexpected integrate node received");
	if (waitingJoins.remove(msg.peer) == false && DEBUG)
	  Debug.printtagln(tag, "Unexpected response message: "+msg);
	if (waitingJoins.size() == 0) {
	  //all joins finished, now start routes
	  FullTestMsg ftm;
	  SecureHash id;
	  LinkedList li = new LinkedList();
	  for (int i=0; i<attached_nodes.size(); i++) {
	    id = genID();
	    ftm = new FullTestMsg((NodeId) attached_nodes.get(i),
				  FullTestMsg.ROUTE_REQUEST, id, 0, guids);
	    ftm.test = DYNAMIC_ROUTE;
	    li.add(ftm);
	    waitingRoutes.add(attached_nodes.get(i));
	  }
	  dispatchList(li);
	}
      } else if (msg.requestType == FullTestMsg.ROUTE_REQUEST) {
	if (!msg.success()) {
	  failedNodes.add(msg.peer);
	  if (DEBUG)
	    Debug.printtagln(tag, "Failed route request from node: "+msg.peer);
	}
	if (waitingRoutes.remove(msg.getID()) == false && DEBUG)
	  Debug.printtagln(tag, "Unexpected route response message: "+msg);
	if (waitingRoutes.size() == 0) {
	  if (failedNodes.size() == 0)
	    System.err.println("Dynamic Route Test passed...");
	  else
	    System.err.println("Dynamic Route Test failed on nodes:"+
			       failedNodes);
	  next_test();
	}
      } else {
	if (DEBUG)
	  Debug.printtagln(tag, "Unexpected response message: "+msg);
      }

    } else if (currTest == DYNAMIC_PUBLISH) {
    } 
    else if (currTest == MISROUTING_LATENCY) 
    {
	if (msg.requestType == FullTestMsg.ROUTING_PATH_RESULTS)
	{
	    int i, j, k, branching_hops;
	    int num_selected, selected[];
	    String comment;
	    NodeId src_id, dst_ids[];
	    SecureHash id;
	    SecureHash [] dst_guids;
	    FullTestMsg ftm;
	    OverlayPath [] paths = msg.paths;

	    //Put all paths from different nodes in a big array.
	    resp_nodes++;
	    if (paths != null)
	    {
		for (i = 0; i < paths.length; i++)
		{
		    total_paths[num_paths] = paths[i];
		    num_paths++;
		    if (num_paths > total_paths.length)
			break;
		}
	    }
	    Debug.printtagln(tag, "resp_nodes = " + resp_nodes + 
			     "; fixedNodes.size() = " + fixedNodes.size());
	    //After we get responses from all nodes, do routing.
	    if ( resp_nodes == fixedNodes.size() )
	    {
		if (DEBUG)
		    Debug.printtagln(tag, "Get " + num_paths + 
				     " paths resp from all nodes, " +
				     "begin to routing...");
		if (DEBUG)
		{
		    Debug.printtagln(tag, "The set of good paths are as follows:");
		    for (i = 0; i < num_paths; i++)
		    {
			System.out.println(total_paths[i]);
		    }
		}

		num_selected = 100;
		//only use num_selected paths, if there are more than num_selected.
		if (num_paths <= num_selected)
		{
		    num_selected = num_paths;
		    selected = new int[num_selected];
		    for (i = 0; i < num_selected; i++)
			selected[i] = i;
		}
		else
		{
		    selected = new int[num_selected];
		    for (i = 0; i < num_selected; i++)
		    {
			selected[i] = (int)(java.lang.Math.random()*num_paths);
			Debug.printtagln(tag, "selected path " + i + " = " +
					 selected[i]);

		    }

		}
		long waiting_time = 1000;
		for (k = 0; k < num_selected; k++)
		{
		    comment = new String();
		    comment = "Path " + k + " with branch at 0: ";
		    i = selected[k];
		    src_id = (NodeId)total_paths[i].nodeids.get(0);
		    dst_ids = new NodeId[1];
		    dst_ids[0] = (NodeId)total_paths[i].nodeids.get(
					 total_paths[i].nodeids.size()-1);
		    dst_guids = new SecureHash[1];
		    dst_guids[0] = total_paths[i].dst;

		    id = genID();
		    ftm = new FullTestMsg(src_id, FullTestMsg.ROUTE_REQUEST, id, 
					  0, dst_guids, dst_ids, comment);
		    ftm.test = currTest;
		    _classifier.dispatch_later(ftm, waiting_time);
		    waiting_time = waiting_time + 10000;


		    branching_hops = total_paths[i].nodeids.size() - 1;
		    src_id = (NodeId)total_paths[i].nodeids.get(0);
		    dst_ids = new NodeId[1];
		    dst_ids[0] = (NodeId)total_paths[i].nodeids.get(
			                 total_paths[i].nodeids.size()-1);
		    dst_guids = new SecureHash[1];
		    dst_guids[0] = total_paths[i].dst;
		    for (j = 1; j < branching_hops; j++)
		    {
			 _classifier.dispatch_later(new RouterFaultMsg((NodeId)total_paths[i].nodeids.get(j), 
						    100, true,
						    (NodeId)total_paths[i].nodeids.get(j-1)),
						    waiting_time);
			 _classifier.dispatch_later(new InjectFaultMsg((NodeId)total_paths[i].nodeids.get(j), 
						    100, true,
 					            (NodeId)total_paths[i].nodeids.get(j-1)),
						    waiting_time);
			waiting_time = waiting_time + 8000;

			comment = new String();
			comment = "Path " +  k + " with branch at " + j + ": ";
			id = genID();
			ftm = new FullTestMsg(src_id, FullTestMsg.ROUTE_REQUEST, id, 
					      0, dst_guids, dst_ids, comment);
			ftm.test = currTest;
			_classifier.dispatch_later(ftm, waiting_time);
			waiting_time = waiting_time + 10000;

			_classifier.dispatch_later(new RouterFaultMsg((NodeId)total_paths[i].nodeids.get(j), 
								      0, true,
								      (NodeId)total_paths[i].nodeids.get(j-1)),
						   waiting_time);
			_classifier.dispatch_later(new InjectFaultMsg((NodeId)total_paths[i].nodeids.get(j), 
								      0, true,
								      (NodeId)total_paths[i].nodeids.get(j-1)),
						   waiting_time);
		    }
		}
	    }
	}
    }
    else if (currTest == MULTICAST_BANDWIDTH) 
    {
	if (msg.requestType == FullTestMsg.ROUTING_PATH_RESULTS)
	{
	    int i, j, k, branching_hops;
	    String comment;
	    NodeId src_id, dst_ids[];
	    SecureHash id;
	    SecureHash [] dst_guids;
	    FullTestMsg ftm;
	    OverlayPath [] paths = msg.paths;

	    //Put all paths from different nodes in a big array.
	    resp_nodes++;
	    if (paths != null)
	    {
		for (i = 0; i < paths.length; i++)
		{
		    total_paths[num_paths] = paths[i];
		    num_paths++;
		    if (num_paths > total_paths.length)
			break;
		}
	    }
	    if (DEBUG)
		Debug.printtagln(tag, "resp_nodes = " + resp_nodes + 
				 "; fixedNodes.size() = " + fixedNodes.size());
	    //After we get responses from all nodes, do routing.
	    if ( resp_nodes == fixedNodes.size() )
	    {
		Debug.printtagln(tag, "Get " + num_paths + 
				 " paths resp from all nodes, " +
				 "begin to routing...");
		if (DEBUG)
		{
		    Debug.printtagln(tag, "The set of good paths are as follows:");
		    for (i = 0; i < num_paths; i++)
			System.out.println(total_paths[i]);
		}

		long waiting_time = 1000;
		//only use the first 50 paths, if there are more than 10.
		//		if (num_paths > 50)
		//		    num_paths = 50;

		for (i = 0; i < num_paths; i++)
		{
		    Debug.printtagln(tag, "Experiment for path with node size = " +
				     total_paths[i].nodeids.size());
		    comment = new String(); 
		    comment = "Path " + i + " with branch at 0: ";
		    src_id = (NodeId)total_paths[i].nodeids.get(0);
		    dst_ids = new NodeId[1];
		    dst_ids[0] = (NodeId)total_paths[i].nodeids.get(
					 total_paths[i].nodeids.size()-1);
		    dst_guids = new SecureHash[1];
		    dst_guids[0] = total_paths[i].dst;

		    Debug.printtagln(tag, "Dispatch msg for " + comment + "...");
		    id = genID();
		    ftm = new FullTestMsg(src_id, FullTestMsg.ROUTE_REQUEST, id, 
					  0, dst_guids, dst_ids, comment);
		    ftm.test = currTest;
		    _classifier.dispatch_later(ftm, waiting_time);
		    waiting_time = waiting_time + 5000*2;

		    dispatch_clear_bandwidth_requests(comment, waiting_time);
		    waiting_time = waiting_time + 8000*2;
		}

		for (i = 0; i < num_paths; i++)
		{
		    branching_hops = total_paths[i].nodeids.size() - 1;
		    src_id = (NodeId)total_paths[i].nodeids.get(0);
		    dst_ids = new NodeId[1];
		    dst_ids[0] = (NodeId)total_paths[i].nodeids.get(
			                 total_paths[i].nodeids.size()-1);
		    dst_guids = new SecureHash[1];
		    dst_guids[0] = total_paths[i].dst;
		    for (j = 1; j < branching_hops; j++)
		    {
			/*
			 _classifier.dispatch_later(new RouterFaultMsg((NodeId)total_paths[i].nodeids.get(j), 
						    100, true,
						    (NodeId)total_paths[i].nodeids.get(j-1)),
						    waiting_time);
			*/
			 _classifier.dispatch_later(new InjectFaultMsg((NodeId)total_paths[i].nodeids.get(j), 
						    100, true,
 					            (NodeId)total_paths[i].nodeids.get(j-1)),
						    waiting_time);
			waiting_time = waiting_time + 10000*2;

			comment = new String();
			comment = "Path " +  i + " with branch at " + j + ": ";
			Debug.printtagln(tag, "Dispatch msg for " + comment + "...");
			id = genID();
			ftm = new FullTestMsg(src_id, FullTestMsg.ROUTE_REQUEST, id, 
					      0, dst_guids, dst_ids, comment);
			ftm.test = currTest;
			_classifier.dispatch_later(ftm, waiting_time);
			waiting_time = waiting_time + 5000*2;
			
			dispatch_clear_bandwidth_requests(comment, waiting_time);
			waiting_time = waiting_time + 8000*2;
			/*
			_classifier.dispatch_later(new RouterFaultMsg((NodeId)total_paths[i].nodeids.get(j), 
								      0, true,
								      (NodeId)total_paths[i].nodeids.get(j-1)),
						   waiting_time);
			*/
			_classifier.dispatch_later(new InjectFaultMsg((NodeId)total_paths[i].nodeids.get(j), 
								      0, true,
								      (NodeId)total_paths[i].nodeids.get(j-1)),
						   waiting_time);
		    }
		}
	    }
	}
    }
    else if (currTest == STATIC_LOOKUP)
    {
	if (msg.requestType == FullTestMsg.PUBLISH_REQUEST)
	{
	    if (waitingPublishes.remove(msg.peer)==false && DEBUG)
		Debug.printtagln(tag, "Unexpected response message: "+msg);
	    else if (waitingPublishes.size() == 0) 
	    {
		if (DEBUG)
		    Debug.printtagln(tag, "Success in publishing, begin to objects lookup...");

		//		dispatch_lookup_requests_by_group();
		dispatch_lookup_requests(fixedNodes, false);
	    }
	}	
	else if (msg.requestType == FullTestMsg.LOOKUP_REQUEST)
	{
	    boolean lookup_finish = handle_response_lookup_request(msg);
	    if ( lookup_finish )
	    {
		if (DEBUG)
		    Debug.printtagln(tag, "Succeed in lookup. Test finished!");
		next_test();
	    }
	}
    }
    else if (currTest == STATIC_ROUTE)
    {
	if (msg.requestType == FullTestMsg.ROUTE_REQUEST) 
	{
	    boolean routing_finish = handle_response_route_request(msg);
	    if ( routing_finish )
	    {
		if (DEBUG)
		    Debug.printtagln(tag, "Succeed in routing. Test finished!");
		next_test();
	    }
	}
	else if (msg.requestType == FullTestMsg.INTEGRATE_REQUEST) 
	{
	    inserted_nodes++;
	    Debug.printtagln(tag, "Inserted nodes number = " + inserted_nodes);
	    boolean join_finish = handle_response_integrate_request(msg);
	    if ( join_finish ) 
	    {
		long time =  Clock.current_date(_self_node_id).getTime() - insertion_start_time;
		Debug.printtagln(tag, "Succeed in parallel insertion within time " + time + 
				 " ms with insertion times = " + insertion_times);
		if (unattached_nodes.size() <= 0)
		{
		    if (DEBUG)
			Debug.printtagln(tag, "Succeed inserting " + 
					 inserted_nodes + 
					 " nodes into network. Begin to routing...");
		    dispatch_route_requests(currTest);
		}
		else
		    //		    dispatch_single_integrate_request();
		    dispatch_integrate_requests_with_number(20);
	    }
	}
    }
    else if (currTest == ROUTE_SWITCH)
    {
	//	if (msg.requestType == FullTestMsg.ROUTE_REQUEST) 
	{
	    //	    boolean routing_finish = handle_response_route_request(msg);
	    //	    if ( routing_finish )
	    {
		//		if (DEBUG)
		Debug.printtagln(tag, "Faults recovery time = " + msg.integrateTime);
		dispatch_clear_bandwidth_requests(true, 0);
		send_exit_request(10000);
	    }
	}
    }
    else if (currTest == MULTICAST_ROUTE)
    {
	if (msg.requestType == FullTestMsg.ROUTE_REQUEST) 
	{
	    boolean routing_finish = handle_response_route_request(msg);
	    if ( routing_finish )
	    {
		if (DEBUG)
		    Debug.printtagln(tag, "Succeed in routing. Test finished!");
		next_test();
	    }
	}
    }
    else if (currTest == TRACE_DRIVEN)
    {
	if (msg.requestType == FullTestMsg.PUBLISH_REQUEST)
	{
	    if (waitingPublishes.remove(msg.peer)==false && DEBUG)
		Debug.printtagln(tag, "Unexpected response message: "+msg);
	    else if (waitingPublishes.size() == 0) 
	    {
		if (DEBUG)
		    Debug.printtagln(tag, "Success in publishing, begin to objects lookup...");

		dispatch_lookup_requests(fixedNodes, false);
	    }
	}	
	else if (msg.requestType == FullTestMsg.LOOKUP_REQUEST)
	{
	    boolean lookup_finish = handle_response_lookup_request(msg);
	    if ( lookup_finish )
	    {
		if (DEBUG)
		    Debug.printtagln(tag, "Succeed in the baseline lookup. Setup the trace driven data...!");
		begin_trace_driven_test();
		try{Thread.sleep(30000);}catch(InterruptedException e){}
		dispatch_lookup_requests(fixedNodes, true);
	    }
	}
    }
    else if (currTest == PARALLEL_INSERTION)
    {
	if (msg.requestType == FullTestMsg.INTEGRATE_REQUEST)
	{
	    inserted_nodes++;
	    Debug.printtagln(tag, "Inserted nodes " + inserted_nodes +
			     " with IP " + msg.peer + " within " + 
			     msg.integrateTime+ "ms.");
	    if (max_insertion_time < msg.integrateTime)
		max_insertion_time = msg.integrateTime;

	    //For timeout...
	    Object token = waitingTimeouts.get(msg.requestID);
	    if (token != null)
		_classifier.cancel(token);
	    else System.err.println("FullTestMember.HandleTestFoundMsg: "+
				    "No token found for Locate timeout");

	    boolean join_finish = handle_response_integrate_request(msg);
	    if ( join_finish ) 
	    {
		long time =  Clock.current_date(_self_node_id).getTime() - insertion_start_time;
		Debug.printtagln(tag, "Succeed in parallel insertion within time " + time + 
				 " ms with insertion times = " + insertion_times);
		Debug.printtagln(tag, "Max insertion time in round " + insertion_times + 
				 " is " + max_insertion_time);

		max_insertion_time = 0;
		if (insertion_times > 20)
		{
		    Debug.printtagln(tag, "Succeed in nodes insertion with 20 times, test finish!");
		    next_test();
		}
		else
		{
		    //		    dispatch_detach_requests(fixedNodes);
		    dispatch_single_detach_requests(fixedNodes);
		}
	    }
	}
	else if (msg.requestType == FullTestMsg.DETACH_REQUEST) 
	{
	    detached_nodes++;
	    if (DEBUG)
		Debug.printtagln(tag, "Detached nodes number = " + detached_nodes);
	    boolean detach_finish = handle_response_detach_request(msg);
	    /*
	    if ( detach_finish ) 
	    {
		if (DEBUG)
		    Debug.printtagln(tag, "Succeed in nodes deletion, begin to insertion...");
		dispatch_integrate_requests();
	    }
	    */
	    if ( detach_finish )
	    {
		if ( attached_nodes.size() <= fixedNodes.size() )
		{
		    if (DEBUG)
			Debug.printtagln(tag, 
			"Success in nodes deletion, do nodes insertion again...");
		    //		  dispatch_lookup_requests();
		    dispatch_integrate_requests();
		}
		else
		    dispatch_single_detach_requests(fixedNodes);
	    }
	}
    }
    else if (currTest == SINGLE_INSERTION)
    {
	if (msg.requestType == FullTestMsg.INTEGRATE_REQUEST) 
	{
	    Debug.printtagln(tag, "Insertion times = " + insertion_times);
	    boolean join_finish = handle_response_integrate_request(msg);
	    if ( join_finish ) 
	    {
		if (insertion_times > 20)
		{
		    Debug.printtagln(tag, "Succeed in nodes insertion with 20 times, test finish!");
		    next_test();
		}
		else
		    dispatch_detach_requests(fixedNodes);
	    }
	}
	else if (msg.requestType == FullTestMsg.DETACH_REQUEST) 
	{
	    detached_nodes++;
	    if (DEBUG)
		Debug.printtagln(tag, "Detached nodes number = " + detached_nodes);
	    boolean detach_finish = handle_response_detach_request(msg);
	    if ( detach_finish ) 
	    {
		if (DEBUG)
		    Debug.printtagln(tag, "Succeed in nodes deletion, begin to insertion...");
		dispatch_integrate_requests();
	    }
	}
    }
    else if (currTest == SINGLE_INSERTION_BW_REPEAT)
    {
	if (msg.requestType == FullTestMsg.INTEGRATE_REQUEST) 
	{
	    inserted_nodes++;
	    Debug.printtagln(tag, "Inserted nodes number = " + inserted_nodes);
	    boolean join_finish = handle_response_integrate_request(msg);
	    if ( join_finish ) 
	    {
		dispatch_clear_bandwidth_requests(true, 0);
		try{Thread.sleep(10000);}catch(InterruptedException e){}
		if (inserted_nodes > 20)
		{
		    Debug.printtagln(tag, "Succeed in nodes insertion with 20 times, test finish!");
		    next_test();
		}
		else
		    dispatch_detach_requests(fixedNodes);
	    }
	}
	else if (msg.requestType == FullTestMsg.DETACH_REQUEST) 
	{
	    detached_nodes++;
	    if (DEBUG)
		Debug.printtagln(tag, "Detached nodes number = " + detached_nodes);
	    boolean detach_finish = handle_response_detach_request(msg);
	    if ( detach_finish ) 
	    {
		if (DEBUG)
		    Debug.printtagln(tag, "Succeed in nodes deletion, begin to insertion...");
		dispatch_clear_bandwidth_requests(false, 0);
		try{Thread.sleep(10000);}catch(InterruptedException e){}
		dispatch_integrate_requests();
	    }
	}
    }
    else if (currTest == SINGLE_INSERTION_BW)
    {
	if (msg.requestType == FullTestMsg.INTEGRATE_REQUEST) 
	{
	    inserted_nodes++;
	    boolean join_finish = handle_response_integrate_request(msg);
	    if ( join_finish ) 
	    {
		dispatch_clear_bandwidth_requests(true, 0);
		try{Thread.sleep(10000);}catch(InterruptedException e){}
		Debug.printtagln(tag, "Succeed in nodes insertion, print results...");
		next_test();
	    }
	}
    }
    else if (currTest == NETWORK_CHANGE) {
      if (msg.requestType == FullTestMsg.PUBLISH_REQUEST)
	{
	  if (waitingPublishes.remove(msg.peer)==false && DEBUG)
	    Debug.printtagln(tag, "Unexpected response message: "+msg);
	  else if (waitingPublishes.size() == 0) 
	    {
	      if (DEBUG)
		Debug.printtagln(tag, "Success in publishing,begin to nodes integration...");
	      dispatch_integrate_requests();
	    }
	}
      else if (msg.requestType == FullTestMsg.INTEGRATE_REQUEST) {
	inserted_nodes++;
	if (DEBUG)
	  Debug.printtagln(tag, "Inserted nodes number = " + inserted_nodes);
	boolean join_finish = handle_response_integrate_request(msg);
	if ( join_finish ) 
	  {
	    //all joins finished, now start obj location.
	    if (DEBUG)
	      Debug.printtagln(tag, "Success in nodes insertion, begin to obj location...");
	    dispatch_lookup_requests();
	  }
      }
      else if (msg.requestType == FullTestMsg.LOOKUP_REQUEST && lookup_times == 1) 
	{
	  boolean lookup_finish = handle_response_lookup_request(msg);
	  if ( lookup_finish )
	    {
	      //all joins finished, now start obj location.
	      if (DEBUG)
		Debug.printtagln(tag, "\nFinish obj location, begin to nodes deletion...");
	      dispatch_detach_requests();
	    }
	}
      else if (msg.requestType == FullTestMsg.DETACH_REQUEST) 
	{
	  boolean detach_finish = handle_response_detach_request(msg);
	  if ( detach_finish )
	    {
	      /*
		if (attached_nodes.remove(msg.peer))
		unattached_nodes.add(msg.peer);
		else if (DEBUG)
		Debug.printtagln(tag, "Unexpected detach node received");
		if (waitingDetaches.remove(msg.getID())==false && DEBUG)
		Debug.printtagln(tag, "Unexpected response message: "+msg);
		else if (waitingDetaches.size() == 0) 
		{
	    */
	      //Delete 90% of nodes from the network, do objects location again. 
	      if (DEBUG)
		Debug.printtagln(tag, 
				"Success in nodes deletion, do obj location again...");
	      dispatch_lookup_requests();
	    }
	}
	else if (msg.requestType == FullTestMsg.LOOKUP_REQUEST && lookup_times == 2) 
	{
	    boolean lookup_finish = handle_response_lookup_request(msg);
	    if ( lookup_finish )
		next_test();
	}
    }
    else if (currTest == DYNAMIC_LOOKUP) {
      if (msg.requestType == FullTestMsg.PUBLISH_REQUEST)
      {
	  if (waitingPublishes.remove(msg.peer)==false && DEBUG)
	    Debug.printtagln(tag, "Unexpected response message: "+msg);
	  else if (waitingPublishes.size() == 0) 
	  {
	      if (DEBUG)
		  Debug.printtagln(tag, "Success in publishing, begin to objects location...");
	      //	      dispatch_lookup_requests();
	      dispatch_lookup_requests(fixedNodes, false);

	  }
      }
      else if (msg.requestType == FullTestMsg.LOOKUP_REQUEST && lookup_times == 1) 
      {
	  boolean lookup_finish = handle_response_lookup_request(msg);
	  if ( lookup_finish )
	  {
	      if (DEBUG)
		  Debug.printtagln(tag, "\nFinish obj location, begin to nodes insertion...");
	      dispatch_integrate_requests();
	  }
      }
      else if (msg.requestType == FullTestMsg.INTEGRATE_REQUEST) {
	inserted_nodes++;
	if (DEBUG)
	  Debug.printtagln(tag, "Inserted nodes number = " + inserted_nodes);
	boolean join_finish = handle_response_integrate_request(msg);
	if ( join_finish ) 
	{
	    //all joins finished, now start obj location.
	    if (DEBUG)
	      Debug.printtagln(tag, "Success in " + insertion_times + 
			       "th nodes insertion, begin to obj location...");
	    //	    dispatch_lookup_requests();
	    dispatch_lookup_requests(fixedNodes, false);
	}
      }
      else if (msg.requestType == FullTestMsg.LOOKUP_REQUEST && lookup_times ==2) 
	{
	  boolean lookup_finish = handle_response_lookup_request(msg);
	  if ( lookup_finish )
	    {
	      //all joins finished, now start obj location.
	      if (DEBUG)
		Debug.printtagln(tag, "\nFinish obj location, begin to nodes deletion...");
	      dispatch_single_detach_requests(fixedNodes);
	    }
	}
      else if (msg.requestType == FullTestMsg.DETACH_REQUEST) 
      {
	  detached_nodes++;
	  if (DEBUG)
	      Debug.printtagln(tag, "Detached nodes number = " + detached_nodes);
	  boolean detach_finish = handle_response_detach_request(msg);
	  if ( detach_finish )
	  {
	      //Delete 90% of nodes from the network, do objects location again. 
	      if ( attached_nodes.size() <= fixedNodes.size() )
	      {
		  if (DEBUG)

		      Debug.printtagln(tag, 
				       "Success in nodes deletion, do obj location again...");
		  //		  dispatch_lookup_requests();
		  dispatch_lookup_requests(fixedNodes, false);
	      }
	      else
		  dispatch_single_detach_requests(fixedNodes);

	  }
      }
      else if (msg.requestType == FullTestMsg.LOOKUP_REQUEST && lookup_times == 3) 
      {
	  boolean lookup_finish = handle_response_lookup_request(msg);
	  if ( lookup_finish )
	  {
	      if (DEBUG)
		  Debug.printtagln(tag, "\nFinish obj location, begin to nodes insertion...");
	      lookup_times = 1;
	      failedNodes.clear();
	      dispatch_integrate_requests();
	  }
      }
    }
    else if (currTest == NETWORK_MAINTENANCE)
    {

    } 
    else if (currTest == VERIFY_TABLES)
    {

	if( msg.requestType == FullTestMsg.VERIFY_REQUEST ) {

	    VerifyRespMsg vm = (VerifyRespMsg) msg;

	    // add this data to our verifier.
	    verifier.addData( vm.getSrc(), vm.getRoutingTable(),
			      vm.getRevpointers() );
	    waiting_verifications--;

	    if( DEBUG) {
		Debug.printtagln( tag, 
				  "Got VerifyRespMsg from " + msg.getSrc() );
		Debug.printtagln( tag, vm.getRoutingTable().toString() );
	    }
	    // test is done
	    if( waiting_verifications == 0 ) {

		if( verifier.verify() ) {
		    System.err.println("Verify Tables Test passed...");
		} else {
		    System.err.println("Verify Tables Test failed...");
		    verifier.printErrors();
		}

		next_test();

	    }

	} else {
	    Debug.printtagln( tag, "Unexpected message response type: " +
			      msg );
	}

    }
    else if (currTest == NETWORK_OPERATION)
    {
	if (msg.requestType == FullTestMsg.PUBLISH_REQUEST)
	{
	    if (waitingPublishes.remove(msg.peer)==false && DEBUG)
		Debug.printtagln(tag, "Unexpected response message: "+msg);
	    else if (waitingPublishes.size() == 0) 
	    {
		if (DEBUG)
		    Debug.printtagln(tag, "Success in publishing,begin to nodes integration...");

		dispatch_integrate_requests();
	    }
	}
	else if (msg.requestType == FullTestMsg.ROUTE_REQUEST) 
	{
	    boolean routing_finish = handle_response_route_request(msg);
	    if ( routing_finish )
	    {
		if (DEBUG)
		    Debug.printtagln(tag, "Begin to route...");
		dispatch_route_requests(STATIC_ROUTE);
	    }
	}
	else if (msg.requestType == FullTestMsg.INTEGRATE_REQUEST)
	{
	    boolean join_finish = handle_response_integrate_request(msg);
	    if ( join_finish ) 
	    {
		//all joins finished, now start node deletion.
		if (DEBUG)
		    Debug.printtagln(tag, "Success in nodes insertion, begin to nodes deletion...");
		dispatch_detach_requests();
	    }
	}
	else if (msg.requestType == FullTestMsg.DETACH_REQUEST) 
	{
	    boolean detach_finish = handle_response_detach_request(msg);
	    if ( detach_finish )
	    {
		if (DEBUG)
		    Debug.printtagln(tag, "Success in nodes deletion, do nodes insertion ...");
		dispatch_integrate_requests();
	    }
	}
	/*
	else if (msg.requestType == FullTestMsg.INTEGRATE_REQUEST)
	{
	    boolean join_finish = handle_response_integrate_request(msg);
	    if ( join_finish ) 
	    {
		//all joins finished, now start obj location.
		if (DEBUG)
		    Debug.printtagln(tag, "Success in nodes insertion, begin to obj location...");
		dispatch_lookup_requests();
		dispatch_route_requests();
		dispatch_detach_requests();
	    }
	}
	else if (msg.requestType == FullTestMsg.LOOKUP_REQUEST ) 
	{
	    boolean lookup_finish = handle_response_lookup_request(msg);
	    if ( lookup_finish )
	    {
		//all joins finished, now start obj location.
		if (DEBUG)
		    Debug.printtagln(tag, "\nFinish obj location, begin to nodes deletion...");
		dispatch_lookup_requests();
	    }
	}
	else if (msg.requestType == FullTestMsg.DETACH_REQUEST) 
	{
	    boolean detach_finish = handle_response_detach_request(msg);
	    if ( detach_finish )
	    {
		//Delete 90% of nodes from the network, do objects location again. 
		if (DEBUG)
		    Debug.printtagln(tag, "Success in nodes deletion, do nodes insertion ...");
		dispatch_lookup_requests();
		dispatch_route_requests();
		dispatch_integrate_requests();
	    }
	}
	*/
    }

    else {
      if (msg.requestType == FullTestMsg.DETACH_REQUEST) {
	if (attached_nodes.remove(msg.peer))
	  unattached_nodes.add(msg.peer);
	else if (DEBUG)
	  Debug.printtagln(tag, "Unexpected detach node received");
	if (waitingRequests.remove(msg.getID())==null && DEBUG)
	  Debug.printtagln(tag, "Unexpected response message: "+msg);
	
      } else if (msg.requestType == FullTestMsg.INTEGRATE_REQUEST) {
	if (unattached_nodes.remove(msg.peer))
	  attached_nodes.add(msg.peer);
	else if (DEBUG)
	  Debug.printtagln(tag, "Unexpected integrate node received");
	if (waitingRequests.remove(msg.getID())==null && DEBUG)
	  Debug.printtagln(tag, "Unexpected response message: "+msg);
	
      } else if (msg.requestType == FullTestMsg.PUBLISH_REQUEST) {
	if (waitingRequests.remove(msg.getID())==null && DEBUG)
	  Debug.printtagln(tag, "Unexpected publish response message: "+msg);
	
      } else if (msg.requestType == FullTestMsg.UNPUBLISH_REQUEST) {
	if (waitingRequests.remove(msg.getID())==null && DEBUG)
	  Debug.printtagln(tag, "Unexpected unpublish response message: "+msg);
	
      } else if (msg.requestType == FullTestMsg.LOOKUP_REQUEST) {
	if (waitingRequests.remove(msg.getID())==null && DEBUG)
	  Debug.printtagln(tag, "Unexpected lookup response message: "+msg);
	
      } else if (msg.requestType == FullTestMsg.ROUTE_REQUEST) {
	if (waitingRequests.remove(msg.getID())==null && DEBUG)
	  Debug.printtagln(tag, "Unexpected route response message: "+msg);
	
      } else if (msg.requestType == FullTestMsg.DETACH_INTEGRATE_REQUEST) {
	if (waitingRequests.remove(msg.getID())==null && DEBUG)
	  Debug.printtagln(tag, "Unexpected detach-int response message:"+msg);
      }
    }
  }

  static final int DYNAMIC_ROUTE = 0;
  static final int DYNAMIC_PUBLISH = 1;
  static final int NETWORK_CHANGE  = 2;
  static final int STATIC_ROUTE  = 3;
  static final int NETWORK_MAINTENANCE = 4;
  static final int NETWORK_OPERATION = 5;
  static final int ROUTING_TABLE_DEBUG = 6;
  static final int BACK_POINTER_DEBUG = 7;
  static final int REVERSE_POINTER_DEBUG = 8;
  static final int SINGLE_INSERTION  = 9;
  static final int ROUTING_RDP_RESULTS = 10;
  static final int STATIC_LOOKUP  = 11;
  static final int LOOKUP_RDP_RESULTS = 13;
  static final int VERIFY_TABLES = 12;
  static final int SINGLE_INSERTION_BW  = 14;
  static final int PARALLEL_INSERTION  = 15;
  static final int DYNAMIC_LOOKUP = 16;
  static final int SINGLE_INSERTION_BW_REPEAT  = 17;
  static final int WAITING_MSGS_DEBUG  = 18;
  static final int TRACE_DRIVEN  = 19;
  static final int MULTICAST_ROUTE  = 20;
  static final int ROUTE_SWITCH  = 21;
  static final int PROBING_BANDWIDTH= 22;
  static final int REACHABILITY = 23;
  static final int MISROUTING_LATENCY = 24;
  static final int MULTICAST_BANDWIDTH = 25;

  public void next_test() {
    final String tag = "FullTestDriver.next_test";
    String test;

    if (tests.size() != 0) 
    {
      test = (String) tests.removeFirst();

      waitingRequests = new HashMap();
      waitingJoins = new HashSet();
      waitingRoutes = new HashSet();
      waitingMembership = new HashMap();
      waitingResponses = new HashSet();
      failedNodes = new HashSet();

      waitingPublishes   = new HashSet(); 
      waitingLookups     = new HashSet();
      waitingDetaches    = new HashSet();
      objects_in_network = new HashMap();

      on_publishing = false;
      insertion_times = 0;
      detach_times    = 0;
      route_times     = 0;
      lookup_times    = 0;
      publish_times   = 0;
      unpublish_times = 0;
      inserted_nodes  = 0;
      detached_nodes  = 0;
      max_insertion_time = 0;
      if (test.equalsIgnoreCase("dynamic_route")) 
      {
	begin_dynamic_route_test();
      } 
      else if (test.equalsIgnoreCase("dynamic_lookup")) 
      {
	begin_dynamic_lookup_test();
      } 
      else if (test.equalsIgnoreCase("dynamic_publish")) 
      {
	begin_dynamic_publish_test();
      }
      else if ( test.equalsIgnoreCase("verify_tables") )
      {

	begin_verify_tables_test();

      }
      else if ( test.equalsIgnoreCase("rdp_results") )
      {
	  begin_handle_routing_rdp_results();
      }
      else if ( test.equalsIgnoreCase("lookup_results") )
      {
	  begin_handle_lookup_rdp_results();
      }
      else if ( test.equalsIgnoreCase("static_route") )
      {
	  begin_static_route_test();
      }
      else if ( test.equalsIgnoreCase("multicast_route") )
      {
	  begin_multicast_route_test();
      }
      else if ( test.equalsIgnoreCase("route_switch") )
      {
	  begin_route_switch_test();
      }
      else if ( test.equalsIgnoreCase("probing_bandwidth") )
      {
	  begin_probing_bandwidth_test();
      }
      else if ( test.equalsIgnoreCase("reachability") )
      {
	  begin_reachability_test();
      }
      else if ( test.equalsIgnoreCase("misrouting_latency") )
      {
	  begin_misrouting_latency_test();
      }
      else if ( test.equalsIgnoreCase("multicast_bandwidth") )
      {
	  begin_multicast_bandwidth_test();
      }
      else if ( test.equalsIgnoreCase("static_lookup") )
      {
	  begin_static_lookup_test();
      }
      else if ( test.equalsIgnoreCase("parallel_insertion") )
      {
	  begin_parallel_insertion_test();
      }
      else if ( test.equalsIgnoreCase("trace_driven") )
      {
	  trace_driven_test();
      }
      else if ( test.equalsIgnoreCase("single_insertion") )
      {
	  begin_single_insertion_test();
      }
      else if ( test.equalsIgnoreCase("single_insertion_bw") )
      {
	  begin_single_insertion_bw_test();
      }
      else if ( test.equalsIgnoreCase("network_change") )
      {
	  begin_large_network_change();
      }
      else if ( test.equalsIgnoreCase("network_operation") )
      {
	  begin_network_operation_under_heavy_dynamic();
      }
      else if ( test.equalsIgnoreCase("routing_table_debug") )
      {
	  begin_routing_table_debug_test();
      }
      else if ( test.equalsIgnoreCase("back_pointer_debug") )
      {
	  begin_back_pointer_debug_test();
      }
      else if ( test.equalsIgnoreCase("reverse_pointer_debug") )
      {
	  begin_reverse_pointer_debug_test();
      }
    } 
    else 
    {
	//	send_exit_request(0);
	SecureHash id;
	NodeId  node_id;
	int     i, j;

	LinkedList list = new LinkedList();
	for (i=0; i<attached_nodes.size(); i++) 
	{
	    node_id = (NodeId)attached_nodes.get(i);
	    id = genID();
	    FullTestMsg ftm = new FullTestMsg(node_id,
					      FullTestMsg.EXIT_REQUEST,
					      id, 0, null);
	    list.add(ftm);
	}
	for (i=0; i<unattached_nodes.size(); i++) 
	{
	    node_id = (NodeId)unattached_nodes.get(i);
	    id = genID();
	    FullTestMsg ftm = new FullTestMsg(node_id,
					      FullTestMsg.EXIT_REQUEST,
					      id, 0, null);
	    list.add(ftm);
	}
	dispatchList(list);
	//	System.exit(1);
    }
  }


    void send_exit_request(long time)
    {
	SecureHash id;
	NodeId  node_id;
	int     i, j;

	LinkedList list = new LinkedList();
	for (i=0; i<attached_nodes.size(); i++) 
	{
	    node_id = (NodeId)attached_nodes.get(i);
	    id = genID();
	    FullTestMsg ftm = new FullTestMsg(node_id,
					      FullTestMsg.EXIT_REQUEST,
					      id, 0, null);
	    list.add(ftm);
	}
	for (i=0; i<unattached_nodes.size(); i++) 
	{
	    node_id = (NodeId)unattached_nodes.get(i);
	    id = genID();
	    FullTestMsg ftm = new FullTestMsg(node_id,
					      FullTestMsg.EXIT_REQUEST,
					      id, 0, null);
	    list.add(ftm);
	}

	dispatchList (list, time);
    }

  /**
   ** dynamic routing test
   ** Look at list of dynam vs static nodes, if number of static nodes 
   ** is more than 1/2 the nodes, then detach a few of them.  Then insert
   ** all of the dynamic nodes, and then request that every node route to
   ** every other
   */
  public void begin_dynamic_route_test() {
    currTest = DYNAMIC_ROUTE;
    //assert here that both waitingRoutes and waitingMembership are either
    //null or empty
    int adj;
    adj = attached_nodes.size() - unattached_nodes.size();
    if (unattached_nodes.size() == 0) 
      adj = attached_nodes.size() / 2;
    SecureHash id;
    LinkedList list = new LinkedList();
    if (adj > 1) {
      for (int i=0; i< adj; i++) {
	id = genID();
	FullTestMsg ftm = new FullTestMsg((NodeId) attached_nodes.get(i),
					  FullTestMsg.DETACH_REQUEST,
					  id, 0, null);
	waitingMembership.put(ftm.requestID, ftm);
	list.add(ftm);
      }
      dispatchList(list);
    } else {
      //start inserting nodes instead of waiting for detaches to come back
      ListIterator li = unattached_nodes.listIterator();
      LinkedList join = new LinkedList();
      FullTestMsg ftm;
      NodeId nid;
      while (li.hasNext()) {
	nid = (NodeId) li.next();
	id = genID();
	ftm = new FullTestMsg(nid, FullTestMsg.INTEGRATE_REQUEST,
			      id, 0, null,  (NodeId)attached_nodes.get(0));
	join.add(ftm);
	waitingJoins.add(nid);
      }
      dispatchList(join);
    }
  }

  public void begin_dynamic_lookup_test() 
  {
      //initial static network => publish => integration => routing => 
      //lookup => deletion => lookup;

      final String tag = class_tag +".dynamic_lookup";
      currTest = DYNAMIC_LOOKUP;

      if (DEBUG)
	  Debug.printtagln(tag, "Objects redistribution and " +
			   "location when large portion of the network " +
			   "has been changed.");

    int num_snodes, num_dnodes;
    num_snodes =   attached_nodes.size();
    num_dnodes = unattached_nodes.size();
    if (num_snodes < 2)
    {
	if (DEBUG)
	    Debug.printtagln(tag, "Too few static nodes : " + num_snodes);
	return;
    }    

    dispatch_publish_requests(50);

    return;
  }

  public void begin_dynamic_publish_test() {
    currTest = DYNAMIC_PUBLISH;
    
  }

  /**
   * Send out requests for everyone's routing info
   */ 
  public void begin_verify_tables_test() {

      currTest = VERIFY_TABLES;

      verifier = new TapestryVerifier();
      waiting_verifications = 0;

      // send only to attached nodes
      SecureHash msg_id = new SHA1Hash( "no_id" );
      for( int i = 0; i < attached_nodes.size(); i++ ) {
	  waiting_verifications++;
	  dispatch( new FullTestMsg( (NodeId) attached_nodes.get(i), 
				     FullTestMsg.VERIFY_REQUEST, msg_id,
				     0, null ) );
      }

  }

  public void begin_single_insertion_test() 
  {
      final String tag = class_tag +".single_insertion";
      if (DEBUG)
	  Debug.printtagln(tag, "Ling Huang single insertion test with " + 
			   attached_nodes.size() + " nodes ...");
      currTest = SINGLE_INSERTION;

      //      dispatch_route_requests();
      dispatch_integrate_requests();
  }

    public void trace_driven_test()
    {
	final String tag = class_tag +".trace_driven";
	if (DEBUG)
	    Debug.printtagln(tag, "Conitiuous lookup Test with " + 
			     attached_nodes.size() + " static nodes and " +
			     unattached_nodes.size() + " dynamic nodes...");
	currTest = TRACE_DRIVEN;

	dispatch_publish_requests(20);
    }

  public void begin_trace_driven_test() 
  {

      if( DEBUG ) {
	  System.out.println( "Starting trace driven test" );
      }

      //Trace parsing
      GnutellaMinuteExitPDF exit = new GnutellaMinuteExitPDF();
      GnutellaMinuteEnterPDF enter = new GnutellaMinuteEnterPDF();
      final int total_mins = 60 * //mins_per_hour
	  60; // total_hours


      long[][] node_times = new long[unattached_nodes.size()][];

      // for each node, have a different RNG and figure out the
      // come and go times for the next 60 hours
      for( int i = 0; i < unattached_nodes.size(); i++ ) {
	  NodeId currNode = (NodeId) unattached_nodes.get(i);

	  // TODO: input random seed, use that instead of 1000 here
	  Random rng = new Random( 1000 + currNode.hashCode() );
	  int curr_min = 0;
	  int curr_life = 0;
	  CDF curr_cdf = enter;
	  boolean alive = false;
	  Vector v_times = new Vector();
	  while( curr_min < total_mins ) {
	      double cdf_value = curr_cdf.getCDFEntry(curr_life);
	      if( rng.nextDouble() < cdf_value ) {
		  // we're switching modes
		  v_times.add( new Long( curr_min*2*60*1000 ) );
		  alive = !alive;
		  if( false /*DEBUG*/ ) {
		      System.out.println( "Node " + currNode + " now alive (" +
					 alive + ") at " +
					 curr_min + " minutes, life " + 
					 curr_life );
		  }
		  if( alive ) {
		      // we're now alive, so next we should exit
		      curr_cdf = exit;
		  } else {
		      curr_cdf = enter;
		  }
		  curr_life = 0;
	      } else {
		  curr_life++;
	      }
	      curr_min++;
	  }
	  
	  // now convert that into an array
	  node_times[i] = new long[v_times.size()];
	  for( int j = 0; j < v_times.size(); j++ ) {
	      node_times[i][j] = ((Long) v_times.get(j)).longValue();
	  }

      }
      
      SecureHash reqID = new SHA1Hash( "Trace driven" );
      //Send trace to FullTestMember via FullTestMsg
      for( int i = 0; i < unattached_nodes.size(); i++ ) {

	  FullTestMsg ftm = new FullTestMsg( (NodeId) unattached_nodes.get(i),
					     FullTestMsg.TRACE_DATA, reqID,
					     (NodeId) attached_nodes.get(0),
					     node_times[i] );
	  if( false /*DEBUG*/ ) {
	      System.out.println( "Sending times to node " + 
				  (NodeId) unattached_nodes.get(i) + 
				  ": " );
	      for( int j = 0; j < node_times[i].length; j++ ) {
		  System.out.print( node_times[i][j] + " " );
	      }
	      System.out.println( "" );
	  }

	  dispatch( ftm );

      }

      // delay by 30 seconds to give the other messages time to reach
      // their destinations
      try{Thread.sleep(20000);}catch(InterruptedException e){};
      // now that we've sent all that, send a trace start message to everyone
      for( int i = 0; i < unattached_nodes.size(); i++ ) {

	  FullTestMsg ftm = new FullTestMsg( (NodeId) unattached_nodes.get(i),
					     FullTestMsg.TRACE_START, reqID,
					     null, null );
	  dispatch( ftm );
	  //	  _classifier.dispatch_later( ftm, 10000 );
      }
  }

  public void begin_parallel_insertion_test() 
  {
      final String tag = class_tag +".parallel_insertion";
      if (DEBUG)
	  Debug.printtagln(tag, "parallel insertion test with " + 
			   attached_nodes.size() + " nodes ...");

      currTest = PARALLEL_INSERTION;
      //      dispatch_integrate_requests_with_number(25);
      timeout_already = false;
      dispatch_integrate_requests();
  }


  public void begin_single_insertion_bw_test() 
  {
      final String tag = class_tag +".single_insertion_bw";
      if (DEBUG)
	  Debug.printtagln(tag, "Single insertion test with " + 
			   attached_nodes.size() + " nodes ...");
      currTest = SINGLE_INSERTION_BW;

      //      dispatch_route_requests();
      dispatch_clear_bandwidth_requests(false, 0);
      try{Thread.sleep(10000);}catch(InterruptedException e){}
      dispatch_integrate_requests();
  }

  public void begin_static_route_test()
  {
      final String tag = class_tag +".static_route";
      if (DEBUG)
	  Debug.printtagln(tag, "Static route Test with " + 
			   attached_nodes.size() + " static nodes and " +
			   unattached_nodes.size() + " dynamic nodes...");

      currTest = STATIC_ROUTE;
      if (unattached_nodes.size() <= 2)
	  dispatch_route_requests(STATIC_ROUTE);
      else
	  //	  dispatch_single_integrate_request();
	  dispatch_integrate_requests_with_number(20);

  }

  public void begin_multicast_route_test()
  {
      final String tag = class_tag +".multicast_route";
      if (DEBUG)
	  Debug.printtagln(tag, "Multicast route Test with " + 
			   attached_nodes.size() + " static nodes and " +
			   unattached_nodes.size() + " dynamic nodes...");

      int size = attached_nodes.size();
      dispatch_route_requests(1, 0, STATIC_ROUTE);
      //      dispatch_route_requests();
  }

  public void begin_probing_bandwidth_test()
  {
      final String tag = class_tag +".probing_bandwidth";
      Debug.printtagln(tag, "Running exp. to measure the bandwidth of probing...");

      dispatch_clear_bandwidth_requests(false, 0);
      dispatch_clear_bandwidth_requests(true, 10000);

      //      try{Thread.sleep(20000);}catch(InterruptedException e){}
      Debug.printtagln(tag, "Succeed in probing, test finish!");
      send_exit_request(20000);
  }

  public void begin_reachability_test()
  {
      final String tag = class_tag +".reachability";
      if (DEBUG)
	  Debug.printtagln(tag, "Reachability Test with " + 
			   attached_nodes.size() + " static nodes and " +
			   unattached_nodes.size() + " dynamic nodes...");

      currTest = REACHABILITY;

      int i, ports[];
      int size = attached_nodes.size();
      ports = new int[size];
      for (i = 0; i < size; i++)
	  ports[i] = ((NodeId)attached_nodes.get(i)).port();


      if (DEBUG)
	  Debug.printtagln(tag, "Read Topology data from file...");
      _topo_parser = new Parser();
      if ( sim_running )
	  _topo_parser.translate_topology(ports, size);
      else
	  _topo_parser.input_links();
      if (DEBUG)
	  Debug.printtagln(tag, "Inject Fault to Tapestry Network...");

      int waiting_time = 0;
      int failed_paths, total_paths = 40000;
      failed_paths = 0;
      percentage_failed_links = 0.0;
      percentage_failed_paths = 0.0;;
      Debug.printtagln(tag, "percentage_failed_links = " +
		       percentage_failed_links +
		       "; percentage_failed_paths = " +
		       (1.0-percentage_failed_paths));
     
      all_failed_ip_pairs = new HashSet();
      all_failed_paths = new HashSet();

      //      dispatch_route_requests(0.0, 0, REACHABILITY);
      //      waiting_time = waiting_time + 20000;

      for (i = 1; i <= 20; i++)
      {
	  if (sim_running)
	      failed_paths = injectFaultToTapestry_sim( _topo_parser, 0.01, 
					    new Integer(i), waiting_time);
	  else
	  {
	      Debug.printtagln(tag, "Here are some error when running on planetlab!");
	      //	      failed_paths = injectFaultToTapestry(_topo_parser, 0.01, 
	      //			   new Integer(i), waiting_time);
	      return;
	  }
	  waiting_time = waiting_time + 20000;

	  dispatch_route_requests((double)i, waiting_time, REACHABILITY);
	  percentage_failed_links = percentage_failed_links + 0.01;
	  percentage_failed_paths = (double)failed_paths / total_paths;
	  Debug.printtagln(tag, "Total failed paths = " + failed_paths +
			   "percentage_failed_links = " +
			   percentage_failed_links +
			   "; percentage_failed_paths = " +
			   (1.0-percentage_failed_paths));

	  waiting_time = waiting_time + 20000;
      }
  }

  public int injectFaultToTapestry_sim(Parser parser, double percentage, 
				    Integer unique_id, long waiting_time)
  {
      Parser.Path path;
      HashSet broken_paths=null;
      HashSet ip_pairs_table;
      SecureHash id;
      NodeId src_id, dst_id;
      String  id_string;
      int num_ip, num_tap;
      int size, i, j;
      int print = 1;

      final String tag = class_tag +".reachability.InjectFault";
      int failed_links = (int)(percentage*parser.getTotalNumberofLinks());
      failed_links = 1;

      if (DEBUG)
	  Debug.printtagln(tag, "Add failed links = " + 
			   failed_links + " ...");
      try {
	  parser.addFailedLinks(failed_links, unique_id);
      }
      catch (Exception e)
      {
	  Carp.out ("Unique ID = " + unique_id + 
		    " already exists!");
	  return 0;
      }
      try {
	  broken_paths = parser.getFailedPath(unique_id);
      }
      catch (Exception e)
      {
	  Carp.out (" "+e);
	  e.printStackTrace();
	  return 0;
      }
      if (DEBUG)
	  Debug.printtagln(tag, "Get failed paths = " + 
			   broken_paths.size() + "in this round!");
      if (DEBUG)
	  Debug.printtagln(tag, "Build a hash table for failed IP paths...");

      int port;
      InetAddress ip_addr = ((NodeId)attached_nodes.get(0)).address();
      Iterator it = broken_paths.iterator();
      while ( it.hasNext() )
      {
	  path = (Parser.Path)it.next();
	  all_failed_paths.add(path);
	  if (print == 1)
	  {
	      Debug.printtagln(tag, "Failed_paths: path");
	      print = 0;
	  }

	  port = Integer.parseInt(path.from);
	  src_id = new NodeId(port, ip_addr);
	  port = Integer.parseInt(path.to);
	  dst_id = new NodeId(port, ip_addr);
	  _classifier.dispatch_later(new RouterFaultMsg(
				     dst_id, 100, true, src_id),
				     waiting_time);
	  _classifier.dispatch_later(new InjectFaultMsg(
				     dst_id, 100, true, src_id),
				     waiting_time);
      }

      return all_failed_paths.size();
  }  


  public int injectFaultToTapestry(Parser parser, double percentage, 
				    Integer unique_id, long waiting_time)
  {
      Parser.Path path;
      HashSet broken_paths=null;
      HashSet ip_pairs_table;
      SecureHash id;
      NodeId src_id, dst_id;
      String  id_string;
      int num_ip, num_tap;
      int size, i, j;

      final String tag = class_tag +".reachability.InjectFault";
      int failed_links = (int)(percentage*parser.getTotalNumberofLinks());

      if (DEBUG)
	  Debug.printtagln(tag, "Add failed links = " + 
			   failed_links + " ...");
      try {
	  parser.addFailedLinks(failed_links, unique_id);
      }
      catch (Exception e)
      {
	  Carp.out ("Unique ID = " + unique_id + 
		    " already exists!");
	  return 0;
      }
      try {
	  broken_paths = parser.getFailedPath(unique_id);
      }
      catch (Exception e)
      {
	  Carp.out (" "+e);
	  e.printStackTrace();
	  return 0;
      }
      if (DEBUG)
	  Debug.printtagln(tag, "Get failed paths = " + 
			   broken_paths.size() + " ...");
      if (DEBUG)
	  Debug.printtagln(tag, "Build a hash table for failed IP paths...");
      num_ip = 0;
      ip_pairs_table = new HashSet();
      Iterator it = broken_paths.iterator();
      while ( it.hasNext() )
      {
	  path = (Parser.Path)it.next();

	  id_string = path.from.toString() + path.to.toString();
	  id = new SHA1Hash(id_string);
	  ip_pairs_table.add(id);
	  if ( !all_failed_ip_pairs.contains(id) )
	  {
	      all_failed_ip_pairs.add(id);
	      num_ip++;
	  }
      }
      if (DEBUG)
	  Debug.printtagln(tag, "The total # of failed IP path = " + 
			   num_ip);

      if (DEBUG)
	  Debug.printtagln(tag, "Mapping from failed IP paths to Tapestry path ...");
      num_tap = 0;
      size = attached_nodes.size();
      for (i = 0; i < size; i++)
      {
	  src_id = (NodeId)attached_nodes.get(i);
	  for (j = 0; j < size; j++)
	  {
	      dst_id = (NodeId)attached_nodes.get(j);
	      if ( NodeIdPairInIPPairsTable(
		   src_id, dst_id, ip_pairs_table) )
	      {
		  num_tap++;
		  _classifier.dispatch_later(new RouterFaultMsg(
					     dst_id, 100, true, src_id),
					     waiting_time);
		  _classifier.dispatch_later(new InjectFaultMsg(
					     dst_id, 100, true, src_id),
					     waiting_time);
	      }
	  }
      }
      if (DEBUG)
	  Debug.printtagln(tag, "The total # of failed TAP path = " + 
			   num_tap);

      return num_ip;
  }  

  boolean NodeIdPairInIPPairsTable(NodeId src_id, NodeId dst_id, 
				   HashSet ip_pairs_table)
  {
      String id_string;
      SecureHash id;

      id_string = src_id.address().getHostAddress() + 
	          dst_id.address().getHostAddress();
      id = new SHA1Hash(id_string);
      return ip_pairs_table.contains(id);
  }

  public void begin_route_switch_test()
  {
      final String tag = class_tag +".route_switch";

      //Old configuration: before 04/01/03
      //12.155.161.147:6600   Berkeley, source
      //131.179.112.70:3601 : (SHA1Hash:9ebcd1be d35b) ucla
      //128.95.219.192:3629 : (SHA1Hash:9ea850a4 b4bc) UW
      //18.31.0.190:3636 : (SHA1Hash:9d0d8157 1746)    MIT, destination

      //New configuration: after 04/01/03
      //12.155.161.147:6600   Berkeley, source
      //131.179.112.70:3602 : (SHA1Hash:f1cea4ba e3b7)    ucla
      //128.95.219.192:3628 : (SHA1Hash:f139caee 60a9)    uw
      //18.31.0.190:3608 : (SHA1Hash:fbe09e7c b278) MIT, destination

      currTest = ROUTE_SWITCH;

      if (DEBUG)
	  Debug.printtagln(tag, "route switch test with " + 
			   attached_nodes.size() + " static nodes and " +
			   unattached_nodes.size() + " dynamic nodes...");

      int size = attached_nodes.size();
      dispatch_route_requests(1, 0, ROUTE_SWITCH);

      /*
      int         samples = 50;
      int         i, j, k;
      String []   node_str1, node_str2, node_str3, node_str4;
      String      str1, str2, str3, str4;
      String      str20,str30,str40;
      NodeId      node1, node2, node3, node4;
      SecureHash  [] guid1, guid2, guid3, guid4;

      node_str1 = new String[100];
      node_str2 = new String[100];
      node_str3 = new String[100];
      node_str4 = new String[100];
      guid1 = new SecureHash[100];
      guid2 = new SecureHash[100];
      guid3 = new SecureHash[100];
      guid4 = new SecureHash[100];
      try {
	  for (i = 0; i < samples; i++)
	  {
	      node_str1[i] = "169.229.51.252:" + Integer.toString(3600+i);
	      node1 = new NodeId (node_str1[i]);
	      guid1[i] = new SHA1Hash (node1);
	      
	      node_str2[i] = "131.179.112.70:" + Integer.toString(3600+i); //ucla
	      node2 = new NodeId (node_str2[i]);
	      guid2[i] = new SHA1Hash (node2);

	      node_str3[i] = "128.95.219.192:" + Integer.toString(3600+i); //UW
	      node3 = new NodeId (node_str3[i]);
	      guid3[i] = new SHA1Hash (node3);
	      
	      node_str4[i] = "18.31.0.190:" + Integer.toString(3600+i); //MIT
	      node4 = new NodeId (node_str4[i]);
	      guid4[i] = new SHA1Hash (node4);
	      
	      System.err.println (node_str1[i] + " : " + guid1[i].toString() + "    " + 
				  node_str2[i] + " : " + guid2[i].toString() + "    " + 
				  node_str3[i] + " : " + guid3[i].toString() + "    " + 
				  node_str4[i] + " : " + guid4[i].toString());
	  }
      } catch( Exception e ) {
	  Carp.die( "Exception creating NodeId! " + e );
      }

      for (i = 0 ; i < samples; i ++)
      {
	  str2 = (guid2[i].toString()).substring(10, 11);
	  for (j = 0 ; j < samples; j ++)
	  {
	      str3 = (guid3[j].toString()).substring(10, 11);
	      for (k = 0 ; k < samples; k ++)
	      {
		  str4 = (guid4[k].toString()).substring(10, 11);
		  if (str2.compareTo(str3) == 0 &&
		      str3.compareTo(str4) == 0 )
		  {
		      str20 = (guid2[i].toString()).substring(11, 12);
		      str30 = (guid3[j].toString()).substring(11, 12);
		      str40 = (guid4[k].toString()).substring(11, 12);
		      if (str20.compareTo(str30) == 0 &&
			  str30.compareTo(str40) != 0 )
		      {

			  System.err.println ("\n"); 
			  System.err.println ("Find the guid we want:"); 
			  System.err.println (node_str2[i] + " : " + guid2[i].toString() + "    " + 
					      node_str3[j] + " : " + guid3[j].toString() + "    " + 
					      node_str4[k] + " : " + guid4[k].toString());
			  return;
		      }
		  }
	      }
	  }
      }
      */
  }


  public void begin_misrouting_latency_test()
  {
      final String tag = class_tag +".misrouting_latency";
      if (DEBUG)
	  Debug.printtagln(tag, "Misrouting_latency " + 
			   attached_nodes.size() + " static nodes and " +
			   unattached_nodes.size() + " dynamic nodes...");

      currTest = MISROUTING_LATENCY;
      dispatch_routing_path_requests();
	
      Iterator     srcit; 
      FullTestMsg  ftm;
      NodeId       srcid;
      SecureHash   id;

      total_paths = new OverlayPath[40000];
      resp_nodes = 0;
      num_paths = 0;
      srcit = fixedNodes.iterator();
      while ( srcit.hasNext() ) 
      {
	  srcid = (NodeId)srcit.next();
	  id = genID();
	  ftm = new FullTestMsg(srcid, FullTestMsg.ROUTING_PATH_RESULTS, 
				id, 0, 4);
	  _classifier.dispatch_later(ftm, 30000);
      }
  }

  public void begin_multicast_bandwidth_test()
  {
      final String tag = class_tag +".multicast_bandwidth";
      if (DEBUG)
	  Debug.printtagln(tag, "multicast_bandwidth " + 
			   attached_nodes.size() + " static nodes and " +
			   unattached_nodes.size() + " dynamic nodes...");

      Debug.printtagln(tag, "Type code for RouteMsg: " + 
		       TypeTable.type_code("ostore.tapestry.impl.RouteMsg"));

      currTest = MULTICAST_BANDWIDTH;
      dispatch_routing_path_requests();
	
      Iterator     srcit; 
      FullTestMsg  ftm;
      NodeId       srcid;
      SecureHash   id;

      total_paths = new OverlayPath[500];
      resp_nodes = 0;
      num_paths = 0;
      srcit = fixedNodes.iterator();
      while ( srcit.hasNext() ) 
      {
	  srcid = (NodeId)srcit.next();
	  id = genID();
	  ftm = new FullTestMsg(srcid, FullTestMsg.ROUTING_PATH_RESULTS, 
				id, 0, 5);
	  _classifier.dispatch_later(ftm, 30000);
      }
  }

  public void begin_static_lookup_test()
  {
      final String tag = class_tag +".static_lookup";
      if (DEBUG)
	  Debug.printtagln(tag, "Static lookup Test with " + 
			   attached_nodes.size() + " static nodes and " +
			   unattached_nodes.size() + " dynamic nodes...");

      currTest = STATIC_LOOKUP;
      if (unattached_nodes.size() <= 5)
	  dispatch_publish_requests_on_single_node();
      else
	  dispatch_single_integrate_request();
  }


  public void begin_large_network_change() 
  {
      //initial static network => publish => integration => routing => 
      //lookup => deletion => lookup;

      final String tag = class_tag +".network_change";
      currTest = NETWORK_CHANGE;

      if (DEBUG)
	  Debug.printtagln(tag, "Routing, objects redistribution and " +
			   "location when large portion of the network " +
			   "has been changed.");

    int num_snodes, num_dnodes;
    num_snodes =   attached_nodes.size();
    num_dnodes = unattached_nodes.size();
    if (num_snodes < 2)
    {
	if (DEBUG)
	    Debug.printtagln(tag, "Too few static nodes : " + num_snodes);
	return;
    }    
    /*    if (num_dnodes < 1)
    {
	if (DEBUG)
	    Debug.printtagln(tag, "Too few dynamic nodes : " + num_dnodes);
	return;
    }
    */
    //start to publish objects.
    dispatch_publish_requests(20);

    return;
  }

    public void begin_network_maintence_under_heavy_load()
    {
	final String tag = class_tag +".network_maintenance";
	currTest = NETWORK_CHANGE;
	
	if (DEBUG)
	    Debug.printtagln(tag, "Network maintenance under heavy load...");

    }

    public void begin_network_operation_under_heavy_dynamic()
    {
	final String tag = class_tag +".network_operation";

	currTest = NETWORK_OPERATION;
	
	if (DEBUG)
	    Debug.printtagln(tag, "Network operation(routing, lookup...)" +
			     " under heavy dynamic...");

	int num_snodes, num_dnodes;
	num_snodes =   attached_nodes.size();
	num_dnodes = unattached_nodes.size();
	if (num_snodes < 2)
	{
	    if (DEBUG)
		Debug.printtagln(tag, "Too few static nodes : " + num_snodes);
	    return;
	}    
	/*    if (num_dnodes < 1)
	{
	if (DEBUG)
	Debug.printtagln(tag, "Too few dynamic nodes : " + num_dnodes);
	return;
	}
	*/
	//start to publish objects.
	dispatch_publish_requests(4);
	dispatch_route_requests(DYNAMIC_ROUTE);
	//	dispatch_integrate_requests();
	
	return;
    }

    public void dispatch_integrate_requests()
    {
	final String tag = class_tag +".integrate_request";
	ListIterator li = unattached_nodes.listIterator();
	LinkedList join = new LinkedList();
	FullTestMsg ftm;
	SecureHash id;
	NodeId nid;
	int    i;

	inserted_nodes = 0;
	insertion_times++;
	waitingJoins.clear();

	Debug.printtagln(tag, "The size of current stable network: "+ 
			 attached_nodes.size() + " nodes.");
	if (DEBUG)
	    Debug.printtagln(tag, "The # of dynamic nodes waiting for join: "+ unattached_nodes.size());
	i = 0;
	while (li.hasNext()) 
	{
	    i++;
	    nid = (NodeId) li.next();
	    id = genID();
	    ftm = new FullTestMsg(nid, FullTestMsg.INTEGRATE_REQUEST,
				  id, 0, null,  (NodeId)attached_nodes.get(1));
	    join.add(ftm);
	    waitingJoins.add(nid);
	    if (DEBUG)
		Debug.printtagln(tag, "Send " + i +" th INTEGRATE_REQUEST for " + 
				 nid + " with "+ftm );

	    //For timeout...
	    long waiting_time = 300000 + 100*i;
	    Object token = 
		_classifier.dispatch_later(new TimeOutEvent(ftm.peer, ftm.requestID,
							    ftm.type), waiting_time);
	    waitingTimeouts.put(ftm.requestID, token);
	}
	dispatchList(join);

	insertion_start_time =  Clock.current_date(_self_node_id).getTime();
	return;
    }

    public void dispatch_integrate_requests_with_number(int node_num)
    {
	final String tag = class_tag +".integrate_request";
	ListIterator li = unattached_nodes.listIterator();
	LinkedList join = new LinkedList();
	FullTestMsg ftm;
	SecureHash id;
	NodeId nid;
	int    i;

	inserted_nodes = 0;
	insertion_times++;
	waitingJoins.clear();

	Debug.printtagln(tag, "The size of current stable network: "+ 
			 attached_nodes.size() + " nodes.");
	if (DEBUG)
	    Debug.printtagln(tag, "The # of dynamic nodes waiting for join: "+ unattached_nodes.size());
	i = 0;
	while (li.hasNext()) 
	{
	    i++;
	    nid = (NodeId) li.next();
	    id = genID();
	    ftm = new FullTestMsg(nid, FullTestMsg.INTEGRATE_REQUEST,
				  id, 0, null,  (NodeId)attached_nodes.get(1));
	    join.add(ftm);
	    waitingJoins.add(nid);
	    if (DEBUG)
		Debug.printtagln(tag, "Send " + i +" th INTEGRATE_REQUEST for " + 
				 nid + " with "+ftm );
	    if (i >= node_num)
		break;
	}
	dispatchList(join);

	insertion_start_time =  Clock.current_date(_self_node_id).getTime();
	return;
    }


    public void dispatch_single_integrate_request()
    {
	final String tag = class_tag +".single_integrate_request";
	ListIterator li = unattached_nodes.listIterator();
	FullTestMsg ftm;
	SecureHash id;
	NodeId nid;
	int    i;

	waitingJoins.clear();

	Debug.printtagln(tag, "The # of dynamic nodes waiting for join: "+ unattached_nodes.size());
	if (unattached_nodes.size() > 0)
	{
	    nid = (NodeId) unattached_nodes.get(0);
	    id = genID();
	    ftm = new FullTestMsg(nid, FullTestMsg.INTEGRATE_REQUEST,
				  id, 0, null,  (NodeId)attached_nodes.get(0));
	    waitingJoins.add(nid);
	    dispatch(ftm);
	}

	return;
    }

    public void dispatch_clear_bandwidth_requests(boolean useful, long waiting_time)
    {
	//Object location finished, delete all nodes which don't hava 
	//physical data stored.
	final String tag = class_tag +".single_integrate_request";
	String comment;
	SecureHash id, node_guid;
	NodeId     node_id;
	LinkedList list = new LinkedList();
	int  adj;

	if (useful)
	    comment = "USEFUL_DATA";
	else
	    comment = "USELESS_DATA";
	adj = attached_nodes.size();
	if (adj > 1) 
	{
	    for (int i=0; i< adj; i++) 
	    {
		node_id = (NodeId)attached_nodes.get(i);
		id = genID();
		FullTestMsg ftm = new FullTestMsg((NodeId) node_id,
						  FullTestMsg.CLEAR_BANDWIDTH_COUNTS,
						  id, 0, null, null, comment);
		waitingDetaches.add(ftm.requestID);
		list.add(ftm);
		if (DEBUG)
		    Debug.printtagln(tag, "Send CLEAR_BANDWIDTH_COUNTS for " + 
				     node_id + " with "+ftm );
	    }
	    if (waiting_time == 0)
		dispatchList(list);
	    else
		dispatchList(list, waiting_time);
	}

	return;
    }     

    public void dispatch_clear_bandwidth_requests(String comment, long waiting_time)
    {
	//Object location finished, delete all nodes which don't hava 
	//physical data stored.
	final String tag = class_tag +".single_integrate_request";
	SecureHash id, node_guid;
	NodeId     node_id;
	LinkedList list = new LinkedList();
	int  adj;

	adj = attached_nodes.size();
	if (adj > 1) 
	{
	    for (int i=0; i< adj; i++) 
	    {
		node_id = (NodeId)attached_nodes.get(i);
		id = genID();
		FullTestMsg ftm = new FullTestMsg((NodeId) node_id,
						  FullTestMsg.CLEAR_BANDWIDTH_COUNTS,
						  id, 0, null, null, comment);
		waitingDetaches.add(ftm.requestID);
		list.add(ftm);
		if (DEBUG)
		    Debug.printtagln(tag, "Send CLEAR_BANDWIDTH_COUNTS for " + 
				     node_id + " with "+ftm );
	    }
	    if (waiting_time == 0)
		dispatchList(list);
	    else
		dispatchList(list, waiting_time);
	}

	return;
    }     

    public void dispatch_detach_requests()
    {
	//Object location finished, delete all nodes which don't hava 
	//physical data stored.
	SecureHash id, node_guid;
	NodeId     node_id;
	LinkedList list = new LinkedList();
	int  adj;

	detached_nodes = 0;
	detach_times++;
	adj = attached_nodes.size();
	if (adj > 1) 
	{
	    for (int i=0; i< adj; i++) 
	    {
		node_id = (NodeId)attached_nodes.get(i);
		node_guid = (SecureHash) nodemap.get(node_id);
		if ( !objects_in_network.containsKey(node_guid) )
		{
		    id = genID();
		    FullTestMsg ftm = new FullTestMsg((NodeId) node_id,
						      FullTestMsg.DETACH_REQUEST,
						      id, 0, null);
		    waitingDetaches.add(ftm.requestID);
		    list.add(ftm);
		}
	    }
	    dispatchList(list);
	}

	return;
    }     

    //Detach all nodes except those in HashSet static_nodes.
    public void dispatch_detach_requests(HashSet static_nodes)
    {
	//Object location finished, delete all nodes which don't hava 
	//physical data stored.
	final String tag = class_tag +".detach_request";

	SecureHash id, node_guid;
	NodeId     node_id;
	LinkedList list = new LinkedList();
	int  adj, j;

	detached_nodes = 0;
	detach_times++;
	adj = attached_nodes.size();
	if (adj > 1) 
	{
	    j = 0;
	    for (int i=0; i< adj; i++) 
	    {
		node_id = (NodeId)attached_nodes.get(i);
		if ( !static_nodes.contains(node_id) )
		{
		    j++;
		    node_guid = (SecureHash) nodemap.get(node_id);
		    id = genID();
		    FullTestMsg ftm = new FullTestMsg((NodeId) node_id,
						      FullTestMsg.DETACH_REQUEST,
						      id, 0, null);
		    waitingDetaches.add(ftm.requestID);
		    list.add(ftm);
		    if (DEBUG)
			Debug.printtagln(tag, "Send " + j +" th DETACH_REQUEST for " + 
					 node_id + " with "+ftm );

		}
	    }
	    dispatchList(list);
	}

	return;
    }     

    //Detach all nodes (one by one) except those in HashSet static_nodes.
    public void dispatch_single_detach_requests(HashSet static_nodes)
    {
	//Object location finished, delete all nodes which don't hava 
	//physical data stored.
	final String tag = class_tag +".detach_request";

	SecureHash id, node_guid;
	NodeId     node_id;
	LinkedList list = new LinkedList();
	int  adj, j;

	adj = attached_nodes.size();
	for (int i=0; i< adj; i++) 
	{
	    node_id = (NodeId)attached_nodes.get(i);
	    if ( !static_nodes.contains(node_id) )
	    {
		node_guid = (SecureHash) nodemap.get(node_id);
		id = genID();
		FullTestMsg ftm = new FullTestMsg((NodeId) node_id,
						  FullTestMsg.DETACH_REQUEST,
						  id, 0, null);
		waitingDetaches.add(ftm.requestID);
		dispatch(ftm);
		break;
	    }
	}

	return;
    }     

    public void dispatch_routing_path_requests()
    {
      final String tag = class_tag +".routing_path_request";
      Iterator     srcit, destit; 
      FullTestMsg  ftm;
      SecureHash   id, srcguid, destguid;
      SecureHash   guids[];
      NodeId       srcid, destid;
      int          i, size, diff;
      
      size = fixedNodes.size();
      if (DEBUG)
	  Debug.printtagln(tag, "Fixed nodes in current Tapestry: " + size);
      i = 0;
      srcit = fixedNodes.iterator();
      while ( srcit.hasNext() ) 
      {
	  srcid = (NodeId)srcit.next();
	  srcguid = (SecureHash) nodemap.get( srcid );
	  destit = fixedNodes.iterator();
	  while ( destit.hasNext() )
	  {
	      destid = (NodeId)destit.next();
	      if ( !srcid.equals(destid) )
	      {
		  destguid = (SecureHash) nodemap.get( destid );
		  diff = Router.guidCompare(srcguid, destguid);
		  //		  if (diff < 35)
		  {
		      guids = new SecureHash[2];
		      guids[0] = srcguid;
		      guids[1] = destguid;
		      id = genID();
		      ftm = new FullTestMsg(srcid, FullTestMsg.ROUTING_PATH_REQUEST, 
					    id, 0, guids);
		      dispatch(ftm);
		      //		      return;
		  }
	      }
	  }
      }
    }

    public void dispatch_route_requests(int test)
    {
      final String tag = class_tag +".route_request";
      Iterator     it; 
      FullTestMsg  ftm;
      SecureHash   id, dest[];
      NodeId       nodeid, dest_id[];
      int          size, i;
      
      route_times++;
      waitingRoutes.clear();

      size = fixedNodes.size();
      dest = new SecureHash[size];
      dest_id = new NodeId[size];
      if (DEBUG)
	  Debug.printtagln(tag, "Fixed nodes in current Tapestry: " + size);
      i = 0;
      it = fixedNodes.iterator(); 
      while ( it.hasNext() ) 
      {
	  nodeid = (NodeId)it.next();
	  dest_id[i] = nodeid;
	  dest[i] = (SecureHash) nodemap.get( nodeid );
	  if (DEBUG)
	      Debug.printtagln(tag, "Node " + i + " : " + nodeid + dest[i]);

	  i++;
      }

      LinkedList li = new LinkedList();
      it = fixedNodes.iterator();
      while ( it.hasNext() )
	  //      if ( it.hasNext() )
      {
	  id = genID();
	  nodeid = (NodeId)it.next();
	  ftm = new FullTestMsg(nodeid, FullTestMsg.ROUTE_REQUEST, id, 0, dest, dest_id);
	  ftm.test = test;
	  li.add(ftm);
	  waitingRoutes.add(nodeid);
	  if (DEBUG)
	      Debug.printtagln(tag, "Send out a ROUTE_REQUEST: " + ftm);
	  
      }

      dispatchList(li);
      return;
    }

    //This is for the rechability test.
    public void dispatch_route_requests(double failure, int waiting, int test)
    {
      final String tag = class_tag +".route_request";
      Iterator     it; 
      FullTestMsg  ftm;
      SecureHash   id, dest[];
      NodeId       nodeid, dest_id[];
      int          size, i;
      
      route_times++;
      waitingRoutes.clear();

      size = fixedNodes.size();
      dest = new SecureHash[size];
      dest_id = new NodeId[size];
      if (DEBUG)
	  Debug.printtagln(tag, "Fixed nodes in current Tapestry: " + size);
      i = 0;
      it = fixedNodes.iterator(); 
      while ( it.hasNext() ) 
      {
	  nodeid = (NodeId)it.next();
	  dest_id[i] = nodeid;
	  dest[i] = (SecureHash) nodemap.get( nodeid );
	  //	  if (DEBUG)
	  //	      Debug.printtagln(tag, "Node " + i + " : " + nodeid + dest[i]);

	  i++;
      }

      LinkedList li = new LinkedList();
      it = fixedNodes.iterator();
      while ( it.hasNext() )
      {
	  id = genID();
	  nodeid = (NodeId)it.next();
	  ftm = new FullTestMsg(nodeid, FullTestMsg.ROUTE_REQUEST, id, 0, 
				dest, dest_id);
	  ftm.failure = failure;
	  ftm.test = test;
	  li.add(ftm);
	  waitingRoutes.add(nodeid);
	  //	  if (DEBUG)
	  //	      Debug.printtagln(tag, "Send out a ROUTE_REQUEST: " + ftm);
	  
	  _classifier.dispatch_later(ftm, waiting);
      }

      //      dispatchList(li);
      return;
    }

    //This is for route_switch and multicast test.
    public void dispatch_route_requests(int src_index, int dest_index, int test)
    {
      final String tag = class_tag +".route_request";
      Iterator     it; 
      FullTestMsg  ftm;
      SecureHash   id, dest[];
      NodeId       nodeid, dest_id[], src_id=null;
      int          size, i;
      
      route_times++;
      waitingRoutes.clear();

      size = 1;
      dest = new SecureHash[size];
      dest_id = new NodeId[size];
      i = 0;
      it = fixedNodes.iterator(); 
      while ( it.hasNext() ) 
      {
	  nodeid = (NodeId)it.next();
	  //Dest is at mit
	  if ( nodeid.toString().equalsIgnoreCase("18.31.0.190:3608") )
	  {
	      dest_id[i] = nodeid;
	      dest[i] = (SecureHash) nodemap.get( nodeid );
	      if (DEBUG)
		  Debug.printtagln(tag, "Node " + i + " : " + nodeid + dest[i]);
	      
	      i++;
	      break;
	  }
      }

      LinkedList li = new LinkedList();
      it = fixedNodes.iterator();
      while ( it.hasNext() )
      {
	  id = genID();
	  nodeid = (NodeId)it.next();
	  //source is at berkeley
	  if ( nodeid.toString().equalsIgnoreCase("12.155.161.147:6600") )
	  {
	      src_id = nodeid;
	      ftm = new FullTestMsg(nodeid, FullTestMsg.ROUTE_REQUEST, id, 0, 
				    dest, dest_id);
	      ftm.test = test;
	      ftm.value = 20;
	      li.add(ftm);
	      waitingRoutes.add(nodeid);
	      if (DEBUG)
		  Debug.printtagln(tag, "Send out a ROUTE_REQUEST: " + ftm);

	      break;
	  }
	  
      }
      //      dispatchList(li);

      it = fixedNodes.iterator();
      while ( it.hasNext() )
      {
	  id = genID();
	  nodeid = (NodeId)it.next();
	  //Inject fault to uw machine:
	  if ( nodeid.toString().equalsIgnoreCase("128.95.219.192:3628") )
	  {
	      RouterFaultMsg rfmsg100, rfmsg0;
	      InjectFaultMsg ifmsg100, ifmsg0;
	      rfmsg100 = new RouterFaultMsg(nodeid, 100, true, src_id); 
	      ifmsg100 = new InjectFaultMsg(nodeid, 100, true, src_id);
	      rfmsg0   = new RouterFaultMsg(nodeid,   0, true, src_id); 
	      ifmsg0   = new InjectFaultMsg(nodeid,   0, true, src_id);

	      long time = (long)(probing_frequency*rand.nextDouble());
	      if (DEBUG)
		  Debug.printtagln(tag, "At " + time +
				   " Inject fault to node: " + nodeid);
	      
	      time = 10000 + time;
	      _classifier.dispatch_later(rfmsg100, time);
	      _classifier.dispatch_later(ifmsg100, time);

	      /*
	      _classifier.dispatch_later(rfmsg0,   10000);
	      _classifier.dispatch_later(ifmsg0,   10000);
	      _classifier.dispatch_later(rfmsg100, 16000);
	      _classifier.dispatch_later(ifmsg100, 16000);
	      _classifier.dispatch_later(rfmsg0,   22000);
	      _classifier.dispatch_later(ifmsg0,   22000);

	      _classifier.dispatch_later(rfmsg100, 28000);
	      _classifier.dispatch_later(ifmsg100, 28000);
	      _classifier.dispatch_later(rfmsg0,   34000);
	      _classifier.dispatch_later(ifmsg0,   34000);

	      _classifier.dispatch_later(rfmsg100, 40000);
	      _classifier.dispatch_later(ifmsg100, 40000);
	      _classifier.dispatch_later(rfmsg0,   46000);
	      _classifier.dispatch_later(ifmsg0,   46000);

	      _classifier.dispatch_later(rfmsg100, 52000);
	      _classifier.dispatch_later(ifmsg100, 52000);
	      _classifier.dispatch_later(rfmsg0,   58000);
	      _classifier.dispatch_later(ifmsg0,   58000);

	      _classifier.dispatch_later(rfmsg100, 64000);
	      _classifier.dispatch_later(ifmsg100, 64000);
	      _classifier.dispatch_later(rfmsg0,   70000);
	      _classifier.dispatch_later(ifmsg0,   70000);

	      _classifier.dispatch_later(rfmsg100, 76000);
	      _classifier.dispatch_later(ifmsg100, 76000);
	      _classifier.dispatch_later(rfmsg0,   82000);
	      _classifier.dispatch_later(ifmsg0,   82000);

	      _classifier.dispatch_later(rfmsg100, 88000);
	      _classifier.dispatch_later(ifmsg100, 88000);
	      _classifier.dispatch_later(rfmsg0,   94000);
	      _classifier.dispatch_later(ifmsg0,   94000);

	      _classifier.dispatch_later(rfmsg100, 100000);
	      _classifier.dispatch_later(ifmsg100, 100000);
	      _classifier.dispatch_later(rfmsg0,   106000);
	      _classifier.dispatch_later(ifmsg0,   106000);

	      _classifier.dispatch_later(rfmsg100, 112000);
	      _classifier.dispatch_later(ifmsg100, 112000);
	      _classifier.dispatch_later(rfmsg0,   118000);
	      _classifier.dispatch_later(ifmsg0,   118000);
	      */

	      break;
	  }
	  
      }


      return;
    }

    //parameter: num is the number of objects that each node will publish to the network.
    public void dispatch_publish_requests(int num)
    {
	final String tag = class_tag +".publish_request";
	ListIterator li = attached_nodes.listIterator();
	int num_objects;
	LinkedList publish = new LinkedList();
	FullTestMsg ftm;
	NodeId nid;
	SecureHash id, node_guid, object_guids[];
	int  i;

	publish_times++;
	waitingPublishes.clear();

	num_objects = num;
	if (DEBUG)
	    Debug.printtagln(tag, "Every static node publish " + 
			     num_objects + " objects...");
	while (li.hasNext()) 
	{
	    nid = (NodeId) li.next();
	    node_guid = (SecureHash) nodemap.get(nid);
	    id = genID();
	    if (DEBUG)
		Debug.printtagln(tag, "Objects on node " + nid + node_guid + " :");
	    object_guids = new SecureHash[num_objects];
	    for (i = 0; i < num_objects; i++)
	    {
		object_guids[i] = genID();
		//		if (DEBUG)
		//		    Debug.printtagln(tag, "Object " + i + " : " + object_guids[i]);
	    }
	    objects_in_network.put(node_guid, object_guids);
	    ftm = new FullTestMsg(nid, FullTestMsg.PUBLISH_REQUEST,
				  id, 0, object_guids);
	    publish.add(ftm);
	    waitingPublishes.add(nid);
	    dispatch(ftm);
	}
	on_publishing = true;

	return;
    }

    //Publish 30000 objects by attached_nodes[0].
    public void dispatch_publish_requests_on_single_node()
    {
	final String tag = class_tag +".publish_request";
	ListIterator li = attached_nodes.listIterator();
	int num_snodes, num_objects;
	LinkedList publish = new LinkedList();
	FullTestMsg ftm;
	NodeId nid;
	SecureHash id, node_guid, object_guids[];
	String ip_addr;
	int  i, good_node;

	publish_times++;
	waitingPublishes.clear();

	nid = null;
	num_objects = 10000;

	good_node = 0;
	while (li.hasNext()) 
	{
	    nid = (NodeId) li.next();
	    ip_addr = (nid.address()).getHostAddress();
	    //	    if (ip_addr.equalsIgnoreCase("128.42.6.143"))
	    //	    if (ip_addr.equalsIgnoreCase("131.243.254.35")) //lbl machine.
	    if (ip_addr.equalsIgnoreCase("171.64.64.216")) //stanford machine.
	    {
		good_node = 1;
		break;
	    }
	}

	//If cannot find the specific node, 
	//get the first one in attached_nodes.
	if ( good_node == 0 )
	{
	    li = attached_nodes.listIterator();
	    nid = (NodeId) li.next();
	}
	publish_server = nid;
	node_guid = (SecureHash) nodemap.get(nid);
	id = genID();
	if (DEBUG)
	    Debug.printtagln(tag, "Node :" + nid + node_guid + 
			     " publishes " + num_objects + " objects...");
	object_guids = new SecureHash[num_objects];
	for (i = 0; i < num_objects; i++)
	    object_guids[i] = genID();
	
	objects_in_network.put(node_guid, object_guids);
	ftm = new FullTestMsg(nid, FullTestMsg.PUBLISH_REQUEST,
			      id, 0, object_guids);
	publish.add(ftm);
	waitingPublishes.add(nid);
	dispatch(ftm);

	on_publishing = true;

	return;
    }
    
    public void dispatch_lookup_requests()
    {
	final String tag = class_tag +".lookup_request";
	LinkedList  object_guids;
	FullTestMsg ftm;
	SecureHash  id, sh[];
	NodeId      nid;
	int i;

	lookup_times++;
	waitingLookups.clear();

	object_guids = new LinkedList();
	for (Iterator it=objects_in_network.values().iterator(); it.hasNext(); ) 
	{
	    sh = (SecureHash []) it.next();
	    for (i=0; i<sh.length; i++) 
		object_guids.add(sh[i]);
	}
	    
	i = 0;
	sh = new SecureHash[object_guids.size()];
	for (Iterator it = object_guids.iterator(); it.hasNext(); )
	    sh[i++] = (SecureHash)it.next();

	LinkedList li = new LinkedList();
	i = 0;
	for (i=0; i<attached_nodes.size(); i++) 
	{
	    nid = (NodeId) attached_nodes.get(i);
	    //	    if ( nid.toString().equalsIgnoreCase("128.32.37.205:3631") )
	    {
		id = genID();
		ftm = new FullTestMsg(nid, FullTestMsg.LOOKUP_REQUEST, 
				      id, 0, sh);
		
		ftm.repeat_lookups = false;
		li.add(ftm);
		waitingLookups.add(nid);
	    }
	}
	dispatchList(li);

	return;
    }

    public void dispatch_lookup_requests(HashSet fixednodes, boolean repeat)
    {
	final String tag = class_tag +".lookup_request";
	LinkedList  object_guids;
	FullTestMsg ftm;
	SecureHash  id, sh[];
	NodeId      nid;
	int i;

	lookup_times++;
	waitingLookups.clear();

	object_guids = new LinkedList();
	for (Iterator it=objects_in_network.values().iterator(); it.hasNext(); ) 
	{
	    sh = (SecureHash []) it.next();
	    for (i=0; i<sh.length; i++) 
		object_guids.add(sh[i]);
	}
	i = 0;

	Debug.printtagln(tag, "The # of objects in the network: " + object_guids.size());
	sh = new SecureHash[object_guids.size()];
	for (Iterator it = object_guids.iterator(); it.hasNext(); )
	    sh[i++] = (SecureHash)it.next();

	LinkedList li = new LinkedList();
	Iterator it = fixedNodes.iterator(); 
	while ( it.hasNext() ) 
	{
	    nid = (NodeId)it.next();
	    id = genID();
	    ftm = new FullTestMsg(nid, FullTestMsg.LOOKUP_REQUEST, 
				  id, 0, sh);
	    ftm.repeat_lookups = repeat;
	    li.add(ftm);
	    waitingLookups.add(nid);
	}

	dispatchList(li);
	return;
    }
    
    public void dispatch_lookup_requests_by_group()
    {
	final String tag = class_tag +".lookup_request";
	LinkedList  object_guids;
	FullTestMsg ftm;
	SecureHash  id, sh[];
	NodeId      nid, grouped_nodes[][];
	int         num_group, num_machine;
	int i, j;

	waitingLookups.clear();

	object_guids = new LinkedList();
	for (Iterator it=objects_in_network.values().iterator(); it.hasNext(); ) 
	{
	    sh = (SecureHash []) it.next();
	    for (i=0; i<sh.length; i++) 
		object_guids.add(sh[i]);
	}
	i = 0;

	Debug.printtagln(tag, "The # of objects in the network: " + object_guids.size());
	sh = new SecureHash[object_guids.size()];
	for (Iterator it = object_guids.iterator(); it.hasNext(); )
	    sh[i++] = (SecureHash)it.next();

	num_group = 30;
	num_machine = 40;
	grouped_nodes = machines_grouped_by_uniform_latency(publish_server, num_group, num_machine);

	for (i=0; i<num_group; i++) 
	{
	    if (DEBUG)
		Debug.printtagln(tag, "Machines of latency between " + i +
				 " and " + i*300/num_group);
	    for (j=0; j<5; j++) 
	    {
		nid = grouped_nodes[i][j];
		if (nid != null)
		{
		    if (DEBUG)
			Debug.printtagln(tag, "Machines[" + j + "] = " + nid);
		} 
	    }
	}

	LinkedList li = new LinkedList();
	for (j=0; j<5; j++) 
	{
	    for (i=0; i<num_group; i++) 
	    {
		nid = grouped_nodes[i][j];
		if (nid != null)
		{
		    id = genID();
		    ftm = new FullTestMsg(nid, FullTestMsg.LOOKUP_REQUEST, 
					  id, 0, sh);
		    
		    li.add(ftm);
		    waitingLookups.add(nid);
		}
	    }
	}
	dispatchList(li);

	return;
    }
    
    public void dispatch_lookup_requests_single_node(NodeId client)
    {
	final String tag = class_tag +".lookup_request";
	LinkedList  object_guids;
	FullTestMsg ftm;
	SecureHash  id, sh[];
	NodeId      nid;
	int i;

	waitingLookups.clear();

	object_guids = new LinkedList();
	for (Iterator it=objects_in_network.values().iterator(); it.hasNext(); ) 
	{
	    sh = (SecureHash []) it.next();
	    for (i=0; i<sh.length; i++) 
		object_guids.add(sh[i]);
	}
	    
	i = 0;
	sh = new SecureHash[object_guids.size()];
	for (Iterator it = object_guids.iterator(); it.hasNext(); )
	    sh[i++] = (SecureHash)it.next();

	LinkedList li = new LinkedList();
	nid = client;
	id = genID();
	ftm = new FullTestMsg(nid, FullTestMsg.LOOKUP_REQUEST, 
			      id, 0, sh);
		
	li.add(ftm);
	waitingLookups.add(nid);
	dispatchList(li);

	return;
    }

    public void dispatch_unpublish_request()
    {
	unpublish_times++;
	return;
    }

    public boolean handle_response_route_request(FullTestRespMsg msg)
    {
	final String tag = class_tag +".route_request";

	boolean routing_finish = false;

	if (!msg.success()) 
	{
	    failedNodes.add(msg.peer);
	    if (DEBUG)
		Debug.printtagln(tag, "Failed route request from node: "+msg.peer);
	}
	if (waitingRoutes.remove(msg.peer)==false && DEBUG)
	    Debug.printtagln(tag, "Unexpected route response message: "+msg);
	if (waitingRoutes.size() == 0) 
	{
	    if (failedNodes.size() == 0)
		System.err.println("Route Test passed!\n");
	    else
	    {
		System.err.println("Route Test failed on nodes:"+
				   failedNodes + "!\n");
		System.exit(0);
	    }
	    routing_finish = true;
	}
	
	return routing_finish;
    }

    public boolean handle_response_lookup_request(FullTestRespMsg msg)
    {
	final String tag = class_tag +".lookup_request";

	boolean lookup_finish = false;

	if (!msg.success()) 
	{
	    failedNodes.add(msg.peer);
	    if (DEBUG)
		Debug.printtagln(tag, "Failed lookup request from node: "+msg.peer);
	}
	if (waitingLookups.remove(msg.peer)==false && DEBUG)
	    Debug.printtagln(tag, "Unexpected route response message: "+msg);
	if (waitingLookups.size() == 0) 
	{
	    if (failedNodes.size() == 0)
		System.err.println("Lookup Test passed!\n");
	    else
		System.err.println("Lookup Test failed on nodes:"+
				   failedNodes + "!\n");
	    lookup_finish = true;
	}

	return lookup_finish;
    }

    public boolean handle_response_integrate_request(FullTestRespMsg msg)
    {
	final String tag = class_tag +".integrate_request";

	if (unattached_nodes.remove(msg.peer))
	    attached_nodes.add(msg.peer);
	else if (DEBUG)
	    Debug.printtagln(tag, "Unexpected integrate node received");
	if (waitingJoins.remove(msg.peer)==false && DEBUG)
	    Debug.printtagln(tag, "Unexpected response message: "+ msg);
	if (waitingJoins.size() == 0) 
	    return true;
	return false;
    }

    public boolean handle_response_detach_request(FullTestRespMsg msg)
    {
	final String tag = class_tag +".detach_request";
	if (attached_nodes.remove(msg.peer))
	    unattached_nodes.add(msg.peer);
	else if (DEBUG)
	    Debug.printtagln(tag, "Unexpected detach node received");
	if (waitingDetaches.remove(msg.getID())==false && DEBUG)
	    Debug.printtagln(tag, "Unexpected response message: "+msg);
	else if (waitingDetaches.size() == 0) 
	    return true;
	return false;
    }

    public void begin_waiting_msgs_debug_test() {
	final String tag = class_tag + ".begin_waiting_msgs_debug_test";
	System.out.println("Sending msg to FullTestMember asking what kind of msgs do they waiting for...");
	int num_snodes, num_dnodes, num_objects;
	num_snodes = attached_nodes.size();
	//num_dnodes = unattached_nodes.size();
	if (num_snodes < 2) {
	    if (DEBUG)
		Debug.printtagln(tag, "Too few static nodes : " + num_snodes);
	    return;
	}

	LinkedList messages = new LinkedList();
	DebugMsg dm;
	NodeId nid;
	SecureHash id, node_guid;
	int i;

	ListIterator li = attached_nodes.listIterator();
	while (li.hasNext())
	{
	    nid = (NodeId) li.next();
	    id = genID();
	    dm = new DebugMsg(nid, DebugMsg.INSERTION_STATE, id);
	    messages.add(dm);
	    waitingResponses.add(nid);
	    _classifier.dispatch_later(dm, 1);
	}

	li = unattached_nodes.listIterator();
	while (li.hasNext())
	{
	    nid = (NodeId) li.next();
	    id = genID();
	    dm = new DebugMsg(nid, DebugMsg.INSERTION_STATE, id);
	    messages.add(dm);
	    waitingResponses.add(nid);
	    _classifier.dispatch_later(dm, 1);
	}

	return;
    }


    public void begin_routing_table_debug_test() {
	final String tag = class_tag + ".begin_routing_table_debug_test";
	currTest = ROUTING_TABLE_DEBUG;
	
	System.out.println("Starting Routing Table Debug Test");

	int num_snodes, num_dnodes, num_objects;
	num_snodes = attached_nodes.size();
	//num_dnodes = unattached_nodes.size();
	if (num_snodes < 2) {
	    if (DEBUG)
		Debug.printtagln(tag, "Too few static nodes : " + num_snodes);
	    return;
	}
	ListIterator li = attached_nodes.listIterator();
	LinkedList messages = new LinkedList();
	DebugMsg dm;
	NodeId nid;
	SecureHash id, node_guid;
	int i;

	num_objects = 2 * num_snodes;

	while (li.hasNext())
	{
	    nid = (NodeId) li.next();
	    //node_guid = (SecureHash) nodemap.get(nid);
	    id = genID();

	    // Send Request for Routing Table
	    dm = new DebugMsg(nid, DebugMsg.ROUTING_TABLE_REQUEST, id, 3);
	    messages.add(dm);
	    waitingResponses.add(nid);
	}

	dispatchList(messages);
	return;
    }

    public void begin_back_pointer_debug_test() {
	final String tag = class_tag + ".begin_back_pointer_debug_test";
	currTest = BACK_POINTER_DEBUG;

	System.out.println("Starting Back Pointer Debug Test");

	ListIterator li = attached_nodes.listIterator();
	LinkedList messages = new LinkedList();
	DebugMsg dm;
	NodeId nid;
	SecureHash id;

	while (li.hasNext()) {
	    nid = (NodeId) li.next();
	    id = genID();

	    // Send Request for Back (Object) Pointers
	    dm = new DebugMsg(nid, DebugMsg.BACK_PTR_REQUEST, id);
	    messages.add(dm);
	}

	dispatchList(messages);
	return;
    }

    public void begin_reverse_pointer_debug_test() {
        final String tag = class_tag + ".begin_reverse_pointer_debug_test";
        currTest = REVERSE_POINTER_DEBUG;

        System.out.println("Starting Reverse Pointer Debug Test");

        ListIterator li = attached_nodes.listIterator();
        LinkedList messages = new LinkedList();
        DebugMsg dm;
        NodeId nid;
        SecureHash id;

        while (li.hasNext()) {
            nid = (NodeId) li.next();
            id = genID();

	    // Send Request for Reverse Pointers
	    dm = new DebugMsg(nid, DebugMsg.REVERSE_PTR_REQUEST, id, 3);
	    messages.add(dm);
	}

	dispatchList(messages);
	return;
    }

    public NodeId [][] machines_grouped_by_uniform_latency(NodeId server, int num_group, int num_machine)
    {
	double pings[][];
	pings = parse_all_pairs_pings("/home/hling/oceanstore/pond/run/x-pl.app");

	Integer bi1=null, bi2=null;
	int index1, index2;
	int    group;
	double rtt_ip, interval;
	NodeId grouped_nodes[][];
	String ip1, ip2;
	int count[], i, j;
	
	
	interval = 300 / num_group;
	count = new int[200];
	grouped_nodes  = new NodeId[num_group][];
	for (i = 0 ; i < num_group; i++)
	{
	    count[i] = 0;
	    grouped_nodes[i]  = new NodeId[num_machine];
	}
	for (i = 0 ; i < num_group; i++)
	    for (j = 0; j < num_machine; j ++)
		grouped_nodes[i][j]  = null;


	ip1 = (server.address()).getHostAddress();
	bi1 = (Integer) server_indices.get( server.address() );
	ListIterator li = attached_nodes.listIterator();
	NodeId client;
	SecureHash id;
	while (li.hasNext()) 
	{
	    client = (NodeId) li.next();
	    ip2 = (client.address()).getHostAddress();
	    if ( !ip1.equalsIgnoreCase(ip2) )
	    {
		bi2 = (Integer) server_indices.get( client.address() );
		if( bi1 == null || bi2 == null) {
		    Carp.die( "Could not find servers in the all_pairs_ping file: " + 
			      server.address() + " and " + client.address() );
		}
		index1 = bi1.intValue();
		index2 = bi2.intValue();
		rtt_ip = pings[index1][index2];
		if (rtt_ip > 1.0 )
		{
		    group = (int)(rtt_ip/interval);
		    if (group < num_group && count[group] < num_machine)
		    {
			grouped_nodes[group][count[group]] = client;
			count[group] = count[group] + 1;
		    }
		    else
			System.err.print("Two many data points: " + group + " " + 
					 count[group] +"\n");
		}
	    }
	}

	return grouped_nodes;
    }

    public NodeId [][] machines_grouped_by_latency(NodeId server, int num_group, int num_machine)
    {
	double pings[][];
	pings = parse_all_pairs_pings("/home/hling/oceanstore/pond/run/x-pl.app");

	Integer bi1=null, bi2=null;
	int index1, index2;
	int    group;
	double rtt_ip, interval;
	NodeId grouped_nodes[][];
	int count[], i, j;
	
	
	interval = 300 / num_group;
	count = new int[200];
	grouped_nodes  = new NodeId[num_group][];
	for (i = 0 ; i < num_group; i++)
	{
	    count[i] = 0;
	    grouped_nodes[i]  = new NodeId[num_machine];
	}
	for (i = 0 ; i < num_group; i++)
	    for (j = 0; j < num_machine; j ++)
		grouped_nodes[i][j]  = null;

	bi1 = (Integer) server_indices.get( server.address() );
	ListIterator li = attached_nodes.listIterator();
	NodeId client;
	SecureHash id;
	while (li.hasNext()) 
	{
	    client = (NodeId) li.next();
	    //if ( !client.equalsIgnoreCase(server) )
	    {
		bi2 = (Integer) server_indices.get( client.address() );
		if( bi1 == null || bi2 == null) {
		    Carp.die( "Could not find servers in the all_pairs_ping file: " + 
			      server.address() + " and " + client.address() );
		}
		index1 = bi1.intValue();
		index2 = bi2.intValue();
		rtt_ip = pings[index1][index2];
		if (rtt_ip > 1.0 )
		{
		    group = (int)(rtt_ip/interval);
		    if (group < num_group && count[group] < num_machine)
		    {
			grouped_nodes[group][count[group]] = client;
			count[group] = count[group] + 1;
		    }
		    else
			System.err.print("Two many data points: " + group + " " + 
					 count[group] +"\n");
		}
	    }
	}

	return grouped_nodes;
    }


    public void begin_handle_lookup_rdp_results()
    {
        currTest = LOOKUP_RDP_RESULTS;

	double pings[][];
	pings = parse_all_pairs_pings("/home/hling/oceanstore/pond/run/x-pl.app");

	BufferedReader br = null;
	try {
	    //	    br = new BufferedReader( new FileReader(
	    //	    "/home/hling/planetlab/stat/origin/lookup/171.64.64.216x30_rdp_raw_pti_1029_large.dat") );
	    //	    br = new BufferedReader( new FileReader(
	    //	    "/home/hling/planetlab/stat/origin/lookup/171.64.64.216x30_rdp_raw_nopti_1024.dat") );
	    br = new BufferedReader( new FileReader(
	    "/home/hling/planetlab/stat/origin/lookup/171.64.64.216x30_rdp_raw_opti444_1119.dat") );

	} catch( FileNotFoundException fnfe ) {
	    Carp.die( "Could not open all_pairs_pings_file: " + fnfe );
	}

	NodeId server1=null, server2=null;
	Integer bi1=null, bi2=null;
	int pos, start, end;
	int index1, index2;
	String ip1, ip2, time_st;
	int    time_int, group;
	double rtt_ip, rtt_tap, rdp;
	double rtt_ip_agg[][], rtt_rdp_agg[][];
	int count[], i, max_pts = 300000;
	
	count = new int[50];
	rtt_ip_agg  = new double[50][];
	rtt_rdp_agg = new double[50][];
	for (i = 0 ; i < 50; i++)
	{
	    count[i] = 0;
	    rtt_ip_agg[i]  = new double[max_pts];
	    rtt_rdp_agg[i] = new double[max_pts];
	}
	try {
	    //	    server2 = new NodeId("18.31.0.190:3630");
	    //	    server2 = new NodeId("193.10.133.128:3630");
	    //	    server2 = new NodeId("131.243.254.35:3630");
	    server2 = new NodeId("171.64.64.216:3630");
	} catch( Exception e ) {
	    Carp.die( "Exception creating NodeId for server2" + ": " + e );
	}
	bi2 = (Integer) server_indices.get( server2.address() );
	try {
	    String line = br.readLine();	
	    while (line != null)
	    {
		pos = line.indexOf("from");
		start = pos + 5;
		end = line.indexOf(":", start);
		ip1 = line.substring(start, end);

		pos = line.indexOf("is", end);
		start = pos + 4;
		end = line.indexOf(" ", start);
		time_st  = line.substring(start, end);
		time_int = (Integer.valueOf(time_st)).intValue();
		rtt_tap = (double)(time_int);
		//		System.err.print("Result: " + ip1 + " " + rtt_tap +"\n");
		if ( !ip1.equalsIgnoreCase( (server2.address()).getHostAddress()) )
		{
		    try {
			server1 = new NodeId( ip1 + ":3630" );
		    } catch( Exception e ) {
			Carp.die( "Exception creating NodeId for " + ip1 + ": " + e );
		    }
		    bi1 = (Integer) server_indices.get( server1.address() );
		    if( bi1 == null || bi2 == null) {
			Carp.die( "Could not find servers in the all_pairs_ping file: " + 
				  server1.address() + " and " + server2.address() );
		    }
		    index1 = bi1.intValue();
		    index2 = bi2.intValue();
		    rtt_ip = pings[index1][index2];
		    if (rtt_ip > 0 && rtt_tap < 5001)
		    {
			group = (int)(rtt_ip/10.0);
			if (group < 50 && count[group] < max_pts)
			{
			    rtt_ip_agg[group][count[group]] = rtt_ip;
			    rdp = rtt_tap / rtt_ip;
			    if (rdp < 1.0)
				rdp = 1.0;
			    rtt_rdp_agg[group][count[group]] = rdp;
			    count[group] = count[group] + 1;
			    //			    System.err.print(" " + rtt_ip + " " + (rtt_tap/rtt_ip) + "\n");
			}
			//			else
			//			    System.err.print("Two many data points: " + group + " " + count[group] +"\n");
		    }
		}

		line = br.readLine();
	    }

	    double mean_ip, mean_rdp, std_rdp;
	    for (i = 0; i < 50; i++)
	    {
		mean_ip  = 0.0;
		mean_rdp = 0.0;
		std_rdp  = 0.0;
		if (count[i] > 0)
		{
		    mean_ip  = calc_mean(rtt_ip_agg[i], count[i]);
		    mean_rdp = calc_mean(rtt_rdp_agg[i], count[i]);
		    std_rdp  = calc_std(rtt_rdp_agg[i], count[i]);
		    System.err.print("mean_ip mean_rdp std_rdp: " + mean_ip + " " 
				     + mean_rdp + " " + std_rdp + "\n");
		}
	    }
	} catch( IOException ioe ) {
	    Carp.die( "Could not read line from all_pairs_pings file: " + 
		      ioe );
	}

	System.exit(0);
    }

    public void begin_handle_routing_rdp_results()
    {
        currTest = ROUTING_RDP_RESULTS;

	double pings[][];
	pings = parse_all_pairs_pings("/home/hling/oceanstore/pond/run/x-pl.app");

	BufferedReader br = null;
	try {
	    br = new BufferedReader( new FileReader(
	       "/home/hling/planetlab/stat/origin/routing/routing_rdp_raw_1010.dat") );
	} catch( FileNotFoundException fnfe ) {
	    Carp.die( "Could not open all_pairs_pings_file: " + fnfe );
	}

	NodeId server1=null, server2=null;
	Integer bi1=null, bi2=null;
	int pos, start, end;
	int index1, index2;
	String ip1, ip2, time_st;
	int    time_int, group;
	double rtt_ip, rtt_tap, rdp;
	double rtt_ip_agg[][], rtt_rdp_agg[][];
	int count[], i;
	int max_pt = 300000;
	
	count = new int[50];
	rtt_ip_agg  = new double[50][];
	rtt_rdp_agg = new double[50][];
	for (i = 0 ; i < 50; i++)
	{
	    count[i] = 0;
	    rtt_ip_agg[i]  = new double[max_pt];
	    rtt_rdp_agg[i] = new double[max_pt];
	}
	try {
	    String line = br.readLine();	
	    while (line != null)
	    {
		pos = line.indexOf("between");
		start = pos + 8;
		end = line.indexOf(":", start);
		ip1 = line.substring(start, end);

		pos = line.indexOf("and", end);
		start = pos + 4;
		end = line.indexOf(":", start);
		ip2 = line.substring(start, end);

		pos = line.indexOf("is", end);
		start = pos + 4;
		end = line.indexOf(" ", start);
		time_st  = line.substring(start, end);
		time_int = (Integer.valueOf(time_st)).intValue();

		rtt_tap = (double)(time_int);
		if ( !ip1.equalsIgnoreCase(ip2) )
		{
		    try {
			server1 = new NodeId( ip1 + ":3630" );
			server2 = new NodeId( ip2 + ":3630" );
		    } catch( Exception e ) {
			Carp.die( "Exception creating NodeId for " + ip1 +
				  " and " + ip2 + ": " + e );
		    }
		    //		    bi1 = (Integer) server_indices.get( server1 );
		    //		    bi2 = (Integer) server_indices.get( server2 );
		    bi1 = (Integer) server_indices.get( server1.address() );
		    bi2 = (Integer) server_indices.get( server2.address() );
		    if( bi1 == null || bi2 == null) {
			Carp.die( "Could not find servers in the all_pairs_ping file: " + 
				  server1.address() + " and " + server2.address() );
		    }
		    index1 = bi1.intValue();
		    index2 = bi2.intValue();
		    rtt_ip = pings[index1][index2];
		    if (rtt_ip > 1.0 && rtt_tap < 2001)
		    {
			group = (int)(rtt_ip/10.0);
			if (group < 50 && count[group] < max_pt)
			{
			    rtt_ip_agg[group][count[group]] = rtt_ip;
			    rdp = rtt_tap / rtt_ip;
			    //			    if (rdp < 1.0)
			    //				rdp = 1.0;
			    rtt_rdp_agg[group][count[group]] = rdp;
			    count[group] = count[group] + 1;
			    if (group == 0)
				System.err.print("Group " + group + " " + rtt_ip + " " + 
						 rtt_tap + " " + (rtt_tap/rtt_ip) + "\n");
			}
		    }
		}
		line = br.readLine();
	    }

	    double mean_ip, mean_rdp, std_rdp;
	    for (i = 0; i < 50; i++)
	    {
		mean_ip  = 0.0;
		mean_rdp = 0.0;
		std_rdp  = 0.0;
		if (count[i] > 0)
		{
		    mean_ip  = calc_mean(rtt_ip_agg[i], count[i]);
		    mean_rdp = calc_mean(rtt_rdp_agg[i], count[i]);
		    std_rdp  = calc_std(rtt_rdp_agg[i], count[i]);
		    System.err.print("mean_ip mean_rdp std_rdp: " + mean_ip + " " 
				     + mean_rdp + " " + std_rdp + "\n");
		}
	    }
	} catch( IOException ioe ) {
	    Carp.die( "Could not read line from all_pairs_pings file: " + 
		      ioe );
	}

	System.exit(0);
	/*
	// now we need to find the line corresponding to our pings
	Integer bi = (Integer) server_indices.get( _self_node_id.address() );
	if( bi == null ) {
	    Carp.die( "Could not find ourselves in the all_pairs_ping file" );
	}

	int self_index = bi.intValue();
	
	StringTokenizer pingsST = null;
	try {
	    for( int j = 0; j < self_index; j++ ) {
		br.readLine();
	    }
	    pingsST = new StringTokenizer( br.readLine() );
	} catch( IOException ioe ) {
	    Carp.die( "Could not read line from all_pairs_pings file: " + 
		      ioe );
	}
	*/
    }
    // instead of pinging everyone, just use a file to tell us what our ping
    // times are.
    protected double [][] parse_all_pairs_pings(String pings_file) 
    {
	final String tag = class_tag + ".parse_all_pairs_pings";
	if (DEBUG)
	    Debug.printtagln (tag, "parsing the all_pairs_pings file" );

	BufferedReader br = null;
	try {
	    br = new BufferedReader( new FileReader( pings_file ) );
	} catch( FileNotFoundException fnfe ) {
	    Carp.die( "Could not open all_pairs_pings_file: " + fnfe );
	}

	// first, we read in a line containing the names of all servers
	StringTokenizer serversST = null;
	try {
	    // throw away time
	    br.readLine();
	    // throw away ping count
	    br.readLine();
	    serversST = new StringTokenizer( br.readLine() );
	} catch( IOException ioe ) {
	    Carp.die( "Could not read line from all_pairs_pings file: " + 
		      ioe);
	}

	StringTokenizer pingsST = null;
	String[] servers = new String[ serversST.countTokens() ];
	double pings[][] = new double[servers.length][];
	// need a map to index servers to number
	server_indices = new HashMap();
	for( int i = 0; i < servers.length; i++ ) 
	{
	    servers[i] = serversST.nextToken();
	    NodeId currNode = null;
	    try {
		currNode = new NodeId( servers[i] + ":3630" );
	    } catch( Exception e ) {
		Carp.die( "Exception creating NodeId for " + servers[i] +
			  ": " + e );
	    }
	    server_indices.put( currNode.address(), new Integer(i) );

	    if( DEBUG )
		Debug.printtagln( tag, "creating entry for " + currNode );
	    try 
	    {
		//		br.readLine();
		pingsST = new StringTokenizer( br.readLine() );
	    } catch( IOException ioe ) {
		Carp.die( "Could not read line from all_pairs_pings file: " + 
			  ioe );
	    }

	    // parse out our ping times
	    pings[i] = new double[ pingsST.countTokens() ];
	    for( int j = 0; j < pings.length; j++ ) 
	    {
		// read out av ping time (the one in the middle)
		StringTokenizer slashST = new StringTokenizer( 
					  pingsST.nextToken(), "/" );
		// throw away min time
		slashST.nextToken();
		// now get get av time
		pings[i][j] = Double.parseDouble( slashST.nextToken() );
	    }
	    // they gotta be the same length
	    if( pings[i].length != servers.length ) 
	    {
		Carp.die( "The length of the pings row (" + pings[i].length +
			  ") does not equal the length of the servers row (" +
			  servers.length + ")." );
	    }
	}

	return pings;
    }

/************************************************/
/* calculate mean of a double array of size n	*/
/************************************************/

    double calc_mean(double array[], int n) 
    {
	double sum;
	int i;

	sum = 0.0;
	for (i = 0; i < n; i++)
	    sum += array[i];

	return (sum / n);
    }

/************************************************/
/* calculate standard deviation of a 		*/
/* double array of size n			*/
/************************************************/

    double calc_std(double array[], int n) 
    {
	double sum1, sum2, mean, S2, S;
	int i;

	sum1 = 0.0;
	for (i = 0; i < n; i++)
	    sum1 += array[i];

	mean = sum1 / n;

	sum2 = 0.0;
	for (i = 0; i < n; i++)
	    sum2 += (array[i] - mean) * (array[i] - mean);
	S2 = sum2 / n;                  // biased estimator for sigma^2
	S = Math.sqrt(S2);

	return S;
    }


  /**
   * generates a random securehash based on the time
   * @return SecureHash generated from the current time
   */
  public SecureHash genID() {
      String time = new Long(rand.nextLong()).toString();
      byte [] b = time.getBytes();
      return new SHA1Hash(b);
  }

  public void destroy() throws Exception {
    // TODO
  }

    /* Specified by seda.sandStorm.api.ProfilableIF */
    public int profileSize() {
	return this_sink.size();
    }

    private boolean sim_running;
    private int probing_frequency;
    private HashSet all_failed_paths;
    private HashSet all_failed_ip_pairs;
    private double percentage_failed_links;
    private double percentage_failed_paths;
    private OverlayPath [] total_paths;
    private int resp_nodes;
    private int num_paths;
  private Parser _topo_parser;
  private boolean timeout_already;
  private HashMap waitingTimeouts;
  private long insertion_start_time;
  private long max_insertion_time;
  private NodeId  publish_server;
  private HashMap server_indices;
  private int inserted_nodes;
  private int detached_nodes;
  private int insertion_times;
  private int detach_times;
  private int route_times;
  private int publish_times;
  private int lookup_times;
  private int unpublish_times;
  private int epoch;
  private boolean on_publishing;
  private Random rand;

  //general waiting list, map: requestID->FullTestMsg
  private HashMap waitingRequests; 
  private HashMap waitingMembership;  //for generic membership changes
  private HashSet waitingJoins;
  private HashSet waitingResponses;
  private HashSet failedNodes;   //nodes which failed this test
  private HashSet fixedNodes;  //a set of nodes frequently join in and detach from network.

  //keep the status of all objects in the netowrk.
  private HashMap objects_in_network; //in pair of <node guid, object guid>
  //waiting lists for Dynamic_Route test
  private HashSet waitingRoutes; //list of nodeids

  //waiting lists for publish test
  private HashSet waitingPublishes; //list of object ids
  //waiting lists for objects locating test
  private HashSet waitingLookups; //list of object ids
  //waiting lists for objects detaching test
  private HashSet waitingDetaches; //list of object ids

  private LinkedList tests;  //list of tests waiting to be run
  private int currTest;   //String name of current test
  private int init_nodes;    //counter of how many nodes have announced
  private NodeId [] nodes;   //list of nodes identified by some index
  private ArrayList unattached_nodes;
  private ArrayList attached_nodes;
  private SecureHash [] guids;  //array of all guids from the nodes
  private HashMap nodemap; //map: nodeid-->guid

  int responseCount;
  private SinkIF this_sink;

  // verifier
  private TapestryVerifier verifier;
  private int waiting_verifications;

  private NodeId _self_node_id;
  private Classifier _classifier;
}



