/**
 *  FullTestMember.java
 *
 *  @author     Ben Y. Zhao
 *  @version    $Id: FullTestMember.java,v 1.30 2003/05/22 23:45:26 hling Exp $
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.impl;
import java.util.*;
import java.io.*;
import java.security.*;
import ostore.util.*;
import ostore.network.*;
import ostore.dispatch.*;
import ostore.tapestry.*;
import ostore.tapestry.api.*;
import seda.sandStorm.api.*;
import seda.sandStorm.core.*;
import seda.sandStorm.internal.SinkProxy;
import cryptix.jce.provider.random.DevRandom;
import ostore.network.patchwork.InjectFaultMsg;
import ostore.tapestry.patchwork.RouterFaultMsg;
import seda.sandStorm.api.SinkException;
import ostore.tapestry.channel.ICMPPongMsg;
import ostore.tapestry.channel.ICMPPingMsg;
import ostore.tapestry.patchwork.FaultController;


public class FullTestMember implements EventHandlerIF, ProfilableIF {

    public String class_tag = "FullTestMember";
    private static final boolean DEBUG = false;
    //    private static final boolean DEBUG = true;

    private void dispatch (QueueElementIF item) {
	try {
	    _classifier.dispatch(item);
	} catch (SinkException e) {
	    Carp.out ("Could not dispatch " + item);
	    System.exit (1);
	}
    }

    public void handleEvents(QueueElementIF items[]) 
    throws EventHandlerException {
	for(int i=0; i<items.length; i++)
	    handleEvent(items[i]);
    }

  public void printtagln(String tag, String msg) {
    System.err.print(tag+":"+msg+"\n");
  }

    public void init (ConfigDataIF config) throws Exception {
	Types.registerAllTypes ();

	//	currTest = FullTestDriver.ROUTE_SWITCH;
	currTest = FullTestDriver.REACHABILITY;
	//	currTest = FullTestDriver.MULTICAST_BANDWIDTH;
	//	currTest = FullTestDriver.MISROUTING_LATENCY;

	sim_running = config.getBoolean( "simulator_running" );

	failures = 0;
	recover_time = new int[20];
	recover_seq = new int[20];

	// for deciding when to drop messages
	_fault_controller = new FaultController();


	rand = new Random(1000);
	_classifier = Classifier.getClassifier (
                new NodeId (config.getString ("node_id")));
	
	_self_node_id = new NodeId (config.getString ("node_id"));

	msg_index = 0;
	class_tag = class_tag + ".["+_self_node_id.port()+"]";
	if (DEBUG) ostore.util.Debug.turnOn(class_tag);

	final String tag = class_tag +".init";
	String on = config.getString("power");
	if (on != null && on.equalsIgnoreCase("off")) return;

	String dynamic = config.getString("dynamic_route");
	if (dynamic == null || dynamic.equalsIgnoreCase("static"))
	  dynamic_route = false;
	else dynamic_route = true;
	String testdriver_name = config.getString("testdriver_nodeID");
	if (testdriver_name != null)
	    driveNodeId = new NodeId(testdriver_name);

	String stageName = config.getStage().getName();
	try 
	{
	  if(config.getBoolean("Profile"))
	  {
	    config.getManager().getProfiler().add("ostore.tapestry.impl.FullTestMember", this);
	    SinkIF sink =
	      config.getManager().getStage(stageName).getSink();
	    this_sink = new SinkProxy(sink,
				      config.getManager(),
				      config.getStage().getWrapper());
	  }
	  else
	    this_sink =
	      config.getManager().getStage(stageName).getSink();
	} 
	catch (NoSuchStageException e) 
	{
	  Carp.die (tag + ": could not get own sink.");
	}

	Filter filter;
	String [] event_types = {
	  "seda.sandStorm.api.StagesInitializedSignal",
	  "ostore.tapestry.impl.TapestryReadyMsg",
	  "ostore.tapestry.api.TapestryDetachConfirm",
	  "ostore.tapestry.impl.TimeOutEvent",
	  "ostore.tapestry.impl.RoutingTable$Available"
	};	  

	for (int i=0; i< event_types.length; i++) {
	  filter = new Filter ();
	  if (! filter.requireType (Class.forName(event_types[i]))) {
	    Carp.out ("Could not require type " + event_types[i]);
	    System.exit (1);
	  }
	  _classifier.subscribe (filter, this_sink);
	}

	String [] msg_types = {
	  "ostore.tapestry.impl.FullTestMsg",
	  "ostore.tapestry.impl.TestReadyMsg",
	  "ostore.tapestry.impl.TestNetworkMsg",
	  "ostore.tapestry.impl.TestLocateMsg",
	  "ostore.tapestry.impl.TestFoundMsg",
	  "ostore.tapestry.api.TapestryLocateFailure",
	  "ostore.tapestry.impl.TestDeleteMsg",
	  "ostore.tapestry.channel.ICMPPongMsg",
	  "ostore.tapestry.patchwork.RouterFaultMsg"
	};

	for (int i=0; i< msg_types.length; i++) {
	  filter = new Filter ();
	  if (! filter.requireType (Class.forName(msg_types[i])))
	    Carp.die ("Could not require type " + msg_types[i]);
	  if (! filter.requireValue("inbound", new Boolean(true)))
	    Carp.die ("Could not require inbound = true for "+
		      msg_types[i]);
	  _classifier.subscribe (filter, this_sink);
	}

	objs = new HashSet();
	paths = new HashMap();
	waitingActions = new HashSet();
	waitingLookups = new HashMap();
	waitingRoutes = new HashMap();
	waitingTimeouts = new HashMap();
	System.out.println("FullTestMember Tapestry stage initialized!!");
    }

  public void handle_fulltest(FullTestMsg msg) {
    final String tag = class_tag +".handle_fulltest";
    if (DEBUG)
	Debug.printtagln(tag, "Get FullTestMsg: " + msg);
    //    currTest = msg.test;
    if (msg.type == msg.EXIT_REQUEST) {
	//      if (DEBUG)
	Debug.printtagln(tag, "Got EXIT_REQUEST...");
	System.gc(); // Clean up objects
	try{Thread.sleep(5000);}catch(InterruptedException e){}
	System.exit(1);
    }
    if (msg.type == msg.INFO_REQUEST) {
      dispatch(new FullTestRespMsg(msg.peer, true, msg.type, msg.requestID,
				   _self_guid, "INFO Message response"));
      
    } 
    else if (msg.type == msg.ROUTING_PATH_RESULTS)
    {
	int i, num;
	OverlayPath tap_path, good_paths[];
	Iterator it;

	Debug.printtagln(tag, "Return a set of paths which have nodes at least " + 
			 msg.hop+1);
	num = 0;
	Collection all_paths = paths.values();
	for ( it = all_paths.iterator(); it.hasNext(); )
	{
	    tap_path = (OverlayPath) it.next();
	    if (tap_path.flag == tap_path.COMPLETE)
	    {
		if (tap_path.nodeids.size() == (msg.hop+1))
		    num++;
	    }
	}

	if (num > 0)
	{
	    good_paths = new OverlayPath[num];
	    i = 0;
	    for ( it = all_paths.iterator(); it.hasNext(); )
	    {
		tap_path = (OverlayPath) it.next();
		if (tap_path.flag == tap_path.COMPLETE)
		{
		    if (tap_path.nodeids.size() >= (msg.hop+1))
		    {
			good_paths[i] = tap_path;
			i++;
			if (i >= num)
			    break;
		    }
		}
	    }
	}
	else
	    good_paths = null;
	dispatch(new FullTestRespMsg(msg.peer, true, msg.type, msg.requestID,
				     _self_guid, "A set of good paths", good_paths));

    }
    else if (msg.type == msg.ROUTING_PATH_REQUEST)
    {
	/*	nodes_in_path = 0;
	waitingPaths.put(msg.requestID, msg);
	Object token = 
	    _classifier.dispatch_later(new TimeOutEvent(msg.peer, msg.requestID,
							msg.type), (long) 5000);
	waitingTimeouts.put(msg.requestID, token);
	*/

	double  id;
	String  id_string;
	SecureHash msg_id;
	if (DEBUG)
	    Debug.printtagln(tag, "Send ICMPPingMsg from " + _self_guid +
			     " to " + msg.guids[1]);
	
	id = java.lang.Math.random();
	id_string = _self_guid.toString() + msg.guids[1].toString() + 
	            Double.toString(id);
	msg_id = new SHA1Hash(id_string);
	dispatch(new ICMPPingMsg(_self_node_id, msg_id, _self_guid, msg.guids[1]));
	
	OverlayPath tap_path = new OverlayPath(_self_guid, msg.guids[1]);
	tap_path.start =  Clock.current_date(_self_node_id).getTime();
	tap_path.id = msg_id;
	tap_path.nodeids.add(_self_node_id);
	tap_path.guids.add(_self_guid);
	tap_path.rtts.add(new Long(0));
	paths.put(msg_id, tap_path);
    }
    else if (msg.type == msg.INJECT_FAULT)
    {
	int loss_rate = msg.value;

	try {
	    long time = Clock.current_date(_self_node_id).getTime();
	    NodeId middle = (NodeId) (msg.gateway).clone();
	    Debug.printtagln(tag, "Get an INJECT_FAULT request for node: " +
			     middle + " with loss = " + loss_rate + 
			     " at time " + time);
	    dispatch( new RouterFaultMsg( middle, loss_rate, true, true) );
	    dispatch( new InjectFaultMsg( middle, loss_rate, true, true ) );
	}
	catch (CloneNotSupportedException e) 
	{
	    Debug.printtagln(tag, ".dispatch: could not dispatch falut msg!");
	}
    }
    else if (msg.type == msg.TRACE_DATA)
    {
	times = msg.time;
	curr_time_index = 0;
	my_gateway = msg.gateway;
	if (my_gateway == null)
	{
	    Debug.printtagln(tag, "No GATEWAY available for me, die...");
	    System.exit(1);
	}
	//Doesn't send resp to Drive node.

	if( DEBUG ) {
	    System.out.println( "Got times: " );
	    for( int j = 0; j < times.length; j++ ) {
		System.out.print( times[j] + " " );
	    }
	    System.out.println( "" );
	}

    }
    else if (msg.type == msg.TRACE_START)
    {

	if( curr_time_index == times.length ) {
	    System.out.println( "Reached the end of my times.  Stopping." );
	    return;
	}

	long later_time = 0;
	// if it's the first integration, just use the first time
	if( curr_time_index == 0 ) {
	    later_time = times[0];
	} else {
	    later_time = times[curr_time_index] - times[curr_time_index-1];
	}

	if( curr_time_index % 2 == 0 ) {

	    //Tell itself to join
	    if( DEBUG ) {
		System.out.println( "Dispatching integration command " +
				    later_time + " ms in the future" );
	    }
	    _classifier.dispatch_later(new TapestryIntegrateMsg(my_gateway),
				       later_time);
	} else {

	    // tell itself to leave
	    if( DEBUG ) {
		System.out.println( "Dispatching detachment command " +
				    later_time + " ms in the future" );
	    }
	    _classifier.dispatch_later(new TapestryNodeDetachMsg(),
				       later_time);
	}

	// send a TRACE_START message to yourself for later
	FullTestMsg ftm = new FullTestMsg( _self_node_id,
					   FullTestMsg.TRACE_START, 
					   msg.requestID,
					   null, null );
	_classifier.dispatch_later(ftm, later_time);
	curr_time_index++;

    }
    else if (msg.type == msg.CLEAR_BANDWIDTH_COUNTS)
    {
	String user_data;

	user_data = msg.comment;
	if (DEBUG)
	    Debug.printtagln(tag, "Get CLEAR_BANDWIDTH_COUNTS for " +
			     user_data + " ...");

	if (sim_running)
	{
	    soss.network.Network.instance().print_byte_counts(user_data);
	    soss.network.Network.instance().clear_byte_counts();
	}
	else
	    dispatch(new ClearBandwidthCountsReq(user_data));
    }
    else if (msg.type == msg.PUBLISH_REQUEST) {
      for (int i=0; i<msg.guids.length; i++) {
	  /*
	dispatch(new TapestryPublishMsg(msg.guids[i], null));
	objs.add(msg.guids[i]);
	  */
	  //Modified by Jemery and Ling for publish optimization.
	  TapestryPublishMsg pub_msg = new TapestryPublishMsg(msg.guids[i], null);
	  //	  Debug.printtagln(tag, "Publishing " + i + "th objects " + msg.guids[i] + " ...");
	  //	  pub_msg.setLocalOptPublishBackup();
	  dispatch(pub_msg);

	  objs.add(msg.guids[i]);
      }
      dispatch(new FullTestRespMsg(msg.peer, true, msg.type, msg.requestID,
				   _self_guid, "PUBLISH Messages sent"));

    } else if (msg.type == msg.UNPUBLISH_REQUEST) {
      HashSet notFound = null;
      for (int i=0; i<msg.guids.length; i++) {
	if (objs.contains(msg.guids[i])) 
	  dispatch(new TapestryUnpublishMsg(msg.guids[i], null));
	else {
	  if (notFound == null)
	    notFound = new HashSet();
	  notFound.add(msg.guids[i]);
	}
      }
      if (notFound == null)
	dispatch(new FullTestRespMsg(msg.peer, true, msg.type, msg.requestID,
				     _self_guid, "All UNPUBLISHES sent"));
      else 
	dispatch(new FullTestRespMsg(msg.peer, false, msg.type,
				     msg.requestID, _self_guid,
				     "UNPUBLISHES FAILED: "+notFound));
      
    } 
    else if (msg.type == msg.LOOKUP_REQUEST) 
    {
	//	Modified by Ling Huang for lookup rdp.
	/*
      for (int i=0; i<msg.guids.length; i++) {
	dispatch(new TestLocateMsg(msg.guids[i], msg.requestID));
      }
	*/
      Debug.printtagln(tag, "For Lookup at time " + lookup_number + "I get " +
		       msg.guids.length + " guids");
      msg.time = new long[msg.guids.length];
      next_lookup = 0;
      int delay;
      while (next_lookup < msg.guids.length)
      {
	  delay = next_lookup*300;
	  msg.time[next_lookup] = Clock.current_date(_self_node_id).getTime() + delay;
	  _classifier.dispatch_later(new TestLocateMsg(_self_node_id, msg.guids[next_lookup], 
				     msg.requestID), (long)delay);
	  next_lookup++;
      }
      waitingLookups.put(msg.requestID, msg);
      Object token = 
	_classifier.dispatch_later(new TimeOutEvent(msg.peer, msg.requestID,
						    msg.type), (long) 500000);
      waitingTimeouts.put(msg.requestID, token);
    } 
    else if (msg.type == msg.ROUTE_REQUEST)
    {
	// For route switch.
	if (currTest == FullTestDriver.ROUTE_SWITCH)
	{
	    long time =  Clock.current_date(_self_node_id).getTime();
	    Debug.printtagln(tag, "Faults end at time = " + time);
	    FullTestRespMsg respmsg = new FullTestRespMsg(driveNodeId, true, msg.type, msg.requestID,
							  _self_guid, "Faults recovery time = ");
	    respmsg.setTime(time - faults_begin_time);
	    dispatch(respmsg);

	    /*
	    prev_seq = -1;
	    interval = msg.value;
	    for (int i=0; i<msg_num; i++) 
	    {
		long time = System.currentTimeMillis();
		Debug.printtagln(tag, "At time = " + time + ", send " + 
				 i + "th TestReadyMsg to "+msg.guids[0]);
		_classifier.dispatch_later(new TestReadyMsg(msg.guids[0], 
					   _self_guid, msg.requestID, i, 
					   true), i*interval);
	    }
	    */
	}
	else if (currTest == FullTestDriver.REACHABILITY ||
		 currTest == FullTestDriver.MULTICAST_BANDWIDTH)
	{
	    //For reachability.
	    //For one-way measurement.
	    TestReadyMsg test_msg;
	    TestNetworkMsg test_net_msg;

	    Debug.printtagln(tag, "Routing request for failure rate = " + 
			     msg.failure);
	    msg.time = new long[msg.guids.length];
	    for (int i=0; i<msg.guids.length; i++) 
	    {
		msg.time[i] = Clock.current_date(_self_node_id).getTime();

		test_msg = new TestReadyMsg(msg.guids[i], 
			  _self_guid, msg.requestID, msg.failure, 
				true);
		//		test_msg.comment = msg.comment;
		dispatch(test_msg);

		test_net_msg = new TestNetworkMsg(msg.nodeids[i], 
						  msg.time[i], msg.failure);
		//		test_net_msg.comment = msg.comment;
		dispatch(test_net_msg);
	    }
	}
	else if (currTest == FullTestDriver.MISROUTING_LATENCY ||
		 currTest == FullTestDriver.STATIC_ROUTE ||
		 currTest == FullTestDriver.DYNAMIC_ROUTE)
	{
	    //This is Modified by Ling Huang for routing rdp and misrouting latency.
	    //For round trip time measurement.
	    test_route = msg;
	    test_route.time = new long[msg.guids.length];
	    next_route = 0;
	    test_route.time[next_route] =  Clock.current_date(_self_node_id).getTime();
	    dispatch(new TestReadyMsg(test_route.guids[next_route], _self_guid, 
				test_route.requestID, true));
	    next_route++;

	    waitingRoutes.put(msg.requestID, msg);
	    Object token = 
		_classifier.dispatch_later(new TimeOutEvent(msg.peer, msg.requestID,
					   msg.type), (long) 2000000);
	}
    }
    else if (msg.type == msg.INTEGRATE_REQUEST) {
	if (DEBUG)
	{
	    Debug.printtagln(tag, "FullTestMember Processing integrate, gateway= "+
			     msg.gateway + " after time " + msg.msToWait);
	}
      if (msg.msToWait == 0) {
	waitingAttach = msg;
	dispatch(new TapestryIntegrateMsg(msg.gateway));
	return;
      }
      if (waitingActions.contains(msg)) {
	waitingActions.remove(msg);
	waitingAttach = msg;
	dispatch(new TapestryIntegrateMsg(msg.gateway));
      } else {
	waitingActions.add(msg);
	_classifier.dispatch_later(msg, msg.msToWait);
      }

    } else if (msg.type == msg.DETACH_REQUEST) {
      if (msg.msToWait == 0) {
	waitingDetach = msg;
	dispatch(new TapestryNodeDetachMsg());
	return;
      }
      if (waitingActions.contains(msg)) {
	waitingActions.remove(msg);
	waitingDetach = msg;
	dispatch(new TapestryNodeDetachMsg());
      } else {
	waitingActions.add(msg);
	_classifier.dispatch_later(msg, msg.msToWait);
      }

    } else if (msg.type == msg.DETACH_INTEGRATE_REQUEST) {

    } else if (msg.type == msg.VERIFY_REQUEST) {

	// send back a VerifyMsg
	VerifyRespMsg vm = new VerifyRespMsg( msg.peer, true, msg.type,
					      msg.requestID, _self_guid,
					      "No output", _routing_tbl,
					      _revpointers );
	dispatch( vm );

    } else {
      System.err.println("FullTestMember.HandleFullTestMsg"+
			 "Unknown FullTestMsg Type!");
    }
  }
  
  public void handle_testlocate(TestLocateMsg msg) {
    if (DEBUG) printtagln("FullTestMember.handle_testlocate",
			  "Got TestLocateMsg from "+ msg.peer+" for guid: "+
			  printGuid(msg.guid));
    //now send back a routemsg to the originator of the locatemsg
    dispatch(new TestFoundMsg(msg.querysrc, msg.guid, msg.id));
  }

  public void handle_icmp_pong_msg(ICMPPongMsg msg) 
  {
      if (DEBUG)
	  System.out.println("FullTestMember gets an ICMPPongMsg: " + msg);
      
      long curr =  Clock.current_date(_self_node_id).getTime();
      OverlayPath tap_path = (OverlayPath)paths.get(msg.msg_id);
      if (tap_path.flag == tap_path.INCOMPLETE)
      {
	  tap_path.nodeids.add(msg.nodeid);
	  tap_path.guids.add(msg.guid);
	  tap_path.rtts.add(new Long(curr - tap_path.start));
	  if ( (msg.guid).equals(tap_path.dst) )
	  {
	      tap_path.flag = tap_path.COMPLETE;
	      if (DEBUG)
		  System.out.println(tap_path);
	  }
      }
  }
  public void handle_testfound(TestFoundMsg msg) {
    final String tag = class_tag +".handle_testready";
    if (DEBUG)
	System.out.println("Got TestFoundMsg for obj "+msg.objGuid+" at node"+
			   msg.peer);
    FullTestMsg ftm = (FullTestMsg) waitingLookups.get(msg.id);
    if (ftm == null) {
      //means this msg was removed by a TapestryLocateFailure, ignore
      return;
    }

    int i;
    for (i=0; i<ftm.guids.length; i++) 
    {
      if (ftm.guids[i] != null && msg.objGuid.equals(ftm.guids[i]))
      {
	  long rtt =  Clock.current_date(_self_node_id).getTime() -  ftm.time[i];
	  String lookup_string = "Lookup";
	  if( ftm.repeat_lookups ) {
	      lookup_string += lookup_number;
	  }
	  Debug.printtagln(tag, i + "th Guid: " + lookup_string + " from " + _self_node_id + 
			   " for " + ftm.guids[i] + " at " + msg.peer + " is: " + rtt + " ms.");
	  ftm.guids[i] = null;
      }
    }

    if (next_lookup < ftm.guids.length)
    {
	ftm.time[next_lookup] =  Clock.current_date(_self_node_id).getTime();
	dispatch(new TestLocateMsg(_self_node_id, ftm.guids[next_lookup], 
				   ftm.requestID));
	next_lookup++;
    }

    boolean done = true;
    for (i=0; i<ftm.guids.length; i++) {
      if (ftm.guids[i] !=null) {
	done = false; 
	break;
      }
    }
    if (done) {
      Object token = waitingTimeouts.get(msg.id);
      if (token != null)
	_classifier.cancel(token);
      else System.err.println("FullTestMember.HandleTestFoundMsg: "+
			      "No token found for Locate timeout");
      waitingLookups.remove(msg.id);
      if( !ftm.repeat_lookups ) {
	  dispatch(new FullTestRespMsg(ftm.peer, true, ftm.type, ftm.requestID,
				       _self_guid, "All objects located."));
      }
      else
      {
	  /*
	  // now, if we're going to be repeating, duplicate the lookup message
	  // and send it to yourself again
	  SecureHash id = genID();
	  FullTestMsg new_ftm = new FullTestMsg(_self_node_id,
						FullTestMsg.LOOKUP_REQUEST, 
						id, 0, 
						repeater_guids);
	  new_ftm.repeat_lookups = true;
	  _classifier.dispatch_later( new_ftm, REPEAT_LOOKUPS_INTERVAL );
	  */
      }
    }
  }
  public void handle_testready(TestReadyMsg msg) 
  {
      final String tag = class_tag +".handle_testready";
      if (DEBUG)
	  Debug.printtagln(tag, "Got TestReadyMsg :"+msg);

      if (currTest == FullTestDriver.ROUTE_SWITCH)
      { 
	  //For route switch.
	  long time =  Clock.current_date(_self_node_id).getTime();
	  Debug.printtagln(tag, "At time = " + time + ", got " + 
			   msg.seq + "th TestReadyMsg:"+msg);
	  int diff;
	  diff = msg.seq - prev_seq -1;
	  prev_seq = msg.seq;
	  if (diff > 0)
	  {
	      Debug.printtagln(tag, "==========================" + 
			       "================================");
	      Debug.printtagln(tag, "The recover time of failures is: "+ 
			       diff*interval);
	      Debug.printtagln(tag, "==========================" + 

			       "================================");
	      recover_time[failures] = diff*interval;
	      recover_seq[failures]  = msg.seq;
    	      failures++;
	  }

	  if (msg.seq >= msg_num -1)
	  {
	      int i;
	      Debug.printtagln(tag, "==========================" + 
			       "================================");
	      for (i = 0; i < failures; i++)
	      {
		  Debug.printtagln(tag, "The recover time of failures at " + 
				   recover_seq[i] + 
				   " is: "+ recover_time[i]);
	      }

	      Debug.printtagln(tag, "==========================" + 
			       "================================");
	  }

      }
      else if (currTest == FullTestDriver.REACHABILITY ||
	       currTest == FullTestDriver.MULTICAST_BANDWIDTH)
      {
	  //For Rechability.
	  String comment = " ";
	  if (msg.comment != null)
	      comment = msg.comment;
	  Debug.printtagln(tag, "TestReadyMsg for TAPESTRY " + 
			   comment +" with link failure " 
			   + msg.failure + "%.");
      }
      else if (currTest == FullTestDriver.STATIC_ROUTE ||
	       currTest == FullTestDriver.MISROUTING_LATENCY ||
	       currTest == FullTestDriver.DYNAMIC_ROUTE)
      {
	  //For routing rdp and misrouting latency.
	  if (msg.request)
	  {
	      Debug.printtagln(tag, "Dispatch a responce to " + msg.src +
			       " from " + _self_guid);
	      dispatch(new TestReadyMsg(msg.src, _self_guid, msg.id, false));
	  }
	  else 
	  {
	      if (DEBUG)
		  Debug.printtagln(tag, "This is a routing response from "+msg.peer);
	      FullTestMsg ftm = (FullTestMsg) waitingRoutes.get(msg.id);
	      if (ftm == null) 
	      {
		  //means this msg was removed by a TapestryLocateFailure, ignore
		  if (DEBUG)
		      Debug.printtagln(tag, "We don't find any FullTestMsg indexed by this msg.");
		  return;
	      }

	      if (next_route < test_route.guids.length)
	      {
		  test_route.time[next_route] = Clock.current_date(_self_node_id).getTime();
		  dispatch(new TestReadyMsg(test_route.guids[next_route], _self_guid, test_route.requestID,
					    true));
		  next_route++;
	      }

	      //go through and zero out the entry matching the response we got
	      boolean zeroout = false;
	      if (DEBUG)
		  Debug.printtagln(tag, "We find a FullTestMsg indexed by this msg: "+ ftm);
	      for (int i=0; i<ftm.guids.length; i++) 
	      {
		  if (ftm.guids[i] != null && msg.src.equals(ftm.guids[i]))
		  {
		      if (DEBUG)
			  Debug.printtagln(tag, "This msg zero out an entry in ftm.guids:  "+ 
					   ftm.guids[i]);

		      long rtt =  Clock.current_date(_self_node_id).getTime() -  ftm.time[i];
		      Debug.printtagln(tag, ftm.comment + "RTT between " + _self_node_id + 
				       " and " + ftm.nodeids[i] +
				       " is: " + rtt + " ms.");
		      zeroout = true;
		      ftm.guids[i] = null;
		  }
	      }
	      if (DEBUG && !zeroout)
		  Debug.printtagln(tag, "This msg doesn't zero out any entry in ftm.guids. " +
				   "There must be something wrong!\n"+msg);
	      boolean done = true;
	      //are we all done?
	      int i;
	      for (i=0; i<ftm.guids.length; i++) {
		  if (ftm.guids[i] !=null) {
		      done = false; 
		      break;
		  }
	      }
	      if (done) {
		  Object token = waitingTimeouts.get(msg.id);
		  if (token != null)
		      //if done, then try to cancel the timeout event
		      _classifier.cancel(token);
		  else System.err.println("FullTestMember.handle_testready: "+
					  "No token found for routemsg timeout");
		  waitingRoutes.remove(msg.id);
		  dispatch(new FullTestRespMsg(ftm.peer, true, ftm.type, ftm.requestID,
					       _self_guid, "All route messages acked."));
	      }
	      else {
		  if (DEBUG)
		      Debug.printtagln(tag, "We still don't get response from some des, for example " +
				       ftm.guids[i]);
	      }
	  }
      }
  }
 
  public void handle_test_network(TestNetworkMsg msg) 
  {
      final String tag = class_tag +".handle_test_network";

      long time =  Clock.current_date(_self_node_id).getTime();
      //For route switch.
      //      Debug.printtagln(tag, "At time = " + time + ", got " + 
      //	       msg.seq + "th TestReadyMsg:"+msg);
      //For Rechability.

      Debug.printtagln(tag, "TestNetworkMsg for IP with link failure " + 
		       msg.failure + 
		       "%.");
  } 

  public void handle_timeoutevent(TimeOutEvent toe) {
    if (toe.requestType == FullTestMsg.LOOKUP_REQUEST) {
      FullTestMsg msg = (FullTestMsg) waitingLookups.get(toe.requestID);
      boolean done = true;
      for (int i=0; i<msg.guids.length; i++) {
	if (msg.guids[i] != null) {
	  done = false;
	  break;
	}
      }
      waitingLookups.remove(toe.requestID);
      if (done)
	dispatch(new FullTestRespMsg(msg.peer, true, msg.type, msg.requestID,
				     _self_guid, "All objects Found!"));
      else dispatch(new FullTestRespMsg(msg.peer, false, msg.type, 
					msg.requestID, _self_guid, 
					"Timeout exceeded on lookup"));
    } else {
      System.err.println("FullTestMember.HandleTimeOutEvent"+
			 "Unrecognized TimeOutEvent type");
    }
  }

  public void handle_tapestry_locatefailure(TapestryLocateFailure msg) {
    final String tag = class_tag +".handle_tapestry_locatefailure";
    if (DEBUG)
	Debug.printtagln(tag, "Got LocateFailureMsg for obj "+
			 msg.obj_guid+" at node"+ msg.peer);
    FullTestMsg ftm = (FullTestMsg) waitingLookups.get(msg.id);
    if (ftm == null) {
	//this should never happen
	//	Carp.die("failure on unexpected object");
      return;
    }
    /*
    int i = next_lookup - 1;
    long rtt = System.currentTimeMillis() -  ftm.time[i];
    String lookup_string = "Lookup";
    if( ftm.repeat_lookups ) {
	lookup_string += lookup_number;
    }
    ftm.guids[i] = null;

    if (next_lookup < ftm.guids.length)
    {
	ftm.time[next_lookup] = System.currentTimeMillis();
	dispatch(new TestLocateMsg(_self_node_id, ftm.guids[next_lookup], 
				   ftm.requestID));
	next_lookup++;
    }
*/

    int i;
    for (i=0; i<ftm.guids.length; i++) 
    {
      if (ftm.guids[i] != null && msg.obj_guid.equals(ftm.guids[i]))
      {
	  long rtt =  Clock.current_date(_self_node_id).getTime() -  ftm.time[i];
	  String lookup_string = "Lookup";
	  if( ftm.repeat_lookups ) {
	      lookup_string += lookup_number;
	  }
	  Debug.printtagln(tag, lookup_string + " failed from " + _self_node_id + 
			   " for " + ftm.guids[i] + " at " + msg.peer);
	  ftm.guids[i] = null;
      }
    }

    boolean done = true;
    for (i=0; i<ftm.guids.length; i++) {
      if (ftm.guids[i] !=null) {
	done = false; 
	break;
      }
    }
    if (done) {
      Object token = waitingTimeouts.get(msg.id);
      if (token != null)
	_classifier.cancel(token);
      else System.err.println("FullTestMember.HandleTestFoundMsg: "+
			      "No token found for Locate timeout");
      waitingLookups.remove(msg.id);
      // if this is not a repeater, send a response
      if( !ftm.repeat_lookups ) {
	  dispatch(new FullTestRespMsg(ftm.peer, true, ftm.type, ftm.requestID,
				       _self_guid, "Some objects fails."));
      }
      else
      {
	  /*
	  // now, if we're going to be repeating, duplicate the lookup message
	  // and send it to yourself again
	  SecureHash id = genID();
	  FullTestMsg new_ftm = new FullTestMsg(_self_node_id,
						FullTestMsg.LOOKUP_REQUEST, 
						id, 0, 
						repeater_guids);
	  new_ftm.repeat_lookups = true;
	  _classifier.dispatch_later( new_ftm, REPEAT_LOOKUPS_INTERVAL );
	  */
      }
    }

    /*    final String tag = class_tag +".handle_tapestry_locatefailure";
    if (DEBUG)
	Debug.printtagln(tag, "Got LocateFailureMsg for obj "+
			 msg.obj_guid+" at node"+ msg.peer);
    FullTestMsg ftm = (FullTestMsg) waitingLookups.get(msg.id);
    if (ftm == null) {
	//this should never happen
	Carp.die("failure on unexpected object");
      return;
    }

    int i = next_lookup - 1;
    long rtt = System.currentTimeMillis() -  ftm.time[i];
    String lookup_string = "Lookup";
    if( ftm.repeat_lookups ) {
	lookup_string += lookup_number;
    }
    Debug.printtagln(tag, lookup_string + " failed from " + _self_node_id + 
		     " for " + ftm.guids[i] + " at " + msg.peer);
    ftm.guids[i] = null;

    if (next_lookup < ftm.guids.length)
    {
	ftm.time[next_lookup] = System.currentTimeMillis();
	dispatch(new TestLocateMsg(_self_node_id, ftm.guids[next_lookup], 
				   ftm.requestID));
	next_lookup++;
    }

    boolean done = true;
    for (i=0; i<ftm.guids.length; i++) {
      if (ftm.guids[i] !=null) {
	done = false; 
	break;
      }
    }
    if (done) {
      Object token = waitingTimeouts.get(msg.id);
      if (token != null)
	_classifier.cancel(token);
      else System.err.println("FullTestMember.HandleTestFoundMsg: "+
			      "No token found for Locate timeout");
      waitingLookups.remove(msg.id);
      // if this is not a repeater, send a response
      if( !ftm.repeat_lookups ) {
	  dispatch(new FullTestRespMsg(ftm.peer, true, ftm.type, ftm.requestID,
				       _self_guid, "Some objects fails."));
      }
    }
*/    
  }

  public void handle_tapestry_detachconfirm() {
    if (DEBUG)
      System.err.println("Node detached!!");
    if (waitingDetach != null)
    {
	dispatch(new FullTestRespMsg(waitingDetach.peer, true, waitingDetach.type,
				     waitingDetach.requestID, _self_guid,
				     "Successfully detached from Tapestry mesh."));
	waitingDetach = null;
    }
  }

  public void handleEvent(QueueElementIF item) 
    throws EventHandlerException {

    final String tag = class_tag+".handleEvent";
    if (DEBUG)
	Debug.printtagln(tag, "Handing event :"+item);

    if (item instanceof StagesInitializedSignal) {

      if (driveNodeId != null && dynamic_route) {
	//send msg to fulltestdriver
	FullTestMemberMsg member_msg = 
	  new FullTestMemberMsg(driveNodeId, 
				_self_guid, FullTestMemberMsg.DYNAMIC);
	dispatch(member_msg);
	if (DEBUG)
	    System.err.println("Send msg to driver: nodeid= " + _self_node_id +
			       "driveNodeid=, "+driveNodeId+" , dynamic_route="+
			       dynamic_route);
      } 
    } else if (item instanceof RoutingTable.Available) {

	RoutingTable.Available rta = (RoutingTable.Available) item;
	_routing_tbl = rta.table;
	_revpointers = rta.revpointers;

    } else if (item instanceof RouterFaultMsg) {
	long time =  Clock.current_date(_self_node_id).getTime();
	//	Debug.printtagln (tag, "Faults begin at time = " + time);
	faults_begin_time = time;

	_fault_controller.updateFaults( (RouterFaultMsg) item );
    } else if (item instanceof ICMPPongMsg) {
      handle_icmp_pong_msg((ICMPPongMsg) item);
    } else if (item instanceof FullTestMsg) {
      FullTestMsg msg = (FullTestMsg) item;
      handle_fulltest(msg);
    } else if (item instanceof TestLocateMsg) {
      TestLocateMsg msg = (TestLocateMsg) item;
      handle_testlocate(msg);

    } else if (item instanceof TestFoundMsg) {
      TestFoundMsg msg = (TestFoundMsg) item;
      handle_testfound(msg);
    } else if (item instanceof TestReadyMsg) {
      TestReadyMsg msg = (TestReadyMsg) item;
      handle_testready(msg);

    } 
    else if (item instanceof TestNetworkMsg) 
    {
	// should we drop this message?
	if( _fault_controller.dropMessage(((NetworkMessage)item).peer) ) 
	    return;
	TestNetworkMsg msg = (TestNetworkMsg) item;
	handle_test_network(msg);
    } 
    else if (item instanceof TimeOutEvent) {
      TimeOutEvent toe = (TimeOutEvent) item;
      handle_timeoutevent(toe);

    } else if (item instanceof TapestryReadyMsg) {
      _self_guid = ((TapestryReadyMsg)item).node_guid;
      /*
      if (driveNodeId != null && dynamic_route) {
	//send msg to fulltestdriver
	FullTestMemberMsg member_msg = 
	  new FullTestMemberMsg(driveNodeId, 
				_self_guid, FullTestMemberMsg.DYNAMIC);
	dispatch(member_msg);
	if (DEBUG)
	    System.err.println("Send msg to driver: nodeid= " + _self_node_id +
			       "driveNodeid=, "+driveNodeId+" , dynamic_route="+
			       dynamic_route);
      
      }
      */
      //send msg to fulltestdriver for statically configured nodes
	System.err.println("waitingAttach="+waitingAttach);
      if (waitingAttach != null) {
	FullTestRespMsg ftrm = new FullTestRespMsg(waitingAttach.peer, true,
						   waitingAttach.type, 
						   waitingAttach.requestID, _self_guid,
						   "Integration successful.");
	ftrm.setTime(((TapestryReadyMsg) item).integrateTime);
	dispatch(ftrm);
	System.err.println("Just sent FullTestResp to FullTestDriver");
	waitingAttach = null;
	return;
      }
      if (!dynamic_route && driveNodeId != null) 
      {
	FullTestMemberMsg member_msg = 
	  new FullTestMemberMsg(driveNodeId, 
				_self_guid, FullTestMemberMsg.STATIC);
	dispatch(member_msg);
	if (DEBUG)
	    System.err.println("Send msg to driver: nodeid= " + _self_node_id +
			       "driveNodeid=, "+driveNodeId+" , dynamic_route="+
			       dynamic_route);
      }

    } else if (item instanceof TapestryLocateFailure) {
      TapestryLocateFailure msg = (TapestryLocateFailure) item;
      handle_tapestry_locatefailure(msg);

    } else if (item instanceof TapestryDetachConfirm) {
      handle_tapestry_detachconfirm();

    } else {
      Carp.die ("ostore.tapestry.impl.Test: Received unknown "
		+ "QueueElementIF: " + item);
    }
  }

  /**
   * generates a random securehash based on the time
   * @return SecureHash generated from the current time
   */
  public SecureHash genID() {
      String time = new Long(rand.nextLong()).toString();
      byte [] b = time.getBytes();
      return new SHA1Hash(b);
  }

  
  public void destroy() throws Exception {
    // TODO
  }

    //  methods for ProfilableIF
    /* Specified by seda.sandStorm.api.ProfilableIF */
    public int profileSize()
    {
	return this_sink.size();
    }
    //  end methods for ProfilableIF

    public static String printGuid (SecureHash guid) {
        String out="";
	int [] gdigits = RoutingTable.digits(guid);
	for (int i = 0; i < gdigits.length; ++i) {
	    out += Integer.toHexString (gdigits [i]);
}   
	return out;
    }

  public static final int TRACE_DRIVEN= 0;
  public static final int CONTROLLER_DRIVEN= 0;

    private FaultController _fault_controller;

    public  long faults_begin_time;
    public  long faults_end_time;
    public  int failures = 0;
    public  int [] recover_time;
    public  int [] recover_seq;
    public  int msg_num = 3000;
    public  int interval = 20;
    public  int prev_seq = -1;
    public  boolean sim_running;
  public  int currTest;
  public  int nodes_in_path;
  private Random rand;
  public  NodeId my_gateway; 
  private int curr_time_index;
  private int msg_index;
  private long curr_time;
  private long prev_time;
    private long happen_time;
    private long plan_time;
  private long times[];
  private int  events[];
  private FullTestMsg test_route;
  private int next_route;
  private int next_lookup;
  private boolean dynamic_route;
  private SecureHash _self_guid;
  private HashSet waitingActions;

  private HashMap paths;
  //generic set of messages which are waiting to be executed
  private HashMap waitingLookups;
  //lookups which are waiting
  private FullTestMsg waitingDetach;
  private FullTestMsg waitingAttach;
  private HashMap waitingRoutes;
  private HashMap waitingTimeouts;
  private HashSet objs;
  private int epoch;
  private NodeId driveNodeId;
  
  private SinkIF this_sink;

  private NodeId _self_node_id;
  private Classifier _classifier;

  private RoutingTable _routing_tbl;
  private HashMap _revpointers;

  private static final long REPEAT_LOOKUPS_INTERVAL = 400000;
    //  private static final long REPEAT_LOOKUPS_INTERVAL = 30000;
  private SecureHash repeater_guids[];
  private int lookup_number = -1;

}

