/**
 * FullTestRespMsg.java
 *
 * @author Ben Y. Zhao
 * @version $Id: FullTestRespMsg.java,v 1.10 2003/03/19 17:55:44 hling Exp $
 */

package ostore.tapestry.impl;
import ostore.network.NetworkMessage;
import ostore.util.NodeId;
import ostore.util.SecureHash;
import ostore.util.ByteUtils;
import ostore.util.QSException;
import ostore.util.OutputBuffer;
import ostore.util.InputBuffer;
import ostore.util.QuickSerializable;

/**
 * Message used by a new FullTestMember node to respond to the 
 * FullTestDriver
 */
public class FullTestRespMsg 
  extends NetworkMessage implements QuickSerializable {

  private boolean success;
  public int requestType;
  public SecureHash requestID;
  private String output;
  private SecureHash src;
  public long integrateTime;
  public OverlayPath [] paths=null; //for ROUTING_PATH_RESULT,
                            //overlay hops.


  public void setTime(long time) {
      integrateTime = time;
  }

  public FullTestRespMsg (NodeId peer, boolean success, int requestType,
			  SecureHash requestID, SecureHash src,
			  String output, OverlayPath [] paths){
    super (peer, false /*outbound*/);
    this.success = success;
    this.requestType = requestType;
    this.requestID = requestID;
    this.src = src;
    this.output = output;
    this.paths = paths;
  }

  public FullTestRespMsg (NodeId peer, boolean success, int requestType,
			  SecureHash requestID, SecureHash src, String output){
    super (peer, false /*outbound*/);
    this.success = success;
    this.requestType = requestType;
    this.requestID = requestID;
    this.src = src;
    this.output = output;
  }

  public FullTestRespMsg (InputBuffer buffer) throws QSException {
    super(buffer);
    success = buffer.nextBoolean();
    requestType = buffer.nextInt();
    requestID = (SecureHash) buffer.nextObject();
    src = (SecureHash) buffer.nextObject();
    output = buffer.nextString();
    integrateTime = buffer.nextLong();
    
    //For overlay path request.
    boolean  have_paths = buffer.nextBoolean();
    if ( have_paths )
    {
	int size = buffer.nextInt();
	paths = new OverlayPath[size];
	for (int i = 0 ; i<size; i++) {
	    paths[i] = (OverlayPath) buffer.nextObject();
	}
    }
  }

  public void serialize (OutputBuffer buffer) {
    super.serialize (buffer);
    buffer.add(success);
    buffer.add(requestType);
    buffer.add(requestID);
    buffer.add(src);
    buffer.add(output);
    buffer.add(integrateTime);

    //For overlay path request.
    if (paths == null)
	buffer.add(false);
    else
    {
	buffer.add(true);
	buffer.add(paths.length);
	for (int i=0; i<paths.length; i++) {
	    buffer.add(paths[i]);
	}
    }
  }

  public SecureHash getID() {
    return requestID;
  }

  public SecureHash getSrc() {
    return src;
  }

  public boolean success() {
    return success;
  }

  public Object clone () throws CloneNotSupportedException {
    FullTestRespMsg result = (FullTestRespMsg) super.clone();
    result.success = success;
    result.requestType = requestType;
    result.requestID = requestID;
    result.output = output;
    result.src = src;
    result.integrateTime = integrateTime;
    return result;
  }

  public String toString() {
    if (success)
      return "(TestRespMsg: success on operation:"+requestType+
	", from="+peer+", id="+requestID+", output="+output +")";
    else 
      return "(TestRespMsg: failure on operation:"+requestType+
	", from="+peer+", id="+requestID+", output="+output +")";
  }
}
