/**
 * RouteObjDestMsg.
 *
 * @author     Sean C. Rhea
 * @version    $Id: RouteObjDestMsg.java,v 1.24 2002/08/18 22:39:41 srhea Exp $
 */

package ostore.tapestry.impl;
import ostore.network.NetworkMessage;
import ostore.tapestry.channel.MsgQos;
import ostore.util.AssertionViolatedException;
import ostore.util.CountBuffer;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.QSException;
import ostore.util.SecureHash;
import ostore.util.SHA1Hash;
import ostore.util.OutputBuffer;
import ostore.util.TypeTable;
import ostore.util.QuickSerializable;

public class RouteObjDestMsg 
extends TapestryMessage implements QuickSerializable {

    public SecureHash src;
    public SecureHash dest;
    public SecureHash guid;
    public int hop;
    public int hopInit;

    private QuickSerializable _user_data;
    private int _user_data_length;
    private InputBuffer _user_data_buffer;

    ///This is moved to TapestryMessage by Ling Huang 07/31/02.
    //Added by Ling Huang for reliable msg delivery.
    //    protected int qos;                //The QOS of this message, should it be reliable or un-.
    //protected SecureHash msg_id;      //Unique identity for a message for retransmission.

    public QuickSerializable user_data () throws QSException {
	if (_user_data == null) {
	    _user_data = _user_data_buffer.nextObject ();
	    _user_data_buffer = null; // save memory
	}
	return _user_data;
   }

   public RouteObjDestMsg (NodeId peer, SecureHash src, SecureHash dest,
	    SecureHash guid, int hop, QuickSerializable user_data) {
	super (peer, false /* outbound */);
	this.src = src;
	this.dest = dest;
	this.guid = guid;
	this.hop = hop;
	_user_data = user_data;

	this.qos = MsgQos.UNRELIABLE;
	this.msg_id = null;
    }

   public RouteObjDestMsg (NodeId peer, SecureHash src, SecureHash dest,
	    SecureHash guid, int hop, int user_data_length, 
	    InputBuffer user_data_buffer) {
	super (peer, false /* outbound */);
	this.src = src;
	this.dest = dest;
	this.guid = guid;
	this.hop = hop;
	this._user_data_length = user_data_length;
	this._user_data_buffer = user_data_buffer;

	this.qos = MsgQos.UNRELIABLE;
	this.msg_id = null;
    }

    public RouteObjDestMsg (InputBuffer buffer) throws QSException {
	super (buffer);
	src = (SecureHash) buffer.nextObject ();
	dest = (SecureHash) buffer.nextObject ();
	guid = (SecureHash) buffer.nextObject ();
	hop = buffer.nextInt ();
	hopInit = buffer.nextInt ();
	_user_data_length = buffer.nextInt ();
	_user_data_buffer = buffer.subBuffer (_user_data_length);

	/*This is moved to TapestryMessage by Ling Huang 07/31/02.
	qos = buffer.nextInt ();
	if (qos == MsgQos.RELIABLE)
	    msg_id = (SecureHash) buffer.nextObject ();
	*/
    }

    public void serialize (OutputBuffer buffer) {
	super.serialize (buffer);

	buffer.add (src);
	buffer.add (dest);
	buffer.add (guid);
	buffer.add (hop);
	buffer.add (hopInit);
	if (_user_data != null) {
	    CountBuffer cb = new CountBuffer ();
	    cb.add (_user_data);
	    buffer.add (cb.size ());
	    buffer.add (_user_data);
	}
	else {
	    buffer.add (_user_data_length);
	    try {
		// In order to make the add_to_digest idempotent, I have to
		// clone the _user_data_buffer before reading the data out
		// of it.  Otherwise, it will be empty the next time this
		// function is called.
		buffer.add ((InputBuffer) _user_data_buffer.clone ());
	    }
	    catch (CloneNotSupportedException e) {
		throw new AssertionViolatedException ("must be cloneable");
	    }
	}
	/*This is moved to TapestryMessage by Ling Huang 07/31/02.
	buffer.add (qos);
	if (qos == MsgQos.RELIABLE)
	    buffer.add (msg_id);
	*/
    }

    public Object clone () throws CloneNotSupportedException {
	RouteObjDestMsg result = (RouteObjDestMsg) super.clone ();
	result.src = src;
	result.dest = dest;
	result.guid = guid;
	result.hop = hop;
	result.hopInit = hopInit;
	result._user_data = _user_data;
	result._user_data_length = _user_data_length;
	if (_user_data_buffer != null)
	    result._user_data_buffer = (InputBuffer) _user_data_buffer.clone ();

	//This is moved to TapestryMessage by Ling Huang 07/31/02.
	//	result.qos = qos;
	//	result.msg_id = msg_id;

	return result;
    }

    // Added by Ling Huang for msg retransmission.
    /*
    public int get_qos()                    {   return qos; }
    public boolean is_reliable()
    {
	if (qos == MsgQos.RELIABLE)
	    return true;
	return false;
    }
    public SecureHash get_msg_id()          {   return msg_id; }
    public void set_msg_id(SecureHash id)   {   msg_id = id; }
    */

    public void set_qos(int qos)
    {   
	this.qos = qos;
	//	if (qos == MsgQos.RELIABLE && msg_id == null)
	if (msg_id == null)
	    generate_msg_id();
    }
    public SecureHash generate_msg_id()
    {
	double  id;
	String  id_string;

	id = java.lang.Math.random();
	id_string = src.toString() + dest.toString() + Double.toString(id);
	msg_id = new SHA1Hash(id_string);

	return msg_id;
    }
    public SecureHash generate_msg_id(Object id)
    {
	String  id_string;

	id_string = src.toString() + dest.toString() + id.toString();
	msg_id = new SHA1Hash(id_string);

	return msg_id;
    }

    public String toString () {
	return "(RouteObjDestMsg " + super.toString () + ")";
    }
}













