/**
 * TapestryCarrierMessage.
 *
 * @author     Ling Huang 
 * @version    $Id: TapestryCarrierMessage.java,v 1.1 2002/08/16 15:17:31 hling Exp $
 */

package ostore.tapestry.impl;
import ostore.tapestry.channel.Snippet;
import seda.sandStorm.lib.aSocket.ATcpInPacket;
import java.util.TreeSet;
import ostore.tapestry.channel.Snippet;
import ostore.network.NetworkInputBuffer;
import ostore.network.NetworkOutputBuffer;
import ostore.util.AssertionViolatedException;
import ostore.util.ByteUtils;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.QSArray;
import ostore.util.QSException;
import ostore.util.QuickSerializable;
import ostore.util.SecureHash;
import ostore.util.Debug;
import ostore.util.DebugFlags;

public abstract class TapestryCarrierMessage extends TapestryMessage 
{
    private static final boolean DEBUG = DebugFlags.TAPESTRY_CHANNEL;
    //    private static final boolean DEBUG = false;
    private String class_tag = "TapestryCarrierMessageMsg";

    //Ling Huang: I move data payload here from RouteMsg, 
    //RouteObjRootMsg and RouteObjDesMsg.
    //Modified by Ling Huang on 08/06/02 for even size fragmentation.
    //We need a linked list for geting user_data from NetworkOutputBuffer.
    //    protected byte[] _user_data;
    //    protected int _user_data_size;
    //    protected int _user_data_offset;
    //After deserializing from NetworkInputBuffer,
    //We use following two fields for user payload.
    protected byte [] _user_data;
    protected int _user_data_size;

    //Before serializing to NetworkOutputBuffer and after payload data 
    //fragmentation and RoutMsg construction,
    //we use following a linked list of Snippet for user payload.
    //we also put the total amount of data (in byte) in _user_data_size.
    protected Snippet _head, _tail;
    protected int _snippet_num;


    public TapestryCarrierMessage (NodeId peer, boolean inbound) 
    {
	super(peer, inbound);

	this._head = this._tail = null;
	this._user_data = null;
	this._user_data_size = 0;
	this._snippet_num = 0;
    }

    public TapestryCarrierMessage (NodeId peer, boolean inbound,
				   byte user_data[], int length, int offset)
    {
	super(peer, inbound);

	this._head = this._tail = null;
	this._user_data = null;
	this._user_data_size = 0;
	this._snippet_num = 0;
	this.add_user_data(user_data, length, offset);
    }

    public TapestryCarrierMessage (NodeId peer, boolean inbound, 
				   QuickSerializable msg) 
    {
	super (peer, inbound);

	this._head = this._tail = null;
	this._user_data = null;
	this._user_data_size = 0;
	this._snippet_num = 0;

	//Serialize the msg into NetworkOutputBuffer to
	//get one fragments (byte arraies).
	NetworkOutputBuffer frag_buffer = new NetworkOutputBuffer ();
	frag_buffer.add(msg);
	NetworkOutputBuffer.Snippet walker = frag_buffer._head;
	while (walker != null) 
	{
	    //add to msg as payload
	    add_user_data(walker.data, walker.offset, walker.length);
	    walker = walker.next;
	}
    }

    public TapestryCarrierMessage (InputBuffer buffer)  throws QSException
    {
	//	final String tag = class_tag + ".deserialize()";

	super(buffer);
	_user_data_size = buffer.nextInt ();
	//Modified by Ling Huang on 08/05/02: _user_data
	// change from QuickSerializable to byte [].
	//	_user_data_buffer = buffer.subBuffer (_user_data_size);
	if (_user_data_size > 0)
	{
	    //	    if (DEBUG) Debug.printtagln (tag, "Try to read " + 
	    //		       _user_data_size + " bytes of data from InputBuffer\n");
	    //Use function void nextBytes (byte [] output, int offset, int length).
	    //Attention: we set offset = 0, so that at here _user_data only contains
	    //a fragment of a msg. This is different from the _user_data when it goes
	    //into an OutputBuffer from the RouteMsg.
	    _user_data = new byte[_user_data_size];
	    buffer.nextBytes(_user_data, 0, _user_data_size);
	}

	this._head = this._tail = null;
	this._snippet_num = 0;
	//Modified on 08/05/02 end.
    }

    public void serialize (OutputBuffer buffer) 
    {
	final String tag = class_tag + ".serialize()";

	super.serialize (buffer);

	//Modified by Ling Huang on 08/05/02: for msg fragmentation
	//_user_data change from QuickSerializable to byte [].
	buffer.add(_user_data_size);
	if (_user_data_size > 0)
	{
	    Snippet    walker;
	    String     str_out = new String();
	    int        actual_data_size;
	    byte       data_bytes[];

	    actual_data_size = 0;
	    if (_head != null)
	    {
		walker = _head;
		while (walker != null)
		{
		    if (DEBUG) 
		    {
			str_out = ByteUtils.print_bytes (walker.data, 
				  walker.offset, walker.length);
			Debug.printtagln (tag, str_out);
		    }
		    //Start at position _user_data_offset, read _user_data_size
		    //amount of data from _user_data and put into OutputBuffer.
		    buffer.add (walker.data, walker.offset, walker.length);
		    actual_data_size = actual_data_size + walker.length;
		    walker = walker.next;
		}

		/*
		walker = _head;
		data_bytes = new byte[100000];
		while (walker != null)
		{
		    System.arraycopy(walker.data, walker.offset, 
				     data_bytes, actual_data_size, walker.length);
		    actual_data_size = actual_data_size + walker.length;
		    walker = walker.next;
		}
		buffer.add (data_bytes, 0, actual_data_size);
		if (DEBUG) Debug.printtagln (tag, str_out);
		*/
	    }
	    else
	    {
		buffer.add(_user_data, 0, _user_data_size);
		actual_data_size = _user_data.length;
	    }

	    if (DEBUG) 
	    {
		Debug.printtagln (tag, "user_data check when " + 
				  "serializing TapestryCarrierMessage:");
		if (actual_data_size != _user_data_size)
		{
		    Debug.printtagln (tag,"Serious problem for msg " + msg_id + 
				       " : actual_data_size(" + 
				       actual_data_size + ") <> _user_data_size(" + 
				       _user_data_size + ")");
		}
		else
		{
		    Debug.printtagln (tag,"No problem for msg " + msg_id + 
				       " : actual_data_size(" + 
				       actual_data_size + ") == _user_data_size(" + 
				       _user_data_size + ")");
		}
	    }
	}

	/* 
	if (_user_data != null) 
	{
	    // TODO: find a better way to get size of message.
	    CountBuffer cb = new CountBuffer ();
	    cb.add (_user_data);
	    buffer.add (cb.size ());
	    buffer.add (_user_data);
	}
	else 
	{
	    buffer.add (_user_data_size);
	    try {
		// In order to make the add_to_digest idempotent, I have to
		// clone the _user_data_buffer before reading the data out
		// of it.  Otherwise, it will be empty the next time this
		// function is called.
		buffer.add ((InputBuffer) _user_data_buffer.clone ());
	    }
	    catch (CloneNotSupportedException e) {
		throw new AssertionViolatedException ("must be cloneable");
	    }
	}
	*/
	//Modified on 08/05/02 end.
    }

    public void add_user_data (byte [] input, int offset, int length) 
    {
	if (length == 0) 
	    return;

	//	if (DEBUG) System.out.println ("add array " + input);
	if (_head == null) 
	{
	    _head = _tail = new Snippet (input, offset, length);
	    //	    if (DEBUG) System.out.println ("was empty, added whole " + _head.data);
	}
	else 
	{
	    _tail.next = new Snippet (input, offset, length);
	    _tail = _tail.next;
	    //	    if (DEBUG) System.out.println ("adding buf " + _tail.data);
	}

	_user_data_size += length;
	_snippet_num++;
    }

    /**
     * @deprecated As of 7/24/2002 in favor of msg fragmentation 
     *  by Ling Huang on 08/01/02:
     *  change the payload of RouteMsg to be byte[],
     *  which in many cases is a fragment of a msg.
     *  QuickSerializable is meanless for those fragments
     */
    //It only works after you deserialize from netowrks.
    public QuickSerializable user_data () throws QSException 
    {
	//Modified by Ling Huang for msg Fragmentation.
	/*
	if (_user_data == null) 
	{
	    _user_data = _user_data_buffer.nextObject ();
	    _user_data_buffer = null; // save memory
	}
	return _user_data;
	*/
	final String tag = class_tag + ".user_data()";

	//For temporary usage: get an object from byte array.
	//Finally we should get rid of this function.
	//It only works for msg with one fragments.
	if (frg_num > 1)
	{
	    throw new AssertionViolatedException 
		( tag + ": frg_num > 1, cannot use user_data()!");
	}

	TreeSet fragments = new TreeSet();
	//Add the fragment into the set.
	fragments.add(this);
	return ChannelHandler.reassemble_msg(fragments);

	/*	
	QuickSerializable  msg;
	NetworkInputBuffer inputbuffer;
	ATcpInPacket    packet;
	Snippet         walker;
	int             snippet_id;

	if (_user_data_size > 0)
	{
	    if (_user_data != null)
	    {
		inputbuffer = new NetworkInputBuffer(0);
		//		connection = new ATcpConnection();
		packet = new ATcpInPacket(null, _user_data, 
					  _user_data_size, 
					  false, // no copy
					  0);
		inputbuffer.add_packet(packet);
		msg = inputbuffer.nextObject();
		return msg;
	    }
	    else if (_head != null)
	    {
		byte [] data_buff;
		
		inputbuffer = new NetworkInputBuffer(0);
		snippet_id = 0;
		walker = _head;
		while (walker != null)
		{
		    if (walker.offset == 0)
			data_buff = walker.data;
		    else
		    {
			data_buff = new byte[walker.length];
			System.arraycopy (walker.data, walker.offset,
					  data_buff, 0, walker.length);
			if (DEBUG)
			{
			    Debug.printtagln(tag, 
				  ": An extra copy when reassemble user_data!");
			}
		    }
		    packet = new ATcpInPacket(null, data_buff, 
					      walker.length, 
					      false, // no copy
					      snippet_id );
		    inputbuffer.add_packet(packet);
		    
		    snippet_id ++;
		    walker = walker.next;
		}
		
		msg = inputbuffer.nextObject();
		return msg;
	    }
	    else
	    {
		throw new AssertionViolatedException 
		    ( tag + ": _user_data_size > 0, but no user_data available!");
	    }
	}
	else
	{
	    return null;
	}
	*/
    }
    
    public abstract void set_qos(int qos);
    public abstract SecureHash generate_msg_id();
    public abstract SecureHash generate_msg_id(Object id);

    public Snippet get_user_data_head ()    {    return _head; }
    public Snippet get_user_data_tail ()    {    return _tail; }
    public int     get_user_data_size()     {    return _user_data_size; }
    //Attention: Only use get_user_data() after deserialize.
    public byte [] get_user_data()          {    return _user_data; }

    public void    set_user_data(byte [] user_data)
    {
	_user_data = user_data; 
	return;
    }

    public void set_user_data_head (Snippet head)       
    {    
	_head= head; 
	return;
    }
    public void set_user_data_tail (Snippet tail)       
    {    
	_tail= tail; 
	return;
    }

    public void set_user_data_size(int length)  
    {    
	_user_data_size = length; 
	return;
    }

    public int user_data_check()
    {
	Snippet    walker;
	int        actual_data_size;
	int        ret;

	final String tag = class_tag + ".user_data()";

	actual_data_size = 0;
	if (_head != null)
	{
	    walker = _head;
	    while (walker != null)
	    {
		actual_data_size = actual_data_size + walker.length;
		walker = walker.next;
	    }
	}
	else
	    actual_data_size = _user_data.length;

	if (actual_data_size != _user_data_size)
	{
	    if (DEBUG) 
		Debug.printtagln (tag, "Serious problem for msg " + 
				  msg_id + " : actual_data_size(" + 
				  actual_data_size + ") <> _user_data_size(" + 
				  _user_data_size + ")");
	    ret = 0;
	}
	else
	{
	    if (DEBUG) 
		Debug.printtagln (tag, "No problem for msg " + 
				  msg_id + " : actual_data_size(" + 
				  actual_data_size + ") == _user_data_size(" + 
				  _user_data_size + ")\n");

	    ret = 1;
	}

	return ret;
    }

    public Object clone () throws CloneNotSupportedException 
    {
	TapestryCarrierMessage result = (TapestryCarrierMessage) super.clone ();

	result._head = _head;
	result._tail = _tail;
	result._user_data = _user_data;
	result._user_data_size = _user_data_size;
	result._snippet_num = _snippet_num;

	//	if (_user_data_buffer != null)
	//	    result._user_data_buffer = (InputBuffer) _user_data_buffer.clone ();
	return result;
    }

    public String toString () 
    {
	return "(TapestryCarrierMessage)";
    }
}














