/**
 * TapestryMessage.
 *
 * @author     Ling Huang 
 * @version    $Id: TapestryMessage.java,v 1.6 2003/02/05 22:33:24 hling Exp $
 */

package ostore.tapestry.impl;
import ostore.tapestry.channel.MsgQos;
import ostore.network.NetworkMessage;
import ostore.util.AssertionViolatedException;
import ostore.util.ByteUtils;
import ostore.util.CountBuffer;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.QSArray;
import ostore.util.QSByteArray;
import ostore.util.QSException;
import ostore.util.QuickSerializable;
import ostore.util.SecureHash;
import ostore.util.SHA1Hash;
import ostore.util.TypeTable;

public class TapestryMessage extends NetworkMessage 
{
    //Added by Ling Huang for reliable msg delivery.
    protected int qos;                //The QOS of this message, should it be reliable or un-.
    protected SecureHash msg_id;      //Unique identity for a message for retransmission.

    protected int frg_num;          //The total number of segments for a message.     
    protected int frg_id;           //The order of this segment in the message. 
    protected int d_type;          //delievery type: unicast, multicast or anycast.
    public static final int   UNICAST = 1001; 
    public static final int MULTICAST = 1002; 
    public static final int   ANYCAST = 1003; 

    public TapestryMessage (NodeId peer, boolean inbound) 
    {
      super(peer, inbound);

      qos = MsgQos.UNRELIABLE;
      msg_id = null;
      frg_num = -1;
      frg_id = -1;
      d_type = UNICAST;
    }

    public TapestryMessage (InputBuffer buffer)  throws QSException
    {
      super(buffer);

      qos = buffer.nextInt ();
      //      if (qos == MsgQos.RELIABLE)
      boolean id = buffer.nextBoolean();
      if (id) 
	  msg_id = (SecureHash) buffer.nextObject ();
      else 
	  msg_id = null;
      frg_num = buffer.nextInt();
      frg_id  = buffer.nextInt();
      d_type = buffer.nextInt();
    }

    public void serialize (OutputBuffer buffer) 
    {
      super.serialize (buffer);

      buffer.add (qos);
      //      if (qos == MsgQos.RELIABLE)
      if (msg_id == null) 
	  buffer.add(false);
      else {
	  buffer.add(true);
	  buffer.add (msg_id);
      }
      buffer.add (frg_num);
      buffer.add (frg_id);
      buffer.add (d_type);
    }

    public Object clone () throws CloneNotSupportedException 
    {
	TapestryMessage result = (TapestryMessage) super.clone ();

	result.qos     = qos;
	result.msg_id  = msg_id;
	result.frg_num = frg_num;
	result.frg_id  = frg_id;
	result.d_type  = d_type;

	return result;
    }

    public boolean is_reliable()
    {
	if (qos == MsgQos.RELIABLE)
	    return true;
	return false;
    }

    public int get_qos()                       {   return qos; }
    public int get_frg_num()                   {   return frg_num; }
    public int get_frg_id()                    {   return frg_id; }
    public SecureHash get_msg_id()             {   return msg_id; }
    public int get_delivery_type()             {   return d_type; }

    public void set_msg_id(SecureHash id)      {   msg_id = id; }
    public void set_frg_num(int num)           {   frg_num = num; }
    public void set_frg_id(int id)             {   frg_id = id; } 
    public void set_delivery_type(int flag)    {   d_type = flag; }

    public String toString () 
    {
	return "(TapestryMessage)";
    }
}









