package ostore.tapestry.interweave;
import ostore.tapestry.api.*;
import ostore.util.NodeId;
import ostore.util.SecureHash;
import ostore.util.ByteUtils;
import ostore.util.QSException;
import ostore.util.OutputBuffer;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;

/**
 * InterweaveMetadataMsg.java
 *
 * The message that carries meta data for a requested file.
 *
 * @author Benjamin K. Poon
 */
public class InterweaveMetadataMsg 
    extends TapestryRouteMsg implements QuickSerializable {

    public SecureHash from;
    public String filename;
    public String login;
    public long lastModified;
    public long size;
    public String type;

    public InterweaveMetadataMsg (SecureHash to, SecureHash from, 
				  String filename, String login, long lastModified, long size, String type) {
	super(to);
	this.from = from;
	this.filename = filename;
	this.login = login;
	this.lastModified = lastModified;
	this.size = size;
	this.type = type;
    }

    public InterweaveMetadataMsg (InputBuffer buffer) throws QSException {
	super(buffer);
	from = (SecureHash)buffer.nextObject();
	filename = buffer.nextString();
	login = buffer.nextString();
	lastModified = buffer.nextLong();
	size = buffer.nextLong();
	type = buffer.nextString();
    }
    
    public void serialize (OutputBuffer buffer) {
	super.serialize (buffer);
	buffer.add(from);
	buffer.add(filename);
	buffer.add(login);
	buffer.add(lastModified);
	buffer.add(size);
	buffer.add(type);
    }
    
    public Object clone () throws CloneNotSupportedException {
	InterweaveMetadataMsg result = (InterweaveMetadataMsg) super.clone();
	result.from = from;
	result.filename = filename;
	result.login = login;
	result.lastModified = lastModified;
	result.size = size;
	result.type = type;
	return result;
    }

    public String toString() {
	return "(InterweaveMetadataMsg: from="+from+"filename="+
	    filename+", login="+login+", lastModified="+lastModified+", size="+size+
	    ", type="+type+")";
    }

}
