/**
 * OneStoneStage.java
 *
 * @author	Jeremy Stribling
 * @version	$Id: OneStoneStage.java,v 1.3 2003/02/09 22:22:10 strib Exp $
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ostore.tapestry.patchwork;

import ostore.apps.gnutella_trace.GnutellaPDF;
import ostore.apps.gnutella_trace.GnutellaHourPDF;
import ostore.util.InvalidConfigFileException;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.network.patchwork.InjectFaultMsg;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.network.NetworkMessage;
import ostore.util.NodeId;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.UnknownEventException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkException;   
import seda.sandStorm.api.SinkFullException;   
import seda.sandStorm.api.StagesInitializedSignal; 
import seda.sandStorm.main.SandstormConfig;
import java.lang.reflect.Constructor;

/**
 * @author 	Jeremy Stribling
 * @version	$Id: OneStoneStage.java,v 1.3 2003/02/09 22:22:10 strib Exp $
 */  
public class OneStoneStage implements EventHandlerIF
{
    ////////////////////       CONSTANTS	////////////////////

    private static final boolean DEBUG = false;

    private Classifier classifier;

    private NodeId self_node_id;

    private NodeId nodes[];

    private SinkIF my_sink;

    ////////////////////  STATIC CONSTRUCTORS	////////////////////
    ////////////////////      CONSTRUCTORS	////////////////////
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Specified by seda.sandStorm.api.EventHandlerIF 
     */
    public void handleEvents(QueueElementIF elemarr[]) 
	throws EventHandlerException
    {
	for( int i=0; i<elemarr.length; i++ ) {
	    handleEvent( elemarr[i] );
	}
    }

    /** 
     * Specified by seda.sandStorm.api.EventHandlerIF 
     */
    public void handleEvent(QueueElementIF element) 
	throws EventHandlerException
    {
	if( element instanceof StagesInitializedSignal ) {

	    // send two fault messages to each, 
	    // telling them to die gracefully
	    for( int i = 0; i < nodes.length; i++ ) {
		dispatch( new InjectFaultMsg(nodes[i], 100, true, true ) );
		dispatch( new RepairMonitorStage.DetachMessage( nodes[i], 
								true ) );
	    }
	    // just to be safe, send out an alarm before we die
	    dispatch_later( new AlarmMsg(), 10000 );

	} else if( element instanceof AlarmMsg ) {

	    System.out.println( "Tapestry: ready" );
	    System.exit(0);

	} else {
	    
	    Carp.error( "Dropping unexpected message: " + element );
	    
	}
    }

    /** 
     * Specified by seda.sandStorm.api.EventHandlerIF 
     */
    public void init(ConfigDataIF config) throws Exception
    {

	///////////////////
	// get config info
	String nodeNames[] = config.getStringList( "nodes" );
	nodes = new NodeId[ nodeNames.length ];
	self_node_id = new NodeId (config.getString ("node_id"));
	for( int i = 0; i < nodeNames.length; i++ ) {
	    nodes[i] =  new NodeId( nodeNames[i] );
	}

	classifier = Classifier.getClassifier( self_node_id );
	my_sink = config.getStage().getSink();
	Filter filter;

	// Subscribe to all my favorite events

	String[] eventNames = {
	    "ostore.tapestry.patchwork.OneStoneStage$AlarmMsg",
	    "seda.sandStorm.api.StagesInitializedSignal"
	};

	for( int i = 0; i < eventNames.length; i++ ) {
	  filter = new Filter();
	  Class eventClass;
          try {
            eventClass = Class.forName( eventNames[i] );
	    if( !filter.requireType( eventClass ) || 
	        !classifier.subscribe( filter, my_sink ) ) {

	      Carp.die( "Could not require class " + eventNames[i] ); 

	    } 

          } catch ( ClassNotFoundException cnfe ) {
            Carp.die( cnfe );
          }

        }

	String[] inb_eventNames = {
	};


	// require inbounders
	for( int i = 0; i < inb_eventNames.length; i++ ) {
	  filter = new Filter();
	  Class eventClass;
          try {
            eventClass = Class.forName( inb_eventNames[i] );
	    if( !filter.requireType( eventClass ) ||
		!filter.requireValue ( "inbound", new Boolean(true) ) || 
	        !classifier.subscribe( filter, my_sink ) ) {

	      Carp.die( "Could not require class " + inb_eventNames[i] ); 

	    } 

          } catch ( ClassNotFoundException cnfe ) {
            Carp.die( cnfe );
          }

        }
	
    }

    /**
     * An internal method used for dispatching events
     */
    private void dispatch( QueueElementIF event ) {
                                                
      try {
        classifier.dispatch( event );
      } catch ( SinkFullException sfe ) {
        //TODO: Better way to deal with SinkFull
        try {
          Thread.currentThread().sleep( 5000 );
        } catch( InterruptedException ie ) {
          //ignore
        }
      }                                         
      
    }

    /**
     * An internal method used for dispatching events
     */
    private void dispatch_later( QueueElementIF event, long millis ) {
                                                
        classifier.dispatch_later( event, millis );

    }

    /** 
     * Specified by seda.sandStorm.api.EventHandlerIF 
     */
    public void destroy()
    {
    }

    private class AlarmMsg implements QueueElementIF {

    }

}
