/**
 * PatchworkBuilder.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.tapestry.patchwork;

import soss.core.BuilderIF;
import ostore.util.Carp;
import seda.sandStorm.main.SandstormConfig;
import seda.sandStorm.api.StageNameAlreadyBoundException;
import java.io.IOException;
import java.util.StringTokenizer;

/**
 * A configuration builder for Patchwork clients
 * Modelled after <code>ostore.bloom.DualBuilder</code>.
 *
 * @author	Jeremy Stribling
 * @version	$Id: PatchworkBuilder.java,v 1.11 2002/12/20 23:54:24 strib Exp $
 */
public class PatchworkBuilder implements BuilderIF
{
  ////////////////////       CONSTANTS	////////////////////
  ////////////////////    STATIC VARIABLES 	////////////////////
  ////////////////////       VARIABLES	////////////////////

  /**
   * The node id of the Federation. */
  private String _federation;
  private String _gateway;

  ////////////////////  STATIC CONSTRUCTORS	////////////////////
  ////////////////////      CONSTRUCTORS	////////////////////

  /**
   * Set global parameters. */
  public PatchworkBuilder(String[] args)
  {
    if (args.length != 2) 
      Carp.die ("usage: PatchworkBuilder <fed> <gateway>");
	
    _federation = args[0];
    _gateway = args[1];
  }

  ////////////////////     STATIC METHODS 	////////////////////
  //////////////////// 	METHODS		////////////////////

  /**
   * Create a configuration object for a single virtual machine. */
  public SandstormConfig build_node(String[] args)
      throws StageNameAlreadyBoundException, IOException
  {
      if( (args.length == 28 && !args[19].equals("me")) ||
	  ((args.length != 22) && (args.length != 28)) )
    {
      String str = new String();
      for(int i = 0; i < args.length; i++)
	  str += args[i] + " ";
      Carp.die ("usage: PatchworkBuilder.build_node <node_id>" +
		" <sim_node?> <respond_period> <loss_period>" + 
		" <loss_op> <loss_factor> <latency_period>" + 
		" <latency_op> <latency_factor> <bandwidth_period>" +
		" <bandwidth_op> <bandwidth_factor> <recalc_period>" +
		" <desired_successes> <second_chances_thresh> " +
		" <second_chances_resp> <timeout_packets> <random_seed>" + 
		" <no_extra_stages> <repair_test> <dynamic> <deletable>" + 
		" [<node1_node2_etc.>] [<fail_and_replace?>]" + 
		" [<blank1_blank2_etc>] [<cdf_class>] [<gateway>]" + 
		" [<time_to_run>]");
    }

    String node_id = args[0];
    boolean sim_node = args[1].equals( "true" );
    String respond_period = args[2];
    String loss_period = args[3];
    String loss_op = args[4];
    String loss_factor = args[5];
    String latency_period = args[6];
    String latency_op = args[7];
    String latency_factor = args[8];
    String bandwidth_period = args[9];
    String bandwidth_op = args[10];
    String bandwidth_factor = args[11];
    String recalc_period = args[12];
    String desired_successes = args[13];
    String second_chances_thresh = args[14];
    String second_chances_resp = args[15];
    String timeout_packets = args[16];
    String random_seed = args[17];
    boolean no_extra_stages = args[18].equals( "1" );
    boolean repair_monitor = args[19].equals( "me" );
    String dynamic = args[20];
    String deletable = args[21];
    boolean repair_test;
    if( repair_monitor ) {
	repair_test = true;
    } else {
	repair_test = args[19].equals( "true" );
    }
    String nodeNames = null;
    String blankNames = null;
    String cdf_class = null;
    String gateway = null;
    String time_to_run = null;
    String fail_and_replace = null;
    if( repair_monitor ) {
	nodeNames = args[22];
	fail_and_replace = args[23];
	blankNames = args[24];
	cdf_class = args[25];
	gateway = args[26];
	time_to_run = args[27];
    }
    StringTokenizer nodeST = null;
    StringTokenizer blankST = null;
    
    if( repair_monitor ) {
	nodeST = new StringTokenizer( nodeNames, "_" );
	blankST = new StringTokenizer( blankNames, "_" );
    }

    SandstormConfig result = new SandstormConfig();
    
    // Router
    {
      String[] init_args = {
	  "node_id=" + node_id,
	  "dynamic_route=" + dynamic,
	  "fake_keys=true"		
        };
      result.addStage("Router", 
		      "ostore.tapestry.impl.Router",
		      init_args);
    }

    // StaticTClient
    {
      String[] init_args = {
	  "node_id=" + node_id,
	  "dynamic_route=" + dynamic,
	  "federation=" + _federation
        };
      result.addStage("StaticTClient", 
		      "ostore.tapestry.impl.StaticTClient",
		      init_args);
    }

    // DynamicTClient
    {
	String[] init_args = {
	    "node_id=" + node_id,
	    "dynamic_route=" + dynamic,
	    "gateway=" + _gateway
	};
	result.addStage("DynamicTClient", 
			"ostore.tapestry.impl.DynamicTClient",
			init_args);
    }

    // Patchwork
    {
      String[] init_args = {
	  "node_id=" + node_id,
	  "loss_period=" + loss_period,
	  "loss_op=" + loss_op,
	  "loss_factor=" + loss_factor,
	  "latency_period=" + latency_period,
	  "latency_op=" + latency_op,
	  "latency_factor=" + latency_factor,
	  "bandwidth_period=" + bandwidth_period,
	  "bandwidth_op=" + bandwidth_op,
	  "bandwidth_factor=" + bandwidth_factor,
	  "recalc_period=" + recalc_period,
	  "respond_period=" + respond_period,
	  "second_chances_thresh=" + second_chances_thresh,
	  "second_chances_resp=" + second_chances_resp,
	  "timeout_packets=" + timeout_packets,
	  "optimize_on=true",
	  "accept_faults=true"
        };
	  result.addStage("PatchworkStage", 
			  "ostore.tapestry.patchwork.PatchworkStage",
			  init_args);
    }

    if( sim_node ) {

	{

	    // PatchworkSimulation 
	    String[] init_args = {
		"node_id=" + node_id,
		"deletable=" + deletable
	    };
		result.addStage("PatchworkSimulationStage", 
			  "ostore.tapestry.patchwork.PatchworkSimulationStage",
				init_args);
	}

    }

    if( !no_extra_stages ) {

	// ArchivalStorageStage
	{
	    String[] init_args = {
		"node_id=" + node_id,
		"AlarmTimeout=120000",
		"StorageServer=true",
		"LogEvents=false",
		"LogFilename=/dev/null",
		"LogSampleRate=1"
	    };
	    result.addStage("ArchivalStorageStage",
			    "ostore.archive.ArchivalStorageStage",
			    init_args);
	}

	// Cache
	{
	    String[] init_args = {
		"node_id=" + node_id,
		"CacheSize=104857600",
		"cachedir=/tmp/fakecache",
		"Publish=false",
		"one_cache=true"
	    };
	    result.addStage("MemoryCache",
			    "ostore.cache.MemoryCache",
			    init_args);
	}
	
    }

    if( repair_monitor ) {
	// RepairMonitor
	{
	    
	    String nodes = "";
	    while( nodeST.hasMoreTokens() ) {
		nodes += nodeST.nextToken() + " ";
	    }

	    String blanks = " ";
	    while( blankST.hasMoreTokens() ) {
		blanks += blankST.nextToken() + " ";
	    }
	    
	    String[] init_args = {
		"node_id=" + node_id,
		"num_failures=" + desired_successes,
		"random_seed=" + random_seed,
		"nodes=" + nodes,
		"fail_and_replace=" + fail_and_replace,
		"blanks=" + blanks,
		"cdf_class=" + cdf_class,
		"gateway=" + gateway,
		"builder_name=ostore.tapestry.patchwork.PatchworkBuilder",
		"builder_args=" + gateway,
		"node_args=" + "node_id " + sim_node + " " + respond_period 
		+ " " + loss_period + " " + loss_op + " " + loss_factor + 
		" " + latency_period + " " + latency_op + " " + 
		latency_factor + " " + bandwidth_period + " " + bandwidth_op 
		+ " " + bandwidth_factor + " " + recalc_period + " " +
		desired_successes + " " + second_chances_thresh + " " +
		second_chances_resp + " " + timeout_packets + " " + 
		random_seed + " true dynamic",
		"gateway=" + gateway,
		"time_to_run=" + time_to_run
		
	    };
	    result.addStage("RepairMonitorStage", 
			    "ostore.tapestry.patchwork.RepairMonitorStage",
			    init_args);
	}
    }
    
    return result;
  }  
  ////////////////////     INNER CLASSES	////////////////////
}

