/**
 * StartrekRouter.
 *
 * @author     Sean C. Rhea
 * @version    $Id: StartrekRouter.java,v 1.2 2003/02/19 10:32:31 hweather Exp $
 */

package ostore.tapestry.startrek;
import java.io.FileInputStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.tapestry.api.TapestryLocateFailure;
import ostore.tapestry.api.TapestryLocateMsg;
import ostore.tapestry.api.TapestryLocateReject;
import ostore.tapestry.api.TapestryMacRouteMsg;
import ostore.tapestry.api.TapestryPrefixRouteMsg;
import ostore.tapestry.api.TapestryPublishMsg;
import ostore.tapestry.api.TapestryQuery;
import ostore.tapestry.api.TapestryQueryResult;
import ostore.tapestry.api.TapestryQueryResultMatch;
import ostore.tapestry.api.TapestryQueryResultNoMatch;
import ostore.tapestry.api.TapestryQueryState;
import ostore.tapestry.api.TapestryRouteMsg;
import ostore.tapestry.api.TapestryTag;
import ostore.tapestry.api.TapestryUnpublishMsg;
import ostore.tapestry.impl.NodeGuid;
import ostore.tapestry.impl.PublishInfo;
import ostore.tapestry.impl.PublishMsg;
import ostore.tapestry.impl.RouteMsg;
import ostore.tapestry.impl.RouteObjDestMsg;
import ostore.tapestry.impl.StaticAllNodesMsg;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.QSClone;
import ostore.util.QSException;
import ostore.util.QSIO;
import ostore.util.QuickSerializable;
import ostore.util.SecureHash;
import ostore.util.SHA1Hash;
import ostore.security.QSPublicKey;
import seda.sandStorm.api.StagesInitializedSignal;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.NoSuchStageException;
import seda.sandStorm.api.ProfilableIF;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.internal.SinkProxy;

public class StartrekRouter implements EventHandlerIF {

    protected String tag = "StartrekRouter";
    protected static final boolean DEBUG = false;

    protected SinkIF sink;
    protected NodeId self_node_id;
    protected SecureHash self_guid;
    protected Classifier classifier;
    protected Map node_guid_to_node_id;
    protected Map guid_to_publish_info = new HashMap ();
    protected LinkedList not_ready_queue = new LinkedList ();

    protected void dispatch (QueueElementIF item) {
	if (DEBUG) Debug.printtagln (tag, "dispatching " +
		item.getClass () + " event: " + item + ".");
	try {
	    classifier.dispatch(item);
	} catch (SinkException e) {
	    BUG (tag + ".dispatch: could not dispatch " + item);
	}
    }

    protected static final void BUG (String msg) {
	Carp.die (msg + "  This is a bug in " +
		"ostore.tapestry.startrek.StartrekRouter.  " +
		"Please let Sean know about it ASAP.");
    }

    public void init (ConfigDataIF config) throws Exception {

	tag = config.getStage().getName();
	if (DEBUG) Debug.turnOn (tag);

	self_node_id = new NodeId (config.getString ("node_id"));
	sink = config.getStage().getSink();
	classifier = Classifier.getClassifier(self_node_id);

	String keyfilename = config.getString ("pkey");
	QSPublicKey pkey = null;
	try {
	    FileInputStream keyfile = new FileInputStream (keyfilename);
	    pkey = (QSPublicKey) QSIO.read (keyfile);
	    keyfile.close ();
	}
	catch (Exception e) {
	    Carp.die ("TClient.init:  Caught exception " + e +
		    " while trying to read pkey from disk.");
	}

	self_guid = new SHA1Hash (pkey);

	System.out.println ("Tapestry NodeId is " + self_node_id +
		", GUID is " + self_guid);

	String [] inb_msg_types = {
	    "ostore.tapestry.impl.PublishMsg",
	    "ostore.tapestry.impl.RouteMsg",
	    "ostore.tapestry.impl.RouteObjDestMsg",
	    "ostore.tapestry.impl.StaticAllNodesMsg"
	};

	for (int i = 0; i < inb_msg_types.length; ++i) {
	    Filter filter = new Filter ();
	    if (! filter.requireType (Class.forName (inb_msg_types [i])))
		BUG (tag + ": could not require type " + inb_msg_types [i]);
	    if (! filter.requireValue ("inbound", new Boolean (true)))
		BUG (tag + ": could not require inbound = true for " 
			+ inb_msg_types [i]);
	    classifier.subscribe (filter, sink);
	}

	String [] outb_msg_types = {
	    "ostore.tapestry.api.TapestryLocateMsg",
	    "ostore.tapestry.api.TapestryMacRouteMsg",
	    "ostore.tapestry.api.TapestryPrefixRouteMsg",
	    "ostore.tapestry.api.TapestryRouteMsg"
	};

	for (int i = 0; i < outb_msg_types.length; ++i) {
	    Filter filter = new Filter ();
	    if (! filter.requireType (Class.forName (outb_msg_types [i])))
		BUG (tag + ": could not require type " + outb_msg_types [i]);
	    if (! filter.requireValue ("inbound", new Boolean (false)))
		BUG (tag + ": could not require inbound = false for " 
			+ outb_msg_types [i]);
	    classifier.subscribe (filter, sink);
	}

	String [] event_types = {
	    "seda.sandStorm.api.StagesInitializedSignal",
	    "ostore.tapestry.api.TapestryPublishMsg",
	    "ostore.tapestry.api.TapestryUnpublishMsg"
	};

	for (int i = 0; i < event_types.length; ++i) {
	    Filter filter = new Filter ();
	    if (! filter.requireType (Class.forName (event_types [i])))
		BUG (tag + ": could not require type " + event_types [i]);
	    classifier.subscribe (filter, sink);
	}
    }

    public void destroy() throws Exception {}

    public void handleEvent (QueueElementIF item) throws EventHandlerException {

	if (DEBUG) Debug.printtagln (tag, "got " +
		item.getClass () + " event: " + item + ".");
	    
	if (item instanceof StagesInitializedSignal) {
	    // Let other nodes know the local node guid.
	    // Most stages extract local node guid from
	    // TapestryReadyMsg, but some stages need
	    // to know the local node guid before
	    // the tapestry is ready, e.g. DynamicTClient.
	    dispatch(new NodeGuid(self_guid));

	}
	else if (item instanceof StaticAllNodesMsg) {
	    StaticAllNodesMsg msg = (StaticAllNodesMsg) item;
	    node_guid_to_node_id = new HashMap ();
	    Iterator i = msg.nodes ().keySet ().iterator (); 
	    while (i.hasNext ()) {
		NodeId n = (NodeId) i.next ();
		SecureHash g = (SecureHash) msg.nodes ().get (n);
		node_guid_to_node_id.put (g, n);
	    }

	    while (! not_ready_queue.isEmpty ()) {
		QueueElementIF x = 
		    (QueueElementIF) not_ready_queue.removeFirst ();
		handleEvent (x);
	    }
	    not_ready_queue = null;

	    System.out.println("Tapestry: ready.\n");
	    System.out.flush();

	    return;
	}
	else if (node_guid_to_node_id == null) {
	    not_ready_queue.addLast (item);
	    return;
	}


	if (item instanceof TapestryPublishMsg) {
	    TapestryPublishMsg msg = (TapestryPublishMsg) item;

	    // Publish it locally.
	    publish (self_guid, msg.guid, msg.tag, true /* publish */);

	    // And then send it to everyone else.
	    Iterator i = node_guid_to_node_id.keySet ().iterator ();
	    while (i.hasNext ()) {
		SecureHash g = (SecureHash) i.next ();
		NodeId n = (NodeId) node_guid_to_node_id.get (g);
		PublishMsg p = new PublishMsg (n, self_guid, 
			msg.guid, msg.tag, 0, true /* publish */);
		p.set_msg_id (new SHA1Hash ());
		dispatch (p);
	    }
	}
	else if (item instanceof TapestryUnpublishMsg) {
	    TapestryUnpublishMsg msg = (TapestryUnpublishMsg) item;

	    // Unpublish it locally.
	    publish (self_guid, msg.guid, msg.tag, false /* unpublish */);

	    // And then send it to everyone else.
	    Iterator i = node_guid_to_node_id.keySet ().iterator ();
	    while (i.hasNext ()) {
		SecureHash g = (SecureHash) i.next ();
		NodeId n = (NodeId) node_guid_to_node_id.get (g);
		PublishMsg p = new PublishMsg (n, self_guid, 
			msg.guid, msg.tag, 0, false /* unpublish */);
		p.set_msg_id (new SHA1Hash ());
		dispatch (p);
	    }
	}
	else if (item instanceof PublishMsg) {
	    PublishMsg msg = (PublishMsg) item;
	    publish (msg.src, msg.guid, msg.tag, msg.publish);
	}
	else if (item instanceof TapestryRouteMsg) {
	    TapestryRouteMsg msg = (TapestryRouteMsg) item;
	    NodeId n = (NodeId) node_guid_to_node_id.get (msg.peer);
	    RouteMsg r = new RouteMsg (n, self_guid, msg.peer, 
		    false /* not prefix only */, 0 /* hop */, msg);
	    r.set_msg_id (new SHA1Hash ());
	    dispatch (r);
	}
	else if (item instanceof RouteMsg) {
	    RouteMsg msg = (RouteMsg) item;
	    if (! msg.dest.equals (self_guid)) {
		Carp.die (tag + ": got " + msg + " not intended for us.");
	    }

	    QuickSerializable outb = null;
	    try {
		outb = msg.user_data ();
	    }
	    catch (QSException e) {
		Carp.out (tag + ": " + msg + 
			" threw QSException on deserialization." +
			"  Stack trace follows.");
		e.printStackTrace ();
		Carp.die (tag + ": Dying...");
	    }

	    if (outb instanceof TapestryRouteMsg) {
		TapestryRouteMsg rm = (TapestryRouteMsg) outb;
		rm.peer = msg.src;
		rm.inbound = true;
		dispatch (rm);
	    }
	    else if (outb instanceof TapestryPrefixRouteMsg) {
		TapestryPrefixRouteMsg prm = (TapestryPrefixRouteMsg) outb;
		prm.peer = msg.src;
		prm.inbound = true;
		dispatch (prm);
	    }
	    else {
		Carp.die (tag + ": got " + msg + 
			" with unknown payload " + outb);
	    }
	}
	else if (item instanceof TapestryLocateMsg) {
	    TapestryLocateMsg msg = (TapestryLocateMsg) item;

	    LinkedList l = (LinkedList) guid_to_publish_info.get (msg.guid);
	    boolean stop = false;

	    if (l != null) {
		LinkedList dispatches = new LinkedList ();
		Iterator i = l.iterator ();
		while (i.hasNext ()) {
		    PublishInfo pi = (PublishInfo) i.next ();

		    if (DEBUG) Debug.printtagln (tag, "query: pi=" + pi);

		    SecureHash [] matches = null;
		    if (msg.query != null) {
			TapestryQueryResult qr = msg.query.query (
				self_guid, pi.src, pi.tag, msg.query_state);
			if (qr instanceof TapestryQueryResultMatch) {
			    TapestryQueryResultMatch m = 
				(TapestryQueryResultMatch) qr;
			    matches = m.matches ();
			    stop = m.stop ();
			} 
			else {
			    stop = false;
			}
			if (DEBUG) Debug.printtagln (tag, "query: qr=" 
				+ qr + ", stop=" + stop);
		    }  
		    else {
			if (DEBUG) Debug.printtagln (tag, "query: no query");
			matches = new SecureHash [1];
			matches [0] = pi.src;
			stop = true;
		    }

		    if (matches != null) {
			boolean foundmatch = false;
			for (int j = 0; j < matches.length; ++j) {

			    NodeId next = (NodeId)
				node_guid_to_node_id.get (matches [j]);

			    if (next == self_node_id) {
				foundmatch = true;

				if (DEBUG) Debug.printtagln (tag, 
					"query: local match=" + matches[j]);

				QuickSerializable event = msg;

				TapestryLocateMsg r = (TapestryLocateMsg) 
				    QSClone.qs_clone (msg);
				r.peer = self_guid;
				r.inbound = true;
				r.guid = msg.guid;
				r.query = null;
				r.query_state = null;
				r.hopCount = r.TTL = 0;
				dispatches.addLast (event);
			    } else {
				if (DEBUG) Debug.printtagln (tag, 
					"query: remote match=" + matches[j] +
					" at " + next);

				RouteObjDestMsg outb = new RouteObjDestMsg (
					next, self_guid, pi.src, 
					msg.guid, 0, msg);
				outb.set_msg_id (new SHA1Hash ());
				dispatches.addLast (outb);
			    }
			}
		    }

		    if (stop) break;
		}

		if (! stop) {
		    if (DEBUG) Debug.printtagln (tag, "query: failure");
		    dispatch (new TapestryLocateFailure 
			    (self_guid, msg.id, msg.guid, self_node_id, 0));
		}

		while (! dispatches.isEmpty ()) {
		    QueueElementIF d = 
			(QueueElementIF) dispatches.removeFirst ();
		    dispatch (d);
		}
	    }
	}
	else if (item instanceof RouteObjDestMsg) {
	    RouteObjDestMsg msg = (RouteObjDestMsg) item;
	    TapestryLocateMsg outb = null;
	    try {
		outb = (TapestryLocateMsg) msg.user_data ();
	    }
	    catch (QSException e) {
		Carp.out (tag + ": message " + msg + 
			" caused exception " + e + " on decode." +
			"  Stack trace follows.");
		e.printStackTrace (System.err);
		Carp.die ("Dying...");
	    }
	    outb.peer = msg.src;
	    outb.inbound = true;
	    outb.guid = msg.guid;
	    dispatch (outb);
	}
	else {
	    BUG (tag + ": unknown event " + item);
	}
    }

    public void handleEvents(QueueElementIF items[]) 
    throws EventHandlerException {
	for(int i=0; i<items.length; i++)
	    handleEvent(items[i]);
    }

    protected void publish (SecureHash src, SecureHash guid, 
	    TapestryTag tag, boolean publish) {

	LinkedList ll = (LinkedList) guid_to_publish_info.get (guid);
	if (ll == null) {
	    if (publish) {
		ll = new LinkedList ();
		guid_to_publish_info.put (guid, ll);
	    }
	    else {
		Carp.die (tag + ": got unpublish for unknown guid " +
			guid + ".  Dying...");
	    }
	}
	PublishInfo pi = new PublishInfo (src, tag, null);
	if (publish) {
	    ll.addLast (pi);
	}
	else {
	    if (! ll.contains (pi)) {
		Carp.die (tag + ": got unpublish for unknown guid " +
			guid + ", tag " + tag + ", and src " + src +
			".  Dying...");
	    }

	    ll.remove (pi);
	}
    }
}

