/**
 * ByteArrayOutputBuffer.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.util;

/**
 * Writes objects into a user-provided byte array.
 *
 * @author  Sean C. Rhea
 * @version $Id: ByteArrayOutputBuffer.java,v 1.4 2003/05/28 05:45:33 srhea Exp $
 */
public class ByteArrayOutputBuffer extends OutputBufferImpl
{
    private byte [] _data;
    private int _offset;
    private int _max_offset;

    public ByteArrayOutputBuffer (byte [] data) {
	this (data, 0, data.length);
    }

    public ByteArrayOutputBuffer (byte [] data, int offset, int length) {
	_data = data; _offset = offset; _max_offset = offset + length - 1;
    }

    public int offset () {
	return _offset;
    }

    public byte [] data () {
	return _data;
    }

    public void add (byte input) {
	if (_offset > _max_offset)
	    throw new AssertionViolatedException ("array full");
	_data [_offset++] = input;
    }

    public void add (byte[] value, int offset, int length) {
	if (_offset + length - 1 > _max_offset)
	    throw new AssertionViolatedException ("array too small");
	System.arraycopy (value, offset, _data, _offset, length);
	_offset += length;
    }

    public void reset (byte [] data) {
	_data = data; _offset = 0; _max_offset = _offset + _data.length - 1;
    }

    public int size () {
	return _offset;
    }
}

