/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

/**
 * A class which is used to register all of the basic OceanStore types.  To
 * ensure that a type can be received over the network or from disk in
 * every OceanStore instance regardless of what stages are instanciated,
 * put an entry for it in this file. 
 * 
 * @author   Sean C. Rhea
 * @version  "$Id: Types.java,v 1.236 2003/04/30 07:20:22 ravenben Exp $"
 * 
 * @see ostore.util.TypeTable
 * @see ostore.util.QuickSerializable
 * @see ostore.util.QSException
 */
public class Types
{
    
    // Associate type codes with class names here.

    static public final String [] types = {
	
 	//  ostore.util types
	"ostore.util.NodeId",
	"ostore.util.NonceAckMsg",
	"ostore.util.QSInt",
	"ostore.util.QSLong",
	"ostore.util.QSString",
	"ostore.util.QSDate",
	"ostore.util.QSByteArray",
	"ostore.util.QSBigInteger",
	"ostore.util.QSList",
	"ostore.util.QSArray",
        "ostore.util.QSBool",
	"ostore.util.QSPair",
        "ostore.util.QSWindow",
        "ostore.util.QSVector",
        "ostore.util.QSTreeSet",
        "ostore.util.QSHashMap",
	"ostore.util.SHA1Hash",
        "ostore.util.StatRecorder",
	"ostore.util.VerifiableBlockTag",
	"ostore.util.VID",

	// ostore.network types
	"ostore.network.IdentityMessage",
        "ostore.network.MyNodeId",


	//  ostore.dataobj types
	"ostore.dataobj.PRCert",

	// ostore.tapestry.impl types
	"ostore.tapestry.impl.StaticAliveMsg",
	"ostore.tapestry.impl.StaticAllNodesMsg",
	"ostore.tapestry.impl.StaticReadyMsg",
	"ostore.tapestry.impl.StaticBeginMsg",
	"ostore.tapestry.impl.StaticPingMsg",
	"ostore.tapestry.impl.StaticPongMsg",
	"ostore.tapestry.impl.StaticPangMsg",
	"ostore.tapestry.impl.StaticPtrMsg",
	"ostore.tapestry.impl.RouteMsg",
	"ostore.tapestry.impl.RoutingTable",
	"ostore.tapestry.impl.RouteEntry",
	"ostore.tapestry.impl.RouteObjRootMsg",
	"ostore.tapestry.impl.RouteObjDestMsg",
	"ostore.tapestry.impl.PublishMsg",
	"ostore.tapestry.impl.BloomWrapper",
	"ostore.tapestry.impl.PingMsg",
	"ostore.tapestry.impl.PongMsg",
	"ostore.tapestry.api.TapestryLocateFailure",
	"ostore.tapestry.impl.PublishInfo",
	"ostore.tapestry.impl.TapestryMessage",
	"ostore.tapestry.impl.OverlayPath",
	"ostore.tapestry.impl.GuidAddr",
	"ostore.tapestry.impl.PeriodicRoutePatch",

	// ostore.tapestry.channel types
	"ostore.tapestry.channel.ChannelACKMsg",
	"ostore.tapestry.channel.ICMPPingRouteMsg",
	"ostore.tapestry.channel.RetransmissionCheckMsg",
	"ostore.tapestry.channel.ICMPPingMsg",
	"ostore.tapestry.channel.ICMPPongMsg",

	// ostore.tapestry.impl dynamic msgs
	"ostore.tapestry.impl.DynamicMapMsg",
	"ostore.tapestry.impl.DynamicNotifyMsg",
	"ostore.tapestry.impl.DynamicDMAckMsg",
	"ostore.tapestry.impl.DynamicDMCastMsg",
	"ostore.tapestry.impl.DynamicMCReqMsg",
	"ostore.tapestry.impl.DynamicObjMapMsg",
	"ostore.tapestry.impl.DynamicObjReqMsg",
	"ostore.tapestry.impl.DynamicPingMsg",
	"ostore.tapestry.impl.DynamicPongMsg",
	"ostore.tapestry.impl.DynamicGatewayMsg",
	"ostore.tapestry.impl.WatchList",
	"ostore.tapestry.impl.DynamicSeedMapMsg",
	"ostore.tapestry.impl.DynamicSeedMsg",
	"ostore.tapestry.impl.DynamicSurrogateMsg",
	"ostore.tapestry.impl.DynamicPtrMsg",
	"ostore.tapestry.impl.DynamicBeginDelete",
	"ostore.tapestry.impl.DynamicEndDelete",
	"ostore.tapestry.impl.RepublishRequestMsg",
	"ostore.tapestry.impl.PartialRepublish",
	"ostore.tapestry.impl.DynamicRepairReqMsg",
	"ostore.tapestry.impl.DynamicRepairRespMsg",

	// ostore.tapestry.impl test msg
	"ostore.tapestry.impl.FullTestMemberMsg",
        "ostore.tapestry.impl.FullTestMsg",
        "ostore.tapestry.impl.FullTestRespMsg",
        "ostore.tapestry.impl.VerifyRespMsg",
	"ostore.tapestry.impl.TestFoundMsg",
	"ostore.tapestry.impl.TestLocateMsg",
	"ostore.tapestry.impl.TestReadyMsg",
	"ostore.tapestry.impl.TestNetworkMsg",
	"ostore.tapestry.impl.TestDeleteMsg",
	"ostore.tapestry.impl.test.MicroRouteMsg",
	"ostore.tapestry.impl.test.MicroBarrierMsg",
	"ostore.tapestry.impl.test.TapestryRouteMsgChannel",
	"ostore.tapestry.impl.test.TapestryLocateMsgChannel",
	"ostore.tapestry.impl.test.TapestryPublishMsgChannel",
	"ostore.tapestry.impl.test.TapestryQueryChannel",
	"ostore.tapestry.impl.test.TapestryQueryStateChannel",
	"ostore.tapestry.impl.test.TapestryTagChannel",
	"ostore.tapestry.impl.DebugMsg",
	"ostore.tapestry.impl.DebugRespMsg",
	"ostore.tapestry.impl.NullTag",
	"ostore.tapestry.impl.NullTagQuery",
	"ostore.tapestry.impl.test.LocalOptimizationTest$Response",
	"ostore.tapestry.impl.test.LocalOptimizationTest$LocateMsg",

	// patchwork types
  	"ostore.tapestry.patchwork.PatchworkModel$Delta",
  	"ostore.tapestry.patchwork.PatchworkModel",
  	"ostore.tapestry.patchwork.PatchworkModel$Observation",
  	"ostore.tapestry.patchwork.PatchworkProbeMsg",
  	"ostore.tapestry.patchwork.PatchworkFaultMsg",
  	"ostore.tapestry.patchwork.PatchworkRevCountMsg",
  	"ostore.tapestry.patchwork.RepairMonitorStage$PtrDeletedMsg",
  	"ostore.tapestry.patchwork.RouterFaultMsg",
  	"ostore.tapestry.patchwork.RepairMonitorStage$PtrDeletedMsg",
  	"ostore.tapestry.patchwork.RepairMonitorStage$DetachMessage",
  	"ostore.tapestry.patchwork.RepairMonitorStage$AttachMessage",
  	"ostore.tapestry.patchwork.RMFoundMsg",
  	"ostore.tapestry.patchwork.RMLocateMsg",
  	"ostore.network.patchwork.InjectFaultMsg",
  	"ostore.network.patchwork.PatchworkTable",
  	"ostore.network.patchwork.PatchworkTable$Delta",
  	"ostore.network.patchwork.PatchworkStats",
  	"ostore.network.patchwork.PatchworkStats$StatDelta",
  	"ostore.network.patchwork.ProbeMsg",

        // ostore.security types
	"ostore.security.QSDHPublicKey",
	"ostore.security.QSRsaPrivateKey",
	"ostore.security.QSRsaPublicKey",
	"ostore.security.QSSha1RsaSignature",
	"ostore.security.SafeBigInt",
	"ostore.security.SignedQS",
	"ostore.security.SignedQSTree",
	"ostore.security.SignedQSTree$Root",
	"ostore.security.SignedQSTreePath",
	"ostore.security.thresh.KeyShare",
	"ostore.security.thresh.SigShare",
	"ostore.security.thresh.ThresholdPublicKey",
	"ostore.security.thresh.ThresholdSignature",
	"ostore.security.thresh.ThresholdSignedQS",
	"ostore.security.thresh.ThresholdSignedQSTree",
	"ostore.security.thresh.Verifier",

	// ostore.bloom types
	"ostore.bloom.UpdateId",
	"ostore.bloom.UpdateMessage",
	"ostore.bloom.TopologyDiscoverQuery",
	"ostore.bloom.TopologyDiscoverResponse",
	"ostore.bloom.SimulatedRead",
	"ostore.bloom.SimulatedReadResp",
        "ostore.bloom.QueryMessage",

  	// ostore.mac types
	"ostore.mac.DHMessage",
	"ostore.mac.STSVerify",
	"ostore.mac.QSHMACMD5",

  	// ostore.replica types
	"ostore.replica.ReplicaTag",
	"ostore.replica.ReplicaQuery",
    };

    //**********************************************************************
    //***************** Don't touch anything below here. *******************
    //**********************************************************************

    static public void registerAllTypes ()
    {
	if (registered) return;
	boolean errors = false;
	for (int i = 0; i < types.length; ++i) {
	    try {
		TypeTable.register_type (types [i]);
	    }
	    catch (Exception e)
	    {
		System.err.println ("Caught a fatal exception initializing " +
			   "type (" + types [i]);
		e.printStackTrace ();
		Carp.die ("Dying...");
		errors = true;
	    }
	}
	if (errors)
	    Carp.die ("Dying...");

	registered = true;
    }

    static private boolean registered = false;

    static
    {
	registerAllTypes();
    }
}

