#!/usr/bin/perl -w
#
# $Id: start_remote.pl,v 1.2 2003/02/28 01:55:52 geels Exp $

# Similar in theory to call_remote.pl, but takes advantage of Tools.pm
# infrastructure to track remote processes and clean up properly.
# You probably should not use this script.

use strict;

BEGIN {
    # Find the directory with Tools.pm.  Should this be easier?
    my $run_dir = "$0";
    if( $run_dir !~ m%^/% ) {	# relative pathname
	$run_dir = $ENV{"PWD"} . "/$run_dir";
    } 
    $run_dir =~ s%/(./)?[^/]*$%%;	# strip off "run-experiment"
    $ENV{"OSTORE_RUN_DIR"} = $run_dir;
}

use lib $ENV{"OSTORE_RUN_DIR"};
use Tools;
use Getopt::Long;
$Getopt::Long::order = $REQUIRE_ORDER;	# Stop processing at user command.

# Use a variable for easy changes (or addition of parameters).
my @ssh_cmd = ("ssh");

my %options;
&GetOptions( \%options, "help|?", "host=s@", "wait-for=s")
    or &usage and exit( -1 );

if( $options{"help"} ) {
    &usage and exit( 0 );
}

# Rest goes into command string
my @command_string = @ARGV;
if( ! @command_string ) {
    &usage and exit( -1 );
}

sub usage {
    print "\nUSAGE:\t start_remote.pl [-host <name>] " .
	"[-wait-for <string>] <command> [args]\n\n";
    print "WARNING: Only use the background option for remote scripts\n";
    print "\t\t that properly report their PID.\n\n";
}

my @machines;
if( $options{"host"} ) {
    @machines = @{$options{"host"}};
} elsif( $ENV{"OSTORE_SERVERS"} ) {
    @machines = split /\s+/, $ENV{"OSTORE_SERVERS"};
} else {
    foreach( 1 .. 42 ) {
	push @machines, "x$_.millennium.berkeley.edu";
    }
}

my @child_logs;
my $total = ($#machines+1);
for( my $i=0; $i < $total; $i++ ) {
    my $machine = $machines[$i];

    print "[" . ($i+1) . "/$total] Calling @command_string on $machine\n";
    # Fork and save.
    my $log = "$tmp_dir/start_remote-$$-$machine.log";
    push @child_logs, $log;
    &fork_remote( $log, $machine, @command_string );
}

if( $options{"wait-for"} ) {
    my $string = $options{"wait-for"};
    print "Waiting for \"$string\"\n";
    foreach (@child_logs) {
	&Tools::wait_for( $_, $string );
    }
}

# Wait for threads to exit, or for user exit.
print "\nOnly daemon threads running.  Press Ctrl-C to exit.\n";

$SIG{INT} = \&user_exit;
if( &returned_zero( @Tools::child_pids ) and &is_clean( @Tools::logs ) ) {
    &exit_all( 0 ); # success!
} else {
    &exit_all( 1 ); # failure!
}
