/**
 * MainFrame.java
 *
 * Copyright (c) 2002 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package Sonar;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import Sonar.GraphDisplayPanel;
import Sonar.Dialog.*;
import Sonar.util.NodeInfo;
import Sonar.event.*;
import Sonar.PacketAnalyzers.PacketAnalyzer;

import ostore.apps.visdemo.QTreeStateMsg;
import ostore.apps.visdemo.QTreeUpMsg;
import ostore.apps.visdemo.QTreeDownMsg;
import ostore.apps.visdemo.VisUp;
import ostore.apps.visdemo.VisDown;

//SANDSTORM
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.NodeId;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.dispatch.Signal;

import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.ProfilableIF;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkFullException;
import seda.sandStorm.internal.SinkProxy;
import seda.sandStorm.api.StagesInitializedSignal;

/**
 * This class, a basic JFC 1.1 based application, has been created completely 
 * by visualCafe, and essentially holds all the GUI information.
 *
 * @author Jason Lee <jlee81@cs.berkeley.edu>
 * @author Kamin Whitehouse <kamin@cs.berkeley.edu> 
 */

public class MainFrame extends javax.swing.JFrame implements EventHandlerIF, ProfilableIF //implements java.io.Serializable 
{
    //public MainFrame() {
    protected void inst() {
	// This code is automatically generated by Visual Cafe when you add
	// components to the visual environment. It instantiates and initializes
	// the components. To modify the code, only use code syntax that matches
	// what Visual Cafe can generate, or Visual Cafe may be unable to back
	// parse your Java file into its visual environment.
	
	//{{INIT_CONTROLS
	setDefaultCloseOperation(javax.swing.JFrame.DO_NOTHING_ON_CLOSE);
	setJMenuBar(MainMenuBar);
	//MainMenuBar.setDoubleBuffered(true);
	setTitle("Sonar: Oceanstore Visualization");
	getContentPane().setLayout(new BorderLayout(0,0));
	setSize(700,500);
	setVisible(false);

	//initialize icons
	newIcon = new ImageIcon("images/new.gif");
	openIcon = new ImageIcon("images/open.gif");
	saveIcon = new ImageIcon("images/save.gif");
	cutIcon = new ImageIcon("images/cut.gif");
	copyIcon = new ImageIcon("images/copy.gif");
	pasteIcon = new ImageIcon("images/paste.gif");
	aboutIcon = new ImageIcon("images/about.gif");
	selectIcon = new ImageIcon("images/palette_select.gif");
	handIcon = new ImageIcon("images/palette_hand.gif");
	zoomIcon = new ImageIcon("images/palette_zoom.gif");
	fitIcon = new ImageIcon("images/palette_magnify.gif");	

	saveFileDialog.setMode(FileDialog.SAVE);
	saveFileDialog.setTitle("Save");
	//$$ saveFileDialog.move(24,336);
	openFileDialog.setMode(FileDialog.LOAD);
	openFileDialog.setTitle("Open");
	//$$ openFileDialog.move(0,336);
	MainPanel.setLayout(new FlowLayout(FlowLayout.LEFT,0,0));
	getContentPane().add(BorderLayout.NORTH, MainPanel);
	MainToolBar.setAlignmentY(0.222222F);
	//MainToolBar.setDoubleBuffered(true);
	MainPanel.add(MainToolBar);
	newButton.setDefaultCapable(false);
	newButton.setToolTipText("Create a new document");
	newButton.setMnemonic((int)'N');
	MainToolBar.add(newButton);
	openButton.setDefaultCapable(false);
	openButton.setToolTipText("Open an existing document");
	openButton.setMnemonic((int)'O');
	MainToolBar.add(openButton);
	saveButton.setDefaultCapable(false);
	saveButton.setToolTipText("Save the active document");
	saveButton.setMnemonic((int)'S');
	MainToolBar.add(saveButton);
	MainToolBar.addSeparator();
	//MainToolBar.add(JToolBarSeparator1);
	cutButton.setDefaultCapable(false);
	cutButton.setToolTipText("Cut the selection and put it on the Clipboard");
	cutButton.setMnemonic((int)'T');
	MainToolBar.add(cutButton);
	copyButton.setDefaultCapable(false);
	copyButton.setToolTipText("Copy the selection and put it on the Clipboard");
	copyButton.setMnemonic((int)'C');
	MainToolBar.add(copyButton);
	pasteButton.setDefaultCapable(false);
	pasteButton.setToolTipText("Insert Clipboard contents");
	pasteButton.setMnemonic((int)'P');
	MainToolBar.add(pasteButton);
	MainToolBar.addSeparator();
	//MainToolBar.add(JToolBarSeparator2);
	aboutButton.setDefaultCapable(false);
	aboutButton.setToolTipText("Display program information, version number and copyright");
	aboutButton.setMnemonic((int)'A');
	MainToolBar.add(aboutButton);
	MainToolBar.addSeparator();
	//MainToolBar.add(JToolBarSeparator3);
	selectButton.setSelected(true);
	selectButton.setToolTipText("Use this to select and move nodes");
	selectButton.setMnemonic((int)'P');
	MainToolBar.add(selectButton);
	//handButton.setToolTipText("Use this to scroll with the left mouse button");
	handButton.setToolTipText("Use this group nodes with the left mouse button");
	handButton.setMnemonic((int)'H');
	MainToolBar.add(handButton);
	zoomButton.setToolTipText("Use this to select a region and zoom with the left mouse button");
	zoomButton.setMnemonic((int)'Z');
	MainToolBar.add(zoomButton);
//	dragButtonGroup.add(selectButton);
//	dragButtonGroup.add(handButton);
//	dragButtonGroup.add(zoomButton);
//	dragButtonGroup.setSelected(selectButton, true);
	MainToolBar.addSeparator();
	//MainToolBar.add(JToolBarSeparator4);
	fitButton.setDefaultCapable(false);
	fitButton.setToolTipText("Fit all nodes onto the screen");
	fitButton.setMnemonic((int)'F');
	MainToolBar.add(fitButton);
	MainToolBar.addSeparator();
	//MainToolBar.add(JToolBarSeparator5);
	JSlider1.setMinimum(1);
	JSlider1.setMaximum(5);
	JSlider1.setToolTipText("Slide this to zoom");
	//JSlider1.setBorder(bevelBorder1);
	JSlider1.setValue(1);
	MainToolBar.add(JSlider1);
	//$$ bevelBorder1.move(0,115);
	JLabel3.setText("1.0");
	JLabel3.setToolTipText("This is the current zoom level");
	MainToolBar.add(JLabel3);
	JLabel3.setFont(new Font("Dialog", Font.BOLD, 16));
	MainToolBar.addSeparator();
	//MainToolBar.add(JToolBarSeparator6);
	JLabel4.setText("jlabel");
	JLabel4.setToolTipText("This is the current mouse position");
	MainToolBar.add(JLabel4);
	JLabel4.setSize(51,27);

	getContentPane().add(BorderLayout.CENTER, MainScrollPane);
	MainScrollPane.setOpaque(true);
	
	//GRAPHDISPLAYPANEL 
	MainScrollPane.setViewportView(graphDisplayPanel);
	MainScrollPane.getViewport().add(graphDisplayPanel);
	graphDisplayPanel.setBounds(0,0,430,270);
	graphDisplayPanel.setLayout(null);
	graphDisplayPanel.setAutoscrolls(true);
	//GraphDisplayPanel.setBounds(0,0,100,100);
	//GraphDisplayPanel.setPreferredSize(new Dimension(100,100));
	MainScrollPane.getViewport().add(graphDisplayPanel);
	
	//$$ MainMenuBar.move(168,312);
	fileMenu.setText("File");
	fileMenu.setActionCommand("File");
	fileMenu.setMnemonic((int)'F');
	MainMenuBar.add(fileMenu);
	newItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_N, java.awt.Event.CTRL_MASK));
	newItem.setText("New");
	newItem.setActionCommand("New");
	newItem.setMnemonic((int)'N');
	fileMenu.add(newItem);
	openItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_O, java.awt.Event.CTRL_MASK));
	openItem.setText("Open...");
	openItem.setActionCommand("Open...");
	openItem.setMnemonic((int)'O');
	fileMenu.add(openItem);
	saveItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_S, java.awt.Event.CTRL_MASK));
	saveItem.setText("Save");
	saveItem.setActionCommand("Save");
	saveItem.setMnemonic((int)'S');
	fileMenu.add(saveItem);
	saveAsItem.setText("Save As...");
	saveAsItem.setActionCommand("Save As...");
	saveAsItem.setMnemonic((int)'A');
	fileMenu.add(saveAsItem);
	fileMenu.add(JSeparator1);
	exitItem.setText("Exit");
	exitItem.setActionCommand("Exit");
	exitItem.setMnemonic((int)'X');
	fileMenu.add(exitItem);
	editMenu.setText("Edit");
	editMenu.setActionCommand("Edit");
	editMenu.setMnemonic((int)'E');
	MainMenuBar.add(editMenu);
	JMenuItem2.setText("Sonar Options");
	editMenu.add(JMenuItem2);
	editMenu.add(JSeparator2);
	cutItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_X, java.awt.Event.CTRL_MASK));
	cutItem.setText("Cut");
	cutItem.setActionCommand("Cut");
	cutItem.setMnemonic((int)'T');
	editMenu.add(cutItem);
	copyItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_C, java.awt.Event.CTRL_MASK));
	copyItem.setText("Copy");
	copyItem.setActionCommand("Copy");
	copyItem.setMnemonic((int)'C');
	editMenu.add(copyItem);
	pasteItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_V, java.awt.Event.CTRL_MASK));
	pasteItem.setText("Paste");
	pasteItem.setActionCommand("Paste");
	pasteItem.setMnemonic((int)'P');
	editMenu.add(pasteItem);
	DisplayMenu.setText("Display");
	DisplayMenu.setActionCommand("Display");
	MainMenuBar.add(DisplayMenu);
	JCheckBoxMenuItem8.setSelected(true);
	JCheckBoxMenuItem8.setToolTipText("Always keep the entire network visible (may cause jumping)");
	JCheckBoxMenuItem8.setText("Fit Network to Screen Automatically");
	JCheckBoxMenuItem8.setActionCommand("Fit Network to Screen Automatically");
	DisplayMenu.add(JCheckBoxMenuItem8);
	JMenuItem7.setText("Fit Network to Screen Now");
	JMenuItem7.setActionCommand("Fit Network to Screen");
	DisplayMenu.add(JMenuItem7);
	DisplayMenu.add(JSeparator4);
	JCheckBoxMenuItem7.setSelected(true);
	JCheckBoxMenuItem7.setToolTipText("Refreshing the screen slows processing significantly.  Use \"Refresh Screen Now\" to refresh manually.");
	JCheckBoxMenuItem7.setText("Refresh Screen Automatically");
	JCheckBoxMenuItem7.setActionCommand("Refresh Screen Automatically");
	DisplayMenu.add(JCheckBoxMenuItem7);
	JMenuItem8.setToolTipText("Use this to refresh the screen manually");
	JMenuItem8.setText("Refresh Screen Now");
	JMenuItem8.setActionCommand("Refresh Screen Now");
	DisplayMenu.add(JMenuItem8);
	NetworkMenu.setText("Network");
	NetworkMenu.setActionCommand("Network");
	MainMenuBar.add(NetworkMenu);
	PacketAnalyzersMenu.setText("Packet Analyzers");
	PacketAnalyzersMenu.setActionCommand("Packet Analyzers");
	MainMenuBar.add(PacketAnalyzersMenu);

	/*
	//Removed Actions from the menu
	JMenu1.setText("Actions");
	JMenu1.setActionCommand("Actions");
	MainMenuBar.add(JMenu1);
	JMenuItem5.setText("Inject Packet");
	JMenuItem5.setActionCommand("Inject Packet");
	JMenu1.add(JMenuItem5);
	JMenu1.add(JSeparator13);
	JMenuItem6.setText("Set Transmission Strength");
	JMenuItem6.setActionCommand("Set Transmission Strength");
	JMenu1.add(JMenuItem6);
	*/

	helpMenu.setText("Help");
	helpMenu.setActionCommand("Help");
	helpMenu.setMnemonic((int)'H');
	MainMenuBar.add(helpMenu);
	aboutItem.setText("About...");
	aboutItem.setActionCommand("About...");
	aboutItem.setMnemonic((int)'A');
	helpMenu.add(aboutItem);
	//$$ JMenuBar1.move(168,312);
	saveButton.setIcon(saveIcon);
	newButton.setIcon(newIcon);
	cutItem.setIcon(cutIcon);
	newItem.setIcon(newIcon);
	openButton.setIcon(openIcon);
	openItem.setIcon(openIcon);
	aboutButton.setIcon(aboutIcon);
	selectButton.setIcon(selectIcon);
	handButton.setIcon(handIcon);
	zoomButton.setIcon(zoomIcon);
	fitButton.setIcon(fitIcon);
	pasteButton.setIcon(pasteIcon);
	saveItem.setIcon(saveIcon);
	pasteItem.setIcon(pasteIcon);
	copyItem.setIcon(copyIcon);
	cutButton.setIcon(cutIcon);
	copyButton.setIcon(copyIcon);
	aboutItem.setIcon(aboutIcon);
	//}}

	//{{INIT_MENUS
	//}}

	//{{REGISTER_LISTENERS
	SymWindow aSymWindow = new SymWindow();
	this.addWindowListener(aSymWindow);
	SymAction lSymAction = new SymAction();
	openItem.addActionListener(lSymAction);
	saveItem.addActionListener(lSymAction);
	exitItem.addActionListener(lSymAction);
	aboutItem.addActionListener(lSymAction);
	openButton.addActionListener(lSymAction);
	saveButton.addActionListener(lSymAction);
	aboutButton.addActionListener(lSymAction);
	selectButton.addActionListener(lSymAction);
	handButton.addActionListener(lSymAction);
	zoomButton.addActionListener(lSymAction);
	fitButton.addActionListener(lSymAction);
	//	SymFocus aSymFocus = new SymFocus();
	//	MainMenuBar.addFocusListener(aSymFocus);
	JMenuItem7.addActionListener(lSymAction);
	SymItem lSymItem = new SymItem();
	JCheckBoxMenuItem7.addItemListener(lSymItem);
	JMenuItem8.addActionListener(lSymAction);
	JCheckBoxMenuItem8.addItemListener(lSymItem);
	SymChange lSymChange = new SymChange();
	JSlider1.addChangeListener(lSymChange);
	JMenuItem2.addActionListener(lSymAction);
	//}}
    }	

    /**
     * Creates a new instance of JFrame1 with the given title.
     * @param sTitle the title for the new frame.
     * @see #JFrame1()
     */
    /*
    public MainFrame(String sTitle) {
	this();
	setTitle(sTitle);
    }
    */

    /**
     * The entry point for this application.
     * Sets the Look and Feel to the System Look and Feel.
     * Creates a new JFrame1 and makes it visible.
     */	
    //static public void main(String args[])
    //{
    //	try {
    // Add the following code if you want the Look and Feel
    // to be set to the Look and Feel of the native system.
    /*
       try {
       UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
       } 
       catch (Exception e) { 
       }
*/
    //Create a new instance of our application's frame, and make it visible.    
    //		(new MainFrame()).setVisible(true);
    //	} 
    //	catch (Throwable t) {
    //		t.printStackTrace();
    //		//Ensure the application exits with an error condition.
    //		System.exit(1);
    //	}
    //}

    /**
     * Notifies this component that it has been added to a container
     * This method should be called by <code>Container.add</code>, and 
     * not by user code directly.
     * Overridden here to adjust the size of the frame if needed.
     * @see java.awt.Container#removeNotify
     */
    public void addNotify() {
	// Record the size of the window prior to calling parents addNotify.
	Dimension size = getSize();
	super.addNotify();
	if (frameSizeAdjusted)
		return;
	frameSizeAdjusted = true;
		
	// Adjust size of frame according to the insets and menu bar
	javax.swing.JMenuBar menuBar = getRootPane().getJMenuBar();
	int menuBarHeight = 0;
	if (menuBar != null)
	    menuBarHeight = menuBar.getPreferredSize().height;
	Insets insets = getInsets();
	setSize(insets.left + insets.right + size.width, insets.top + insets.bottom + size.height + menuBarHeight);
    }

    // Used by addNotify
    boolean frameSizeAdjusted = false;

    //{{DECLARE_CONTROLS
    javax.swing.ImageIcon newIcon = null; 
    javax.swing.ImageIcon openIcon = null; 
    javax.swing.ImageIcon saveIcon = null; 
    javax.swing.ImageIcon cutIcon = null;
    javax.swing.ImageIcon copyIcon = null; 
    javax.swing.ImageIcon pasteIcon = null; 
    javax.swing.ImageIcon aboutIcon = null; 
    javax.swing.ImageIcon selectIcon = null; 
    javax.swing.ImageIcon handIcon = null; 
    javax.swing.ImageIcon zoomIcon = null; 
    javax.swing.ImageIcon fitIcon = null;

    java.awt.FileDialog saveFileDialog = new java.awt.FileDialog(this);
    java.awt.FileDialog openFileDialog = new java.awt.FileDialog(this);
    javax.swing.JPanel MainPanel = new javax.swing.JPanel();
    javax.swing.JToolBar MainToolBar = new javax.swing.JToolBar();
    javax.swing.JButton newButton = new javax.swing.JButton();
    javax.swing.JButton openButton = new javax.swing.JButton();
    javax.swing.JButton saveButton = new javax.swing.JButton();
    //com.symantec.itools.javax.swing.JToolBarSeparator JToolBarSeparator1 = new com.symantec.itools.javax.swing.JToolBarSeparator();
    javax.swing.JButton cutButton = new javax.swing.JButton();
    javax.swing.JButton copyButton = new javax.swing.JButton();
    javax.swing.JButton pasteButton = new javax.swing.JButton();
    //com.symantec.itools.javax.swing.JToolBarSeparator JToolBarSeparator2 = new com.symantec.itools.javax.swing.JToolBarSeparator();
    javax.swing.JButton aboutButton = new javax.swing.JButton();
    //com.symantec.itools.javax.swing.JToolBarSeparator JToolBarSeparator3 = new com.symantec.itools.javax.swing.JToolBarSeparator();
    javax.swing.JToggleButton selectButton = new javax.swing.JToggleButton();
    javax.swing.JToggleButton handButton = new javax.swing.JToggleButton();
    javax.swing.JToggleButton zoomButton = new javax.swing.JToggleButton();
    //com.symantec.itools.javax.swing.JToolBarSeparator JToolBarSeparator4 = new com.symantec.itools.javax.swing.JToolBarSeparator();
    javax.swing.JButton fitButton = new javax.swing.JButton();
    //com.symantec.itools.javax.swing.JToolBarSeparator JToolBarSeparator5 = new com.symantec.itools.javax.swing.JToolBarSeparator();
    javax.swing.JSlider JSlider1 = new javax.swing.JSlider();
    javax.swing.JLabel JLabel3 = new javax.swing.JLabel();
    //com.symantec.itools.javax.swing.JToolBarSeparator JToolBarSeparator6 = new com.symantec.itools.javax.swing.JToolBarSeparator();
    javax.swing.JLabel JLabel4 = new javax.swing.JLabel();
    javax.swing.JScrollPane MainScrollPane = new javax.swing.JScrollPane();

    //GRAPHDISPLAYPANEL
    Sonar.GraphDisplayPanel graphDisplayPanel;
    javax.swing.JMenuBar MainMenuBar = new javax.swing.JMenuBar();
    javax.swing.JMenu fileMenu = new javax.swing.JMenu();
    javax.swing.JMenuItem newItem = new javax.swing.JMenuItem();
    javax.swing.JMenuItem openItem = new javax.swing.JMenuItem();
    javax.swing.JMenuItem saveItem = new javax.swing.JMenuItem();
    javax.swing.JMenuItem saveAsItem = new javax.swing.JMenuItem();
    javax.swing.JSeparator JSeparator1 = new javax.swing.JSeparator();
    javax.swing.JMenuItem exitItem = new javax.swing.JMenuItem();
    javax.swing.JMenu editMenu = new javax.swing.JMenu();
    javax.swing.JMenuItem JMenuItem2 = new javax.swing.JMenuItem();
    javax.swing.JSeparator JSeparator2 = new javax.swing.JSeparator();
    javax.swing.JMenuItem cutItem = new javax.swing.JMenuItem();
    javax.swing.JMenuItem copyItem = new javax.swing.JMenuItem();
    javax.swing.JMenuItem pasteItem = new javax.swing.JMenuItem();
    public javax.swing.JMenu DisplayMenu = new javax.swing.JMenu();
    javax.swing.JCheckBoxMenuItem JCheckBoxMenuItem8 = new javax.swing.JCheckBoxMenuItem();
    javax.swing.JMenuItem JMenuItem7 = new javax.swing.JMenuItem();
    javax.swing.JSeparator JSeparator4 = new javax.swing.JSeparator();
    javax.swing.JCheckBoxMenuItem JCheckBoxMenuItem7 = new javax.swing.JCheckBoxMenuItem();
    javax.swing.JMenuItem JMenuItem8 = new javax.swing.JMenuItem();
    public javax.swing.JMenu NetworkMenu = new javax.swing.JMenu();
    public javax.swing.JMenu PacketAnalyzersMenu = new javax.swing.JMenu();
    javax.swing.JMenu JMenu1 = new javax.swing.JMenu();
    javax.swing.JMenuItem JMenuItem5 = new javax.swing.JMenuItem();
    javax.swing.JSeparator JSeparator13 = new javax.swing.JSeparator();
    javax.swing.JMenuItem JMenuItem6 = new javax.swing.JMenuItem();
    javax.swing.JMenu helpMenu = new javax.swing.JMenu();
    //javax.swing.JMenuItem JMenuItem1 = new javax.swing.JMenuItem();
    //javax.swing.JSeparator JSeparator3 = new javax.swing.JSeparator();
    javax.swing.JMenuItem aboutItem = new javax.swing.JMenuItem();
    //com.symantec.itools.javax.swing.borders.BevelBorder bevelBorder1 = new com.symantec.itools.javax.swing.borders.BevelBorder();
    //}}
    //{{DECLARE_MENUS
    //}}

    void exitApplication()
    {
	final String tag = class_tag + ".exitApplication";
	/*
	try {
	// Beep
	//Toolkit.getDefaultToolkit().beep();
	// Show a confirmation dialog
	int reply = JOptionPane.showConfirmDialog(this, 
	"Do you really want to exit?", "SURGE - Exit",
	JOptionPane.YES_NO_OPTION,
	JOptionPane.QUESTION_MESSAGE);
	// If the confirmation was affirmative, handle exiting.
	if (reply == JOptionPane.YES_OPTION) {
	this.setVisible(false);    // hide the Frame
	this.dispose();            // free the system resources
	System.exit(0);            // close the application
	}
	*/

	//TODO 
	NodeInfo node_info = null;
	/*
	for(Enumeration nodes = graphDisplayPanel.displayManager.getProprietaryNodeInfo().elements();
		nodes.hasMoreElements();) {
	    node_info = (NodeInfo) nodes.nextElement();
	    try {
		_classifier.dispatch(new QTreeStateMsg(node_info.getNodeId(), false, QTreeStateMsg.CLOSE, _self_node_id, 0));
	    } catch (SinkFullException e) {
		e.printStackTrace();
	    }
	}*/
	if(_gateway_node_id != null) {
	    try {
		_classifier.dispatch(new QTreeStateMsg(_gateway_node_id, false, QTreeStateMsg.CLOSE, _self_node_id, 0));
	    } catch (SinkFullException e) {
		e.printStackTrace();
	    }
	    if(DEBUG) Debug.printtagln(tag, "sent a close connection message to gateway node");
	} 
	//TODO send close connection to all other connected nodes
	
	//wait for close connection ack's 
	//Debug.printtagln(tag, "waiting for acks...");

	try {
	    Thread.sleep(3000); 
	} catch (Exception e) {
	    e.printStackTrace();
	}

	try {
	    this.setVisible(false); // hide the frame
	    this.dispose();		// free the system resources
	    System.exit(0);		// close the application
	} catch (Exception e) {
	}
    }

    class SymWindow extends java.awt.event.WindowAdapter
    {
	public void windowClosing(java.awt.event.WindowEvent event)
	{
	    Object object = event.getSource();
	    if (object == MainFrame.this)
		MainFrame_windowClosing(event);
	}
    }

    void MainFrame_windowClosing(java.awt.event.WindowEvent event)
    {
	// to do: code goes here.
	MainFrame_windowClosing_Interaction1(event);
    }

    void MainFrame_windowClosing_Interaction1(java.awt.event.WindowEvent
	    event) {
	try {
	    this.exitApplication();
	} catch (Exception e) {
	}
    }

    class SymAction implements java.awt.event.ActionListener
    {
	public void actionPerformed(java.awt.event.ActionEvent event)
	{
	    Object object = event.getSource();
	    if (object == openItem)
		openItem_actionPerformed(event);
	    else if (object == saveItem)
		saveItem_actionPerformed(event);
	    else if (object == exitItem)
		exitItem_actionPerformed(event);
	    else if (object == aboutItem)
		aboutItem_actionPerformed(event);
	    else if (object == openButton)
		openButton_actionPerformed(event);
	    else if (object == saveButton)
		saveButton_actionPerformed(event);
	    else if (object == aboutButton)
		aboutButton_actionPerformed(event);
	    else if (object == selectButton)
		selectButton_actionPerformed(event);
	    else if (object == handButton)
		handButton_actionPerformed(event);
	    else if (object == zoomButton)
		zoomButton_actionPerformed(event);
	    else if (object == fitButton)
		fitButton_actionPerformed(event);
	    if (object == JMenuItem7)
		JMenuItem7_actionPerformed(event);
	    if (object == JMenuItem8)
		JMenuItem8_actionPerformed(event);
	    else if (object == JMenuItem2)
		JMenuItem2_actionPerformed(event);
	}
    }

    void openItem_actionPerformed(java.awt.event.ActionEvent event)
    {
	// to do: code goes here.
	openItem_actionPerformed_Interaction1(event);
    }

    void openItem_actionPerformed_Interaction1(java.awt.event.ActionEvent event) {
	try {
	    // openFileDialog Show the FileDialog
	    openFileDialog.setVisible(true);
	} catch (Exception e) {
	}
    }

    void saveItem_actionPerformed(java.awt.event.ActionEvent event)
    {
	// to do: code goes here.

	saveItem_actionPerformed_Interaction1(event);
    }

    void saveItem_actionPerformed_Interaction1(java.awt.event.ActionEvent event) {
	try {
	    // saveFileDialog Show the FileDialog
	    saveFileDialog.setVisible(true);
	} catch (Exception e) {
	}
    }

    void exitItem_actionPerformed(java.awt.event.ActionEvent event)
    {
	// to do: code goes here.
			 
	exitItem_actionPerformed_Interaction1(event);
    }

    void exitItem_actionPerformed_Interaction1(java.awt.event.ActionEvent event) {
	try {
	    this.exitApplication();
	} catch (Exception e) {
	}
    }

    void aboutItem_actionPerformed(java.awt.event.ActionEvent event)
    {
	// to do: code goes here.
			 
	aboutItem_actionPerformed_Interaction1(event);
    }

    void openButton_actionPerformed(java.awt.event.ActionEvent event)
    {
	// to do: code goes here.
			 
	openButton_actionPerformed_Interaction1(event);
    }

    void openButton_actionPerformed_Interaction1(java.awt.event.ActionEvent event) {
	try {
	    // openFileDialog Show the FileDialog
	    openFileDialog.setVisible(true);
	} catch (Exception e) {
	}
    }

    void saveButton_actionPerformed(java.awt.event.ActionEvent event)
    {
	// to do: code goes here.
			 
	saveButton_actionPerformed_Interaction1(event);
    }

    void saveButton_actionPerformed_Interaction1(java.awt.event.ActionEvent event) {
	try {
	    // saveFileDialog Show the FileDialog
	    saveFileDialog.setVisible(true);
	} catch (Exception e) {
	}
    }

    void aboutButton_actionPerformed(java.awt.event.ActionEvent event)
    {
	try {
	    // JAboutDialog Create and show as modal
	    {
		JAboutDialog JAboutDialog1 = new JAboutDialog(this);
		JAboutDialog1.setModal(true);
		JAboutDialog1.show();
	    }
	} catch (java.lang.Exception e) {
	}
    }

    void selectButton_actionPerformed(java.awt.event.ActionEvent event)
    {
	// to do: code goes here.
			 
	selectButton_actionPerformed_Interaction1(event);
    }

    void selectButton_actionPerformed_Interaction1(java.awt.event.ActionEvent event) {
	try {
	    // JAboutDialog Create with owner and show as modal
	    {
		selectButton.setSelected(true);
		handButton.setSelected(false);
		zoomButton.setSelected(false);
		
		graphDisplayPanel.displayManager.setSelectMode(true);
		graphDisplayPanel.displayManager.setHandMode(true);
		graphDisplayPanel.displayManager.setZoomMode(true);
		
	    }
	} catch (Exception e) {
	}
    }

    void handButton_actionPerformed(java.awt.event.ActionEvent event)
    {
	// to do: code goes here.
			 
	handButton_actionPerformed_Interaction1(event);
    }

    void handButton_actionPerformed_Interaction1(java.awt.event.ActionEvent event) {
	try {
	    // JAboutDialog Create with owner and show as modal
	    {
		selectButton.setSelected(false);
		handButton.setSelected(true);
		zoomButton.setSelected(false);
		 
		graphDisplayPanel.displayManager.setSelectMode(false);
		graphDisplayPanel.displayManager.setHandMode(true);
		graphDisplayPanel.displayManager.setZoomMode(false);
		
	    }
	} catch (Exception e) {
	}
    }

    void zoomButton_actionPerformed(java.awt.event.ActionEvent event)
    {
	// to do: code goes here.
			 
	zoomButton_actionPerformed_Interaction1(event);
    }

    void zoomButton_actionPerformed_Interaction1(java.awt.event.ActionEvent event) {
	try {
	    // JAboutDialog Create with owner and show as modal
	    {
		selectButton.setSelected(false);
		handButton.setSelected(false);
		zoomButton.setSelected(true);
		 
		graphDisplayPanel.displayManager.setSelectMode(false);
		graphDisplayPanel.displayManager.setHandMode(false);
		graphDisplayPanel.displayManager.setZoomMode(true);
		
	    }
	} catch (Exception e) {
	}
    }

    void fitButton_actionPerformed(java.awt.event.ActionEvent event)
    {
	// to do: code goes here.
			 
	fitButton_actionPerformed_Interaction1(event);
    }

    void fitButton_actionPerformed_Interaction1(java.awt.event.ActionEvent event) {
	try {
	    // JAboutDialog Create with owner and show as modal
	    {
		getGraphDisplayPanel().fitToScreen();
	    }
	} catch (Exception e) {
	}
    }

    /*class SymFocus extends java.awt.event.FocusAdapter
      {
      public void focusGained(java.awt.event.FocusEvent event)
      {
      Object object = event.getSource();
      if (object == MainMenuBar)
      MainMenuBar_focusGained(event);
      }
      }

      void MainMenuBar_focusGained(java.awt.event.FocusEvent event)
      {
      // to do: code goes here.
			 
      MainMenuBar_focusGained_Interaction1(event);
      }

      void MainMenuBar_focusGained_Interaction1(java.awt.event.FocusEvent event)
      {
      try {
      MainClass.suspend();
      } catch (java.lang.Exception e) {
      }
      }*/
	
    //the following code was written by kamin
    public GraphDisplayPanel getGraphDisplayPanel()
    {
	return graphDisplayPanel;
    }

    public javax.swing.JScrollPane getMainScrollPane()
    {
	return MainScrollPane;
    } 

    public javax.swing.JLabel getCoordLabel()
    {
	return JLabel4;
    } 

    void JMenuItem7_actionPerformed(java.awt.event.ActionEvent event)
    {
	// to do: code goes here.
			 
	JMenuItem7_actionPerformed_Interaction1(event);
    }


    class SymItem implements java.awt.event.ItemListener
    {
	public void itemStateChanged(java.awt.event.ItemEvent event)
	{
	    Object object = event.getSource();
	    if (object == JCheckBoxMenuItem7)
		JCheckBoxMenuItem7_itemStateChanged(event);
	    if (object == JCheckBoxMenuItem8)
		JCheckBoxMenuItem8_itemStateChanged(event);
	}
    }

    void JCheckBoxMenuItem7_itemStateChanged(java.awt.event.ItemEvent event)
    {
	// to do: code goes here.
			 
	JCheckBoxMenuItem7_itemStateChanged_Interaction1(event);
    }

    void JCheckBoxMenuItem7_itemStateChanged_Interaction1(java.awt.event.ItemEvent event)
    {
	try {
	    if(event.getStateChange() == ItemEvent.SELECTED)
		{
		    //TODO
		    //replace with handle alarms
		    //graphDisplayPanel.displayManager.start();
		}
	    else
		{
		    //TODO
		    //graphDisplayPanel.displayManager.stop();
		}
	} catch (java.lang.Exception e) {
	}          
    }

    void JMenuItem8_actionPerformed(java.awt.event.ActionEvent event)
    {
	// to do: code goes here.
			 
	JMenuItem8_actionPerformed_Interaction1(event);
    }

    void JMenuItem8_actionPerformed_Interaction1(java.awt.event.ActionEvent event)
    {
	//try {
	//graphDisplayPanel.displayManager.refreshScreenNow();
	paint(getGraphics()); 
	//} catch (java.lang.Exception e) {}
    }

    void JCheckBoxMenuItem8_itemStateChanged(java.awt.event.ItemEvent event)
    {
	// to do: code goes here.
			 
	JCheckBoxMenuItem8_itemStateChanged_Interaction1(event);
    }

    void JCheckBoxMenuItem8_itemStateChanged_Interaction1(java.awt.event.ItemEvent event)
    {
	try {
	    if(event.getStateChange() == ItemEvent.SELECTED)
		{
		    getGraphDisplayPanel().setFitToScreenAutomatically(true);
		}
	    else
		{
		    getGraphDisplayPanel().setFitToScreenAutomatically(false);
		}
	} catch (java.lang.Exception e) {
	}          
    }

    void JMenuItem7_actionPerformed_Interaction1(java.awt.event.ActionEvent event)
    {
	try {
	    getGraphDisplayPanel().fitToScreen();
	} catch (java.lang.Exception e) {
	}
    }
	
    class SymChange implements javax.swing.event.ChangeListener
    {
	public void stateChanged(javax.swing.event.ChangeEvent event)
	{
	    Object object = event.getSource();
	    if (object == JSlider1)
		JSlider1_stateChanged(event);
	}
    }

    void JSlider1_stateChanged(javax.swing.event.ChangeEvent event)
    {
	// to do: code goes here.
			 
	JSlider1_stateChanged_Interaction1(event);
    }

    void JSlider1_stateChanged_Interaction1(javax.swing.event.ChangeEvent event)
    {
	try {
	    // convert int->class java.lang.String
	    graphDisplayPanel.displayManager.multiplyGraphDisplayPanelSize(JSlider1.getValue());
	    JLabel3.setText(java.lang.String.valueOf(JSlider1.getValue()));
	} catch (java.lang.Exception e) {
	}
    }

    void aboutItem_actionPerformed_Interaction1(java.awt.event.ActionEvent event)
    {
	try {
	    // JAboutDialog Create and show as modal
	    {
		JAboutDialog JAboutDialog1 = new JAboutDialog(this);
		JAboutDialog1.setModal(true);
		JAboutDialog1.show();
	    }
	} catch (java.lang.Exception e) {
	}
    }

    void JMenuItem2_actionPerformed(java.awt.event.ActionEvent event)
    {
	// to do: code goes here.
			 
	JMenuItem2_actionPerformed_Interaction1(event);
    }

    void JMenuItem2_actionPerformed_Interaction1(java.awt.event.ActionEvent event)
    {
	showOptionsDialog();
    }

    /**
     * SANDSTORM 
     */
    private static final String class_tag = "MainFrame";
    private static final boolean DEBUG = true;

    private void handleQTreeStateMsgAck(QTreeStateMsg msg) {
	final String tag = class_tag + ".handleQTreeStateMsgAck";

	if(msg.getState() == QTreeStateMsg.ACK) {

	    /** 
	     * HACK: Wait for a few seconds before sending the QTreeDown?
	     * Also expecting to receive the root's guid in the ACK
	     * instead of the nodeid.
	     */

	    if(DEBUG) Debug.printtagln(tag, " received an ACK for the Qtree");
	    /*
	    try {
		if(DEBUG) Debug.printtagln(tag, "dispatching a QTreeDownMsg");
		_classifier.dispatch(new QTreeDownMsg(_gateway_node_id,
			    false, true,
			    new VisDown(true, true, true, msg.getGuid()), 
			    msg.getGuid()));
	    } catch (SinkFullException sfe) {
		sfe.printStackTrace();
	    }
	    */
	} else {
	    if(DEBUG) Debug.printtagln(tag, "received an unknown QTreeStateMsg:"+ msg + ":" +msg.getState());
	}

    }

    /**
     * If new event generated, paints the elm (blocking)?
     */
    public void handleEvent (QueueElementIF item) throws EventHandlerException {
	final String tag = class_tag + ".handleEvent";

	
	if (item instanceof QTreeStateMsg) {
	    handleQTreeStateMsgAck((QTreeStateMsg) item);

	} else if(item instanceof QTreeUpMsg) {
	    try {
		ostore.util.QuickSerializable data = ((QTreeUpMsg) item).user_data();
		if(data instanceof VisUp) {
		    if(!((VisUp) data).hasPayload())
			Debug.printtagln(tag, "received visup with msgname:"
				+((VisUp)data).getMsgName());
		    _classifier.dispatch((VisUp) data);
		} else {
		    if(DEBUG) Debug.printtagln(tag, "received a userdata of type:"
			    +data.getClass().getName()
			    +" in a QTreeUpMsg that isn't a QueueElementIF");
		}
	    } catch(ostore.util.QSException qse) {
		qse.printStackTrace();
	    } catch(SinkFullException sfe) {
		sfe.printStackTrace();
	    }
	} else if (item instanceof NodeEvent) {
	    graphDisplayPanel.handleNodeEvent((NodeEvent) item);
	} else if (item instanceof EdgeEvent) {
	    graphDisplayPanel.handleEdgeEvent((EdgeEvent) item);
	} else if (item instanceof StagesInitializedSignal 
		&& _gateway_node_id != null) {
	    //Send QTree State Message if gateway not null
	    QTreeStateMsg msg;
	    if(_election_bool) {
		msg = new QTreeStateMsg(_gateway_node_id,false,
			QTreeStateMsg.CONNECTION, _self_node_id, 0);
	    } else {
		    //FIXME
		msg = new QTreeStateMsg(_gateway_node_id,false,
			QTreeStateMsg.CONNECTION, _self_node_id, 0);
			//QTreeStateMsg.ROUTING, _self_node_id, 0);
	    }
	    try {
		if(DEBUG) Debug.printtagln(tag, "dispatching QTreeStateMsg to :"+_gateway_node_id);
		_classifier.dispatch(msg);
	    } catch (SinkFullException e) {
		e.printStackTrace();
	    }
	} else if (item instanceof Signal.AlarmSignal) {
	    //displayManager.refreshScreenNow () 
	    paint(getGraphics()); 
	    
	    Signal.getSignal(_self_node_id).register(Signal.ALRM, stageName, _this_sink);
	    Signal.getSignal(_self_node_id).alarm(stageName, graphDisplayPanel.displayManager.getRefreshRate());
	} else if(item instanceof OptionsMenuEvent) {
	    Debug.printtagln(tag, "handling options menu event");
	    addOptionsPanelContributor (((OptionsMenuEvent) item).getSource());
	} else if(item instanceof MainMenuEvent) {
	    Debug.printtagln(tag, "handling main menu event");
	    PacketAnalyzersMenu.add((JMenu) ((MainMenuEvent) item).getSource());
	} else if(item instanceof UpdateGraphEvent) {
	    graphDisplayPanel.displayManager.locationAnalyzer.updateGraph();
	} else if(item instanceof AnalyzerMenuEvent) {
	    graphDisplayPanel.displayManager.handleAnalyzerMenuEvent((AnalyzerMenuEvent) item);
	}
    }

    public void handleEvents (QueueElementIF items[]) throws EventHandlerException {
	final String tag = class_tag + ".handleEvents";
	for(int i=0; i<items.length; i++ ) 
	    handleEvent(items[i]);
    }

    public void init(ConfigDataIF config) throws Exception {
	final String tag = class_tag + ".init";
	
	Filter filter = null;
	String messageTypes[] = {"ostore.apps.visdemo.QTreeStateMsg",
				 "ostore.apps.visdemo.QTreeUpMsg",
	    			 "Sonar.event.NodeEvent",
	    			 "Sonar.event.EdgeEvent",
				 "Sonar.event.OptionsMenuEvent",
				 "Sonar.event.MainMenuEvent",
				 "Sonar.event.AnalyzerMenuEvent",
				 "seda.sandStorm.api.StagesInitializedSignal"
				};
	if (DEBUG) Debug.turnOn(class_tag);
	if (DEBUG) Debug.printtagln (tag, "Debug turned on ");

	_self_node_id = new NodeId(config.getString("node_id"));
	_classifier = Classifier.getClassifier ( _self_node_id);

	stageName = config.getStage().getName();
	if(config.getBoolean("Profile")) {
	    config.getManager().getProfiler().add("Sonar.MainFrame", this);
	    SinkIF sink = config.getStage().getSink();
	    _this_sink = new SinkProxy(sink, config.getManager(), 
		    config.getStage().getWrapper());
	} else 
	    _this_sink = config.getStage().getSink();

	for(int i = 0; i < messageTypes.length; i++) {
	    filter = new Filter();
	    if(!filter.requireType(Class.forName (messageTypes[i])))
		BUG ( tag + ": could not require type==" +
			messageTypes[i] + ".");
	    _classifier.subscribe (filter, _this_sink);
	}

	//also inst a display manager
	graphDisplayPanel = new GraphDisplayPanel(this); 
	graphDisplayPanel.planetlab = config.getBoolean("planetlab");

	//initialize icons, frame
	inst();

	//TESTING
	setVisible(true);
	Signal.getSignal(_self_node_id).register(Signal.ALRM, stageName, _this_sink);
	Signal.getSignal(_self_node_id).alarm(stageName, ALARM_SLEEP);


	//GATEWAY contact below
	if(config.getString("gateway") != null) {
	    _gateway_node_id = new NodeId(config.getString("gateway"));
	} else {
	    _gateway_node_id = null;
	}

	if(_gateway_node_id != null) {
	    if(DEBUG) Debug.printtagln(tag, "gateway node is : "+_gateway_node_id.toString());
	}
	_election_bool = config.getBoolean("election");
	if(DEBUG) Debug.printtagln(tag, "election is :"+_election_bool);

	/*
	 if((gateway_node = configuration.getGateway()) != null) {
	  MainClass.out("Gateway Node:"+gateway_node.toString());
	  self_nodeid = new NodeId(port, hostAddress);
	  QTreeStateMsg msg = null;
	  
	  if(configuration.getElection()) {
	      packetSender.setGatewayNode(gateway_node);
	      msg = new QTreeStateMsg(gateway_node,false,
		QTreeStateMsg.CONNECTION, self_nodeid);
	  } else {
	      msg = new QTreeStateMsg(gateway_node,false,
		QTreeStateMsg.ROUTING, self_nodeid);
	  }
	  packetSender.sendMessage(msg);
	  }
	*/
    }
    private NodeId _gateway_node_id;
    private boolean _election_bool;

    private long ALARM_SLEEP = 100;

    // methods for ProfilableIF
    public int profileSize()
    {
	return _this_sink.size();
    }
    // end methods for ProfilableIF
    
    // methods for EventHandlerIF
    public void destroy () throws Exception { }
    // end methods for EventHandlerIF
    
    private static final void BUG (String msg) {
	Carp.die (msg + " This is a bug in " + class_tag + ". " +
		"Please let Jason know about it ASAP." );
    }

    private NodeId _self_node_id;
    private String stageName;
    private Classifier _classifier;
    private SinkIF _this_sink;
   
    /**
     * MainClass methods
     */
    protected Vector optionPanelContributors = new Vector();
    public void addOptionsPanelContributor(Object pContributor) 
    { 
	optionPanelContributors.add(pContributor); 
    } 

    public void removeOptionsPanelContributor(Object pContributor) 
    { 
	optionPanelContributors.remove(pContributor); 
    }

    public void showOptionsDialog() 
    { 
	TabbedDialog optionsDialog = new TabbedDialog("sonar options");		 
	ActivePanel currentPanel = null; 
	Object contributor; 

	for(Enumeration e = optionPanelContributors.elements(); e.hasMoreElements();) { 
	    contributor = e.nextElement(); 
	    if(contributor instanceof DisplayManager) 
		currentPanel = ((DisplayManager) contributor).getOptionsPanel();
	    else if(contributor instanceof PacketAnalyzer)
		currentPanel = ((PacketAnalyzer) contributor).getOptionsPanel();

	    if(currentPanel != null) 
		optionsDialog.addActivePanel(currentPanel.getTabTitle(), currentPanel);
	} 
	optionsDialog.setModal(false); 
	optionsDialog.show(); 
    } 
}
