/**
 * CapacityAnalyer.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package Sonar.PacketAnalyzers;

import Sonar.*;
import Sonar.util.*;
import Sonar.event.*;
import Sonar.Dialog.*;

import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.zip.*;

import ostore.util.NodeId;
import ostore.tapestry.impl.RouteMsg;
import ostore.tapestry.impl.RouteObjDestMsg;

/**
 * @author   Jason Lee <jlee81@cs.berkeley.edu>
 */

public class CapacityAnalyzer extends PacketAnalyzer implements Runnable {

    protected TwoKeyHashtable proprietary_edge_info;
    protected Thread del_edge_thread;
    
    public CapacityAnalyzer() {
	proprietary_edge_info = new TwoKeyHashtable();
	del_edge_thread = new Thread(this);
	del_edge_thread.start();
    }

    /*
     * Modifies the size of edge width displayed
     */
    public void run() {
	EdgeInfo tempEdge = null;
	while (del_edge_thread == Thread.currentThread()) {
	    long currentTime = Calendar.getInstance().getTime().getTime();
	    for(Enumeration edges = proprietary_edge_info.elements(); 
		    edges.hasMoreElements();) {
		tempEdge = (EdgeInfo) edges.nextElement();
		if(currentTime - MainClass.edgeEventLongevity > tempEdge.getTime()) {
		    if(tempEdge.instances > 1) {
		        tempEdge.instances--;
			MainClass.displayManager.modifyEdgeWidth(tempEdge);
		    } else {
			MainClass.objectMaintainer.triggerEdgeDeletedEvent(
			    new EdgeEvent(this, tempEdge, Calendar.getInstance().getTime()));
			proprietary_edge_info.remove(tempEdge.getSource(), 
				tempEdge.getDest()); 
		    }
		}
	    }
	    /*
	    try {
	        Thread.sleep(2000);
		//MainClass.out("size: "+proprietary_edge_info.size());
	    } catch (InterruptedException ie) { }
	    */
	}
    }
    
    public void packetReceived(PacketEvent e) {
	ostore.tapestry.impl.RouteMsg rt_msg = null;
	NodeId sender = null;
	NodeId recp = null;

	if(MainClass.analyzer_menu.show_along_tapestry) {
	    sender = e.getSender(); 
	    recp   = e.getRecipient(); 
	} else if(e.hasPayload()) {
	    if(e.getPacket() instanceof ostore.apps.visdemo.RoutingTableMessage) return;
	    Object obj = e.getPacket();
	    if(obj instanceof RouteMsg) {
		sender = MainClass.objectMaintainer.guid_to_node(((RouteMsg) e.getPacket()).src);
		recp = MainClass.objectMaintainer.guid_to_node(((RouteMsg) e.getPacket()).dest);
	    } else if(obj instanceof RouteObjDestMsg) {
		sender = MainClass.objectMaintainer.guid_to_node(((RouteObjDestMsg) e.getPacket()).src);
		recp = MainClass.objectMaintainer.guid_to_node(((RouteObjDestMsg) e.getPacket()).dest);
	    }
	} else if(!MainClass.analyzer_menu.show_along_tapestry) {
	    if(e.getMsgType().equals("ostore.apps.visdemo.RoutingTableMessage")) 
		return;

	    sender = MainClass.objectMaintainer.guid_to_node(e.getSourceGuid());
	    recp = MainClass.objectMaintainer.guid_to_node(e.getDestGuid());
	}

	/*
	try {
	if(e.getPacket() instanceof RouteMsg) {
	   RouteMsg rmsg = (RouteMsg) e.getPacket();
	   if(rmsg.user_data() instanceof ostore.tapestry.channel.ICMPPingMsg) {
	    MainClass.out("ICMPPing msg from "+sender+" to "+recp);
	    MainClass.out("src:"+rmsg.src+" and dest:"+rmsg.dest);
	   }
	   if(rmsg.user_data() instanceof ostore.tapestry.channel.ICMPPongMsg) {
	    MainClass.out("ICMPPong msg from "+sender+" to "+recp);
	    MainClass.out("src:"+rmsg.src+" and dest:"+rmsg.dest);
	   }
	} 
	} catch (Exception e1) { e1.printStackTrace();  }
	*/
	if(sender != null && recp != null) {
	    EdgeInfo edgeInfo = null;
	    if((edgeInfo = (EdgeInfo)
	        proprietary_edge_info.get(sender, recp)) != null) {
		edgeInfo.instances++;
		edgeInfo.modifyTime(Calendar.getInstance().getTime().getTime()); 
		//MainClass.displayManager.modifyEdgeWidth(sender, recp, edgeInfo.getInstances(), "CapacityAnalyzer");
		MainClass.displayManager.modifyEdgeWidth(edgeInfo);
	    } else {
		edgeInfo = new EdgeInfo(sender, recp,"CapacityAnalyzer", 
			    Calendar.getInstance().getTime().getTime());
		proprietary_edge_info.put(sender, recp, edgeInfo);
		MainClass.objectMaintainer.triggerEdgeCreatedEvent(new 
		    EdgeEvent(this, edgeInfo, Calendar.getInstance().getTime()));
	    }
	}
    }

}
