/**
 * Identifier.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package Sonar.PacketAnalyzers;

import Sonar.*;
import Sonar.util.*;
import Sonar.event.*;
import Sonar.Dialog.*;

import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.zip.*;

import ostore.util.NodeId;
import ostore.tapestry.impl.RouteMsg;

import ostore.inner.CreateObjectRespMsg;
import ostore.rp.InnerRingCreateReqMsg;
import ostore.archive.dissemination.DisseminateFragMsg;
import ostore.archive.dissemination.DisseminateFragAckMsg;
import ostore.rp.CreateObjectReqMsg;
import ostore.rp.InnerRingKeyReqMsg;
import ostore.rp.InnerRingKeyRespMsg;

/**
 * Registered in MainClass and identifies the nodes by listening to 
 * specified messages. However, it's role is redundant, and should be used
 * if the other identifying anaylzers are not instantiated.
 *      
 * @author   Jason Lee <jlee81cs.berkeley.edu>
 * 
 */

public class Identifier extends PacketAnalyzer { 
    public Identifier() { } 

    public void packetReceived(PacketEvent e) {
	int offset[] = {0};
	int typeData = -1;
	RouteMsg rt_msg = null;

	//TODO : to be implemented
	if(!e.hasPayload()) return;

	if(e.getPacket() instanceof RouteMsg) {
	    rt_msg = (RouteMsg) e.getPacket(); 
	} 

	try { 
	    if(rt_msg != null && rt_msg.user_data() != null) {
		if(rt_msg.user_data() instanceof CreateObjectReqMsg) { //rp
		    CreateObjectReqMsg msg = (CreateObjectReqMsg) 
		    	rt_msg.user_data();
		    ostore.rp.CreateObjectReq req = msg.req();
		    //NodeId rp = (NodeId) MainClass.objectMaintainer.guid_to_node(rt_msg.dest);
		    NodeId client = (NodeId) 
		    	MainClass.objectMaintainer.guid_to_node(req.client);
		    if(client != null) {
			MainClass.objectMaintainer.identifyNode(client, 1);
		    }
		} else if(rt_msg.user_data() instanceof InnerRingCreateReqMsg) {
		    NodeId innerRing = MainClass.objectMaintainer.guid_to_node(
			    rt_msg.dest);
		    if(innerRing != null)
			MainClass.objectMaintainer.identifyNode(innerRing, 2);
		} else if(rt_msg.user_data() instanceof DisseminateFragMsg) {
		    NodeId storageServer = 
		    	MainClass.objectMaintainer.guid_to_node(rt_msg.dest);
		    if(storageServer != null) 
			MainClass.objectMaintainer.identifyNode(
				storageServer,3);
		}
	    }
	} catch (Exception qse) {
		qse.printStackTrace();
	}
    }
}
	
