/**
 * EdgeEvent.java
 *
 * Copyright (c) 2002 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package Sonar.event;

import Sonar.*;
import Sonar.util.EdgeInfo;
import ostore.util.NodeId;
import java.util.*;
import seda.sandStorm.api.QueueElementIF;

/**
 * This event is triggered every time an edge is created or deleted.
 *
 * @author Kamin Whitehouse <kamin@cs.berkeley.edu>
 */
public class EdgeEvent extends java.util.EventObject implements QueueElementIF 
{
    protected NodeId sourceNodeId;
    protected NodeId destinationNodeId;
    protected Date time;//and the time the event was generated
    //protected String type; //for special events in oceanstore
    protected EdgeInfo edgeInfo;
    protected boolean typeSpecified;

    public final static int CREATE = 0;
    public final static int DELETE = 1;
    public final static int MODIFY_WIDTH = 2;
    protected int event_type = CREATE;

    public int getEventType() { return event_type; }
    
    //*****---CONSTRUCTOR---******//
    public EdgeEvent(Object source, EdgeInfo edge, Date pTime) {
	super(source);
	sourceNodeId = edge.getSource();
	destinationNodeId = edge.getDest();
	time = pTime;

	if(!edge.getType().equals("no type")) {
	    typeSpecified = true;
	} else {
	    typeSpecified = false;
	}
	edgeInfo = edge;
    }

    public EdgeEvent(Object source, EdgeInfo edge, Date pTime, int event_type) {
	this(source, edge, pTime);
	this.event_type = event_type; 
    }
    //*****---CONSTRUCTOR---******//
    
	
	
    //*****---Get Functions---******//
    public NodeId getSourceNodeId(){return sourceNodeId;} 
    public NodeId getDestinationNodeId(){return destinationNodeId;}
    public EdgeInfo getEdgeInfo() { return edgeInfo; }
    
    public boolean hasType() {return typeSpecified; }
    //public String getType() {return type; }    
    public Date getTime() { return time; }
    //*****---Get Functions---******//
	
}
