/**
 * PaintTools.java
 *
 * Copyright (c) 2002 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
package Sonar.util; 

import java.awt.Color;
import java.awt.Graphics;
import java.lang.Math;
import java.util.Hashtable;

import Sonar.DisplayManager;
import Sonar.PacketAnalyzers.GridLocationAnalyzer;

import ostore.util.NodeId;


/**
 * To be extended by DisplayManager. Holds specific responses to message events.
 *
 * @author	Jason Lee <jlee81@cs.berkeley.edu>
 */

public class PaintTools {
    protected int max_edge_width = 10;

    public GridLocationAnalyzer locationAnalyzer = null;
    protected boolean show_traffic = false;

    protected boolean show_ICMP = true; 
    protected boolean show_create_obj = true;
    protected boolean show_ir_create = true;
    protected boolean show_disseminate_frag = true;
    protected boolean show_update = true;
    protected boolean show_heartbeat = true;
    protected boolean show_datadrivenhb = true;
    protected boolean show_patchwork = true;
    protected boolean show_imap = true;
    protected boolean show_req_frag = true;
    public Color custom_color = Color.yellow;
    
    public void paintTypedEdge(Hashtable typeToEdge, int x1, int y1, int x2, int y2, Graphics g){
	EdgeInfo typedEdge = null;
	if(!show_traffic) {
	    //*********************TRACEROUTE**************************//
	    //Show overlap
	    if((typedEdge = (EdgeInfo) typeToEdge.get("ostore.tapestry.channel.ICMPPingMsg")) != null && show_ICMP) {
		g.setColor(Color.yellow);
		paintWeightedEdge(x1, y1, x2, y2, typedEdge, g);
	    } 
	    if ((typedEdge = (EdgeInfo) typeToEdge.get("ostore.tapestry.channel.ICMPPongMsg")) != null && show_ICMP) {
		g.setColor(Color.blue); 
		paintWeightedEdge(x1, y1, x2, y2, typedEdge, g);
		//*********************TRACEROUTE**************************//
		
	
		//*****************Remote Replica Messages ****************//
	    } else if((typedEdge = (EdgeInfo) typeToEdge.get("ostore.replica.RemoteReplicaUpdateReq")) != null) {
		g.setColor(Color.orange); 
		paintWeightedEdge(x1, y1, x2, y2, typedEdge, g);
	    } else if((typedEdge = (EdgeInfo) typeToEdge.get("ostore.replica.RemoteReplicaCreateReq")) != null) {
		g.setColor(Color.orange); 
		paintWeightedEdge(x1, y1, x2, y2, typedEdge, g);
		//*****************Remote Replica Messages ****************//
		
		
	    } else if((typedEdge = (EdgeInfo) typeToEdge.get("ostore.rp.CreateObjectReqMsg")) != null && show_create_obj) {
		//Set Time Limit to Longevity of these edges
		g.setColor(Color.yellow);
		//g.drawLine(x1, y1, x2, y2);
		paintWeightedEdge(x1, y1, x2, y2, typedEdge, g);

	    } else if((typedEdge = (EdgeInfo) typeToEdge.get("ostore.rp.InnerRingCreateReqMsg")) != null && show_ir_create) {
		g.setColor(Color.lightGray);
		//g.drawLine(x1, y1, x2, y2);
		paintWeightedEdge(x1, y1, x2, y2, typedEdge, g);
	    } else if((typedEdge = (EdgeInfo) typeToEdge.get("ostore.archive.dissemination.DisseminateFragMsg")) != null && show_disseminate_frag) {
		g.setColor(Color.red);
		//g.drawLine(x1, y1, x2, y2);
		paintWeightedEdge(x1, y1, x2, y2, typedEdge, g);
	    } else if((typedEdge = (EdgeInfo) typeToEdge.get("ostore.inner.UpdateReqMsg")) != null && show_update) {
		g.setColor(Color.magenta);
		//g.drawLine(x1, y1, x2, y2);
		paintWeightedEdge(x1, y1, x2, y2, typedEdge, g);
	    } else if((typedEdge = (EdgeInfo) typeToEdge.get("ostore.archive.RequestFragMsg")) != null && show_req_frag) {
		g.setColor(Color.cyan);
		//g.drawLine(x1, y1, x2, y2);
		paintWeightedEdge(x1, y1, x2, y2, typedEdge, g);
	    } else if((typedEdge = (EdgeInfo) typeToEdge.get("ostore.archive.RequestFragResponseMsg")) != null && show_req_frag)  {
		g.setColor(Color.cyan);
		paintWeightedEdge(x1, y1, x2, y2, typedEdge, g);
	    } else if((typedEdge = (EdgeInfo) typeToEdge.get("ostore.tapestry.impl.PublishMsg")) != null && show_heartbeat) {
		if(locationAnalyzer != null && locationAnalyzer.getCenteredNodeId().equals(typedEdge.getOrigin())) { 
		    g.setColor(Color.yellow);
		    //g.drawLine(x1, y1, x2, y2);
		    paintWeightedEdge(x1, y1, x2, y2, typedEdge, g);
		}
	    } else if((typedEdge = (EdgeInfo) typeToEdge.get("InputAnalyzerMsg")) != null) {
		g.setColor(custom_color);
		paintWeightedEdge(x1, y1, x2, y2, typedEdge, g);

		/* HeartBeat  */
	    } else if((typedEdge = (EdgeInfo) typeToEdge.get("ostore.archive.repair.DataDrivenServerHbMsg")) != null && show_datadrivenhb) {
		g.setColor(Color.yellow);
		paintWeightedEdge(x1, y1, x2, y2, typedEdge, g);
	    } else if((typedEdge = (EdgeInfo) typeToEdge.get("ostore.tapestry.patchwork.PatchworkFaultMsg")) != null && show_patchwork) {
		g.setColor(Color.yellow);
		paintWeightedEdge(x1, y1, x2, y2, typedEdge, g);
	    } else if((typedEdge = (EdgeInfo) typeToEdge.get("ostore.apps.mail.core.test.StartPrefetch")) != null && show_imap) {
		g.setColor(Color.yellow);
		paintWeightedEdge(x1, y1, x2, y2, typedEdge, g);
	    } 
	    
	} else {
	    if((typedEdge = (EdgeInfo) typeToEdge.get("CapacityAnalyzer")) != null) {
		paintWeightedEdge(x1, y1, x2, y2, typedEdge, g);
	    }
	} 
    }
    
    /**
     * Paints an edge with a desired width.
     */
    public void paintWeightedEdge(int x1, int y1, int x2, int y2, EdgeInfo e, Graphics g) {
	int width = e.instances; 
	if(width > max_edge_width) { 
	   width = max_edge_width; 
	} else if(width == 1) {
	    width = 2;
	} else if(width == 0) {
	    //out("DM> Width of 0!");
	}
	
	if(show_traffic) {
	    g.setColor(Color.red);
	}
	int x[] = new int[4];
	int y[] = new int[4];
	if((x1 == x2) && (y1==y2)) { //On top of the other
	    //Do Nothing!
	} else if(y1==y2) { //vertical  
	    if(x1 < x2) {
		x[0] = x1;
		x[1] = x1;
		x[2] = x2;
		x[3] = x2;
	    } else {
		x[0] = x2;
		x[1] = x2;
		x[2] = x1;
		x[3] = x1;
	    }
	    y[0] = y1 + width;
	    y[1] = y1 - width;
	    y[2] = y1 - width;
	    y[3] = y1 + width;
	    g.fillPolygon(x, y, 4);
	} else if(x1==x2) { //horizontal
	    if(y1 < y2) {
		y[0] = y1;
		y[1] = y1;
		y[2] = y2;
		y[3] = y2;
	    } else {
		y[0] = y2;
		y[1] = y2;
		y[2] = y1;
		y[3] = y1;
	    }
	    x[0] = x1 + width;
	    x[1] = x1 - width;
	    x[2] = x1 - width;
	    x[3] = x1 + width;
	    g.fillPolygon(x, y, 4);
	} else {
	    //perpendicular line @ each vertex
	    double slope = ((double) y2 - (double) y1) / ((double) x2 - (double) x1);

	    if(slope == 0) {
		//out("DM> Divide by 0! (1) ");
	    }
	    double inverse = - 1 / slope;
	    double result1 = width;
	    result1 = Math.sqrt(width * width / (inverse * inverse + 1));

	    int x_temp, y_temp;
	    x_temp = new Double(result1 + x1).intValue(); 
	    y_temp = new Double(inverse * (x_temp - x1)).intValue() + y1;

	    if(x_temp < x1) {
		x[0] = x_temp;
		y[0] = y_temp;
		x[1] = 2 * x1 - x[0]; 
		y[1] = 2 * y1 - y[0];
	    } else if(x_temp > x1) {
		x[1] = x_temp;
		y[1] = y_temp;
		x[0] = 2 * x1 - x[1]; 
		y[0] = 2 * y1 - y[1];
	    }

	    if(x_temp == x1) {
		x[0] = x1;
		y[0] = y1 - width;
		x[1] = x1;
		y[1] = y1 + width;

		x[2] = x2;
		y[2] = y2 + width;
		x[3] = x2;
		y[3] = y2 - width;

	    } else {
		x_temp = new Double(x2 + result1).intValue();
		y_temp = new Double(inverse * (x_temp - x2)).intValue() + y2;

		if(x_temp > x2) {
		    x[2] = x_temp;
		    y[2] = y_temp;
		    x[3] = 2 * x2 - x[2];
		    y[3] = 2 * y2 - y[2];
		} else {
		    x[3] = x_temp;
		    y[3] = y_temp;
		    x[2] = 2 * x2 - x[3];
		    y[2] = 2 * y2 - y[3];
		}
	    }

	    if((x1 < x2) && (y1 < y2)) { //upper left to lower right?
		if(show_traffic) g.setColor(Color.lightGray);
		g.fillPolygon(x, y, 4);
	    } else if((x1 > x2) && (y1 < y2)) { // upper right to lower left?
		if(show_traffic) g.setColor(Color.orange);
		g.fillPolygon(x, y, 4);
	    } else if((x1 > x2) && (y1 > y2)) { //lower right to upper left?
		if(show_traffic) g.setColor(Color.blue);
		g.fillPolygon(x, y, 4);
	    } else if((x1 < x2) && (y1 > y2)) { //lower left to upper right?
		if(show_traffic) g.setColor(Color.black);
		g.fillPolygon(x, y, 4);
	    }
	}
    }
	    
    /* Shows whether a tapestry edge is bidirectional or only in one direction.
     *
     */
    public void paintArrow(int x1, int y1, int x2, int y2, Graphics g, int arrowColor) {
	boolean displayEndArrows = false; //Only Defined for nonvertical nor horizontal lines
	double arrowLen = 10;
	double edgeLen = Math.sqrt(Math.pow(x1-x2, 2) + Math.pow(y1-y2, 2));      

	if(y1==y2 && x2!=x1) {
	    //Horizontal Line

	    //40%
	    int x_arrow[] = new int[4];
	    int y_arrow[] = new int[4];
	    double x40;

	    if(x2 < x1) {
		x40 = 0.4*edgeLen + x2;
		x_arrow[0] = (new Double(x40 - arrowLen)).intValue();
	    } else {
		x40 = 0.6*edgeLen + x1;
		x_arrow[0] = (new Double(x40 + arrowLen)).intValue();
	    }
	    x_arrow[1] = (new Double(x40)).intValue();
	    x_arrow[2] = (new Double(x40)).intValue();
	    x_arrow[3] = x_arrow[0];

	    y_arrow[0] = y1;
	    y_arrow[1] = y1+ (new Double(arrowLen)).intValue();
	    y_arrow[2] = y1- (new Double(arrowLen)).intValue();
	    y_arrow[3] = y1;
	    paintArrowHelper(x_arrow, y_arrow, arrowColor, g);

	} else if( x2 != x1) {
	    double slope = ((double) y2 - (double) y1) / ((double) x2 - (double) x1);
	    double inverse = - 1 / slope;

	    //End Arrows	  
	    double x3_dble = Math.sqrt(slope*slope+1);	  
	    if(x2 < x1) {
		x3_dble = arrowLen/x3_dble + x2;
	    } else {
		x3_dble = (edgeLen - arrowLen) / x3_dble + x1;
	    }   
	    double y3_dble = slope * (x3_dble - x2) + y2;

	    int x3 = (new Double(x3_dble)).intValue();
	    int y3 = (new Double(y3_dble)).intValue();

	    double x4_dble = Math.sqrt(inverse*inverse+1);
	    double x5_dble;
	    int x_arrow[] = new int[4];
	    int y_arrow[] = new int[4];

	    if(displayEndArrows) {
		x4_dble = arrowLen / x4_dble + x3_dble;
		x5_dble = 2* x3_dble - x4_dble;

		//For X points along above the edge, below the edge,
		//on the inverse
		x_arrow[0] = x2;
		x_arrow[3] = x2;
		x_arrow[1] = (new Double(x4_dble)).intValue();
		x_arrow[2] = (new Double(x5_dble)).intValue();

		y_arrow[0] = y2;
		y_arrow[3] = y2;
		y_arrow[1] = (new Double(inverse*(x4_dble - x3_dble) + y3_dble)).intValue();
		y_arrow[2] = (new Double(inverse*(x5_dble - x3_dble) + y3_dble)).intValue();
		//Draw polygon, and fill
		paintArrowHelper(x_arrow, y_arrow, arrowColor, g);
	    }
	    //40% of Edge Length, place the arrows	  
	    double x40_dble = Math.sqrt(slope*slope+1);
	    double x40_dble2 = x40_dble;
	    if(x2 < x1) {
		x40_dble = (edgeLen*.4) /x40_dble + x2;
		x40_dble2 = (edgeLen*.4-arrowLen)/x40_dble2 + x2;	      
	    } else {
		x40_dble = (edgeLen *.6) / x40_dble + x1;
		x40_dble2 = (edgeLen *.6+arrowLen) / x40_dble2 + x1;
	    }
	    double y40_dble = slope * (x40_dble - x2) + y2;
	    double y40_dble2 = slope * (x40_dble2 - x2) + y2;

	    int x40 = (new Double(x40_dble2)).intValue();
	    int y40 = (new Double(y40_dble2)).intValue();

	    x4_dble = Math.sqrt(inverse*inverse+1);
	    x4_dble = arrowLen / x4_dble + x40_dble;
	    x5_dble = 2* x40_dble - x4_dble;

	    x_arrow[0] = x40;
	    x_arrow[3] = x40;
	    x_arrow[1] = (new Double(x4_dble)).intValue();
	    x_arrow[2] = (new Double(x5_dble)).intValue();

	    y_arrow[0] = y40;
	    y_arrow[3] = y40;
	    y_arrow[1] = (new Double(inverse*(x4_dble - x40_dble) + y40_dble)).intValue();
	    y_arrow[2] = (new Double(inverse*(x5_dble - x40_dble) + y40_dble)).intValue();	  
	    paintArrowHelper(x_arrow, y_arrow, arrowColor, g);  
	} else if(y1 == y2) {
	    //Error by Analyzer who prompted Edge
	    //g.setColor(Color.red);
	    //g.drawString("?self-edge?", x2, y2);
	} else { 
	    //Vertical Line
	    //40%
	    int x_arrow[] = new int[4];
	    int y_arrow[] = new int[4];
	    double y40;

	    if(y2 < y1) {
		y40 = 0.4*edgeLen + y2;
		y_arrow[0] = (new Double(y40 - arrowLen)).intValue();
	    } else {
		y40 = 0.6*edgeLen + y1;
		y_arrow[0] = (new Double(y40 + arrowLen)).intValue();
	    }
	    y_arrow[1] = (new Double(y40)).intValue();
	    y_arrow[2] = (new Double(y40)).intValue();
	    y_arrow[3] = y_arrow[0];

	    x_arrow[0] = x1;
	    x_arrow[1] = x1+ (new Double(arrowLen)).intValue();
	    x_arrow[2] = x1- (new Double(arrowLen)).intValue();
	    x_arrow[3] = x1;
	    paintArrowHelper(x_arrow, y_arrow, arrowColor, g);
	}
    }

    public void paintArrowHelper (int x_arrow[], int y_arrow[], int arrowColor, Graphics g) {
	g.setColor(Color.black);
	g.drawPolyline( x_arrow, y_arrow, 4); //Outline
	switch(arrowColor) {
	    case 0:
		g.setColor(Color.white);
		break;
	    case 1:
		g.setColor(Color.blue);
		break;
	    case 2:
		g.setColor(Color.cyan);
		break;
	    case 3:
		g.setColor(Color.yellow);
		break;
	    case 4:
		g.setColor(Color.gray);
		break;
	    case 5:
		g.setColor(Color.pink);
		break;
	    case 6:
		g.setColor(Color.magenta);
		break;
	    case 7:	
		g.setColor(Color.orange);
		break;
	    case 8:
		g.setColor(Color.lightGray);
		break;
	    case 9:
		g.setColor(Color.red);
	}
	g.fillPolygon( x_arrow, y_arrow, 4); 
    }

    public class ArrowInfo {
	public int[] xcoords;
	public int[] ycoords;

	public ArrowInfo() {
	    xcoords = new int[4];
	    ycoords = new int[4];
	}
    }
}
