/*
 * Main.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package soss.core;
import seda.sandStorm.api.SandstormConfigIF;
import seda.sandStorm.api.StageIF;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.ManagerIF;
import seda.sandStorm.api.NoSuchStageException;
import seda.sandStorm.api.SignalMgrIF;
import seda.sandStorm.api.ProfilerIF;
import java.util.HashMap;
import java.util.Map;

public class Manager implements ManagerIF {

    public Manager (SandstormConfigIF config) {
	_config = config;
	_stages = new HashMap ();
    }

    public void addStage(String stagename, StageIF stage) {
	_stages.put (stagename, stage);
    }

    public StageIF getStage(String stagename) throws NoSuchStageException {
	StageIF result = (StageIF) _stages.get (stagename);
	if (result == null)
	    throw new NoSuchStageException ();
	return result;
    }

    public StageIF createStage(String stagename, EventHandlerIF eventHandler,
	    String initargs[]) throws Exception {
	throw new IllegalArgumentException ();
    }

    /**
     * Returns a handle to the system signal interface.
     */
    public SignalMgrIF getSignalMgr() {
	throw new IllegalArgumentException ();
    }

    /**
     * Returns a handle to the system profiler. 
     */
    public ProfilerIF getProfiler() {
	throw new IllegalArgumentException ();
    }

    /**
     * Returns a copy of the SandstormConfigIF for this Manager. This
     * contains all of the global options used by the runtime system. Note
     * that modifying any options of this copy does not in fact change the
     * runtime parameters of the system; this is used for informational
     * purposes only.
     */
    public SandstormConfigIF getConfig() {
	return _config;
    }

    /**
     * Note that calling this slightly violates the semantics of this
     * method as defined in ManagerIF.  It does not remove all references
     * to the stage, because the soss Main stage still has a reference
     * to the node's classifier, which contains references to the sink,
     * which has a reference to the EventHandler itself. So the stage won't
     * completely disappear until the reference to the classifier is
     * destroyed (see soss.core.Main.removeNode)
     */
    public void destroyStage(String stagename) throws Exception {
	if (stagename == null) 
	    throw new NoSuchStageException("no such stage: null");
	StageIF stage = (StageIF)_stages.get(stagename);
	if (stage == null) 
	    throw new NoSuchStageException("no such stage: "+stagename);

	// destroy and remove this sucker
	_stages.remove( stagename );
	stage.destroy();
	System.out.println( stagename + ": Stage destroyed" );
    }

    private Map _stages;
    private SandstormConfigIF _config;
}
